#!/bin/bash
# script to create basic C++ wrapper for an MG5 fortran matrix element
# includes a colour-structure (ICOL) map from mg5 matrix element
# for use with VINCIA. E.g., the ICOL for structure T[4,5,6] T[7,8,9] will
# be associated with the integer array values IT[1]=456, IT[2]=789. 
file=dummy.f
baseClass=MG5matrix
if [ $# -ge 1 -a "$1" != "-h" -a "$1" != "--help" ]; then
  file=$1
  if [ $# = 2 ]; then
    # Add MG5 prefix to name of base class if not already given, and
    # remove any .f suffix
    baseClass=`echo $2 | sed s/MG5// | sed s/\.f//`
    baseClass=MG5$baseClass
  fi
fi
if [ -r $file ] ; then
  # Process and file names
  procName=`echo $file | sed s/\.f//`  
  fFile=$procName.f
  ccFile=$procName.cc
  hFile=$procName.h
  # number of legs
  nLegs=`cat $file | grep PARAMETER | grep -m1 NEXTERNAL= | cut -d'=' -f2 | cut -d')' -f1`
  # Create list of process labels that this ME can handle (cut the file
  # at the first DATA statement to ignore duplicate Process statements)
  processList=`cat $file | sed -e '/DATA/,$d' | grep Process | sed s/"Process:"/"Process:BEGIN"/ | cut -d: -f2 | sed s/WEIGHTED/"END:"/ | cut -d: -f1 | sed s/"  "/" "/g` 
  # Create header file
  date=`date`
  year=`date "+%Y"`
  (      
  echo "// ${hFile} is a part of the VINCIA plug-in to the PYTHIA 8 generator."
  echo "// Copyright (C) $year P. Skands."
  echo "// VINCIA is licenced under the GNU GPL version 2."
  echo "// Please respect the MCnet Guidelines for Event Generator Authors and Users, "
  echo "//               http://www.montecarlonet.org/GUIDELINES"
  echo ""
  echo "// This file was created on $date by $0"
  echo "// VINCIA C++ wrapper for $file"
  echo ""
  echo "#ifndef VINCIA_${procName}_H"
  echo "#define VINCIA_${procName}_H"
  echo ""
  echo '#include "Vincia/MG5interface.h"'
  echo ""
  echo "namespace Vincia {"
  echo "using namespace Pythia8;"
  echo ""
  echo "class MG5${procName} : public ${baseClass} {"
  echo ""
  echo " public:"
  echo ""
  echo "  // Constructor"
  echo "  MG5$procName() { initHelMap(); initColMap(); };"
  echo ""
  echo "  // Destructor"
  echo "  virtual ~MG5$procName() {};"
  echo ""
  echo "  // Name"
  echo "  "virtual string name"()"  '{return "'${procName}'";};'
  echo ""
  echo "  // PDG ID multilabel"
  echo "  virtual vector< vector<int> > processList();"
  echo ""
  echo "  // Number of legs"
  echo "  virtual int nLegs() {return ${nLegs};}"
  echo ""
  echo "  // Redefine methods to define/initialise helicity and colour maps"
  echo "  virtual void initHelMap();"
  echo "  virtual void initColMap();"
  echo ""
  echo "  // Store momenta in MG5 format (from input in Pythia Vec4 format)"
  echo "  // (Must be copied in derived classes so that pSav will be the"
  echo "  // variable in the derived class, rather than the one in the base class)"
  echo "  virtual void setMom(vector<Particle> pIn, bool permute=true) {"
  echo "    for (int i=0; i<nLegs(); ++i) {"
  echo "      if (permute) {"
  echo "        pSav[i][0] = pIn[iVincia[i]].e();"
  echo "        pSav[i][1] = pIn[iVincia[i]].px();"
  echo "        pSav[i][2] = pIn[iVincia[i]].py();"
  echo "        pSav[i][3] = pIn[iVincia[i]].pz();"
  echo "      } else {"
  echo "        pSav[i][0] = pIn[i].e();"
  echo "        pSav[i][1] = pIn[i].px();"
  echo "        pSav[i][2] = pIn[i].py();"
  echo "        pSav[i][3] = pIn[i].pz();"
  echo "      }"
  echo "    }"
  echo "  }"
  echo "" 
  echo "  // setMom with vector<Vec4> as input (hardcoded for speed)"
  echo "  virtual void setMom(vector<Vec4> pIn, bool permute=true) {"
  echo "    for (int i=0; i<nLegs(); ++i) {"
  echo "      if (permute) {"
  echo "        pSav[i][0] = pIn[iVincia[i]].e();"
  echo "  	pSav[i][1] = pIn[iVincia[i]].px();"
  echo "        pSav[i][2] = pIn[iVincia[i]].py();"
  echo "        pSav[i][3] = pIn[iVincia[i]].pz();"
  echo "      } else {"
  echo "        pSav[i][0] = pIn[i].e();"
  echo "        pSav[i][1] = pIn[i].px();"
  echo "        pSav[i][2] = pIn[i].py();"
  echo "        pSav[i][3] = pIn[i].pz();"
  echo "      }"
  echo "    }"
  echo "  }"
  echo ""
  echo "  // Print momenta (in MG enumeration)"
  echo "  virtual void printMom() {"
  echo "    for (int i=0; i<nLegs(); ++i) {"
  echo '      cout<<" pMG"<<i<<setprecision(6)<<" = ";'
  echo '      for (int j = 0; j<4; ++j) cout<<" "<<setw(16)<<pSav[i][j];'
  echo "      cout<<endl;"
  echo "    }"
  echo "  }"
  echo ""
  echo "  // Wrapper for MG5 matrix element squared"
  echo "  virtual double getME2(int mcMode);"
  echo ""
  echo " protected: "
  echo ""
  echo "  // Momenta in MadGraph format: p[NLEGS][4]"
  echo "  double pSav[${nLegs}][4];"
  echo ""
  echo "}; // end class MG5${procName}"
  echo ""
  echo "} // end namespace Vincia"
  echo ""
  echo "#endif"
  echo ""
  ) > $hFile
  #------------------------------------------------------------------------
  # Create .cc file
  (
  echo "// ${ccFile} is a part of the VINCIA plug-in to the PYTHIA 8 generator."
  echo "// Copyright (C) ${year} P. Skands."
  echo "// VINCIA is licenced under the GNU GPL version 2."
  echo "// Please respect the MCnet Guidelines for Event Generator Authors and Users,"
  echo "//               http://www.montecarlonet.org/GUIDELINES"
  echo ""
  echo "// This file was created on $date by $0"
  echo "// VINCIA C++ wrapper for ${file}"
  echo ""
  echo '#include "'VinciaMG5/${hFile}'"'
  echo ""
  echo "namespace Vincia {"
  echo ""
  echo "//===================================================================="
  echo ""
  echo "// Wrapper for the MG5 FORTRAN subroutine itself"
  echo ""
  echo 'extern "C" {'
  echo "  extern void shel${procName}_(int *cMode, int *ICOL, double P1[${nLegs}][4], int *HEL, double* ANS);"
  echo "}"
  echo ""
  echo "//--------------------------------------------------------------------"
  echo ""
  echo "// initHelMap(): create map from vector of helicities to IHEL number"
  echo ""
  echo "void MG5${procName}::initHelMap() {"
  echo "  // Hardcoded helicity map extracted from .f file"
  ./mkHelMap.sh $file
  echo "}"
  echo ""
  echo "//--------------------------------------------------------------------"
  echo ""
  echo "// initColMap(): create map from vector of MG indices to ICOL numbers"
  echo ""
  echo "void MG5${procName}::initColMap() {"
  echo "  // Hardcoded colour-index map extracted from .f file"
  ./mkColMap.sh $file
  echo "}"
  echo ""
  echo "//------------------------------------------------------------------"
  echo ""
  echo "// Return vector of all process labels matched by this ME"
  echo ""
  echo "vector< vector<int> > MG5${procName}::processList() {"
  echo "  vector< vector<int> > labels; labels.resize(0);"
  echo "  vector<int> thisLabel;"
  # Loop over components of procName, replacing particle names by ID codes
  # This part should be replaced by MG5 outputting a label directly.
  i=0;  
  echo "Process labels contained in this ME:" >&2 
  while (( i<${#processList} )); do
      particleID="";
      iUsed=1;
      # Skip > separator between IS and FS
      if [ "${processList:$i:1}" == ">" ]; then
	  i=$((i+1))
	  continue;	  
      elif [ "${processList:$i:5}" == "BEGIN" ]; then
	  echo "  thisLabel.resize(0);"
	  i=$((i+5))
	  continue;
      elif [ "${processList:$i:3}" == "END" ]; then	  
	  echo "" >&2
	  echo "  labels.push_back(thisLabel);"
	  i=$((i+3))
	  continue;
      elif [ "${processList:$i:1}" == " " ]; then
	  # Skip whitespace
	  i=$((i+1))
	  continue;    
      fi;
      # Identify common names, starting with longest strings first
      particleID=0;
      if [ "${processList:$i:3}" == "ve~" ]; then
	  particleID=-12;
	  iUsed=3;
      elif [ "${processList:$i:3}" == "vm~" ]; then
	  particleID=-14;
	  iUsed=3;
      elif [ "${processList:$i:3}" == "vt~" ]; then
	  particleID=-16;
	  iUsed=3;
      elif [ "${processList:$i:3}" == "mu-" ]; then
	  particleID=13;
	  iUsed=3;
      elif [ "${processList:$i:3}" == "mu+" ]; then
	  particleID=-13;
	  iUsed=3;
      elif [ "${processList:$i:3}" == "ta-" ]; then
	  particleID=15;
	  iUsed=3;
      elif [ "${processList:$i:3}" == "ta+" ]; then
	  particleID=-15;
	  iUsed=3;
      elif [ "${processList:$i:2}" == "e+" ]; then
	  particleID=-11;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "e-" ]; then
	  particleID=11;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "ve" ]; then
	  particleID=12;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "vm" ]; then
	  particleID=14;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "vt" ]; then
	  particleID=16;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "d~" ]; then
	  particleID=-1;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "u~" ]; then
	  particleID=-2;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "s~" ]; then
	  particleID=-3;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "c~" ]; then
	  particleID=-4;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "b~" ]; then
	  particleID=-5;
	  iUsed=2;
      elif [ "${processList:$i:2}" == "t~" ]; then
	  particleID=-6;
	  iUsed=2;
      elif [ "${processList:$i:1}" == "w+" ]; then
	  particleID=23;
	  iUsed=2;
      elif [ "${processList:$i:1}" == "w-" ]; then
	  particleID=-23;
	  iUsed=2;
      elif [ "${processList:$i:1}" == "d" ]; then
	  particleID=1;
      elif [ "${processList:$i:1}" == "u" ]; then
	  particleID=2;
      elif [ "${processList:$i:1}" == "s" ]; then
	  particleID=3;
      elif [ "${processList:$i:1}" == "c" ]; then
	  particleID=4;
      elif [ "${processList:$i:1}" == "b" ]; then
	  particleID=5;
      elif [ "${processList:$i:1}" == "t" ]; then
	  particleID=6;
      elif [ "${processList:$i:1}" == "g" ]; then
	  particleID=21;
      elif [ "${processList:$i:1}" == "z" ]; then
	  particleID=24;
      elif [ "${processList:$i:1}" == "h" ]; then
	  particleID=25;
      elif [ "${processList:$i:1}" == "a" ]; then
	  particleID=22;
      fi;
      if [ $particleID != 0 ]; then
	  echo "  thisLabel.push_back($particleID);"
	  echo -ne " "$particleID >&2
      fi
      i=$((i+$iUsed));
  done
  echo "  // Sanity check"
  echo "  for (int i=0; i<(int)labels.size(); ++i) {"
  echo "    if (nLegs() != (int)labels[i].size())"
  echo '      cout<<"  MG5${procName}::processList(): ERROR! inconsistent number of legs"<<endl;'
  echo "  }"
  echo "  return labels;"
  echo "}"
  echo ""
  echo "//------------------------------------------------------------------"
  echo ""
  echo "// Wrapper for MG5 ME, with variables set by the other methods"
  echo ""
  echo "double MG5${procName}::getME2(int mcMode) {"
  echo "  // Call wrapped matrix element"
  echo "  double ans;"
  echo "  shel${procName}_(&mcMode, &icolSav, pSav, &ihelSav, &ans);"
  echo "  return ans;"
  echo "}"
  echo ""
  echo "} // End namespace Vincia"
  echo ""
  ) > $ccFile
fi

