// VinciaRootGraph.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the VinciaRootGraph class 

// Include the VinciaRootGraph header
#include "VinciaRoot/VRgraph.h"

namespace Vincia {
  
  //=====================================================================
  // VRgraph Member function definitions 

  //-----------------------------------------------------------------
  // update():
  // Update graph

  bool VRgraph::update(bool normalizeIn, double normFacIn, 
                       double sigmaStat, double scatterX) { 

    // Check histogram not empty
    if (hisPtrSav[0]->GetEntries() <= 0) return false;
    
    // Update
    sumWeights   = 0.0;

    // Check if there is a nontrivial normalization factor applied
    double normFactor = hisPtrSav[0]->GetNormFactor();
    if (normFactor == 0.0) normFactor = 1.0;
    
    // Include overflow bins in normalization
    int nBinHis         = hisPtrSav[0]->GetNbinsX();
    // Add content of under- and overflow bins to sum of weights
    sumWeights += hisPtrSav[0]->GetBinContent(nBinHis+1) 
      * hisPtrSav[0]->GetBinWidth(nBinHis)*normFactor; 
    sumWeights += hisPtrSav[0]->GetBinContent(0) 
      * hisPtrSav[0]->GetBinWidth(1)*normFactor; 
    // End of histogram
    double xHisEnd      = hisPtrSav[0]->GetBinLowEdge(nBinHis) 
      + hisPtrSav[0]->GetBinWidth(nBinHis) ;

    // Allow optional rebinning according to a "rebinAxis"
    int nPointGraph     = nBinHis;
    int firstPointGraph = 0;
    if (rebin) {
      if (rebinAxisPtr != 0) {
        nPointGraph     = rebinAxisPtr->GetN();
        firstPointGraph = 0;
      } else {
        rebin = false;
      }
    }

    int nPoints  =  -1;
    // Do inner and outer loop; 
    // outer: loop over graph bins (according to rebinner or hist, as desired)
    // inner: loop over histogram bins (starting where previous left off)
    // Note: check for half-bins or smaller. Not reliable. Probably only write 
    // that once (initialize to false, then write true, or use Pythia's INFO)
    int iBinFirst = 0;
    int iBinLast  = nBinHis;
    for (int iPointGraph = 0; iPointGraph <= nPointGraph; ++iPointGraph) {
      // Start out with assumption of no rebinning: his points = graph points
      double xGraph    = hisPtrSav[0]->GetBinCenter(min(iPointGraph,nBinHis));
      double exGraphLo = xGraph - hisPtrSav[0]->GetBinLowEdge(min(iPointGraph,nBinHis));
      double exGraphHi = hisPtrSav[0]->GetBinWidth(min(iPointGraph,nBinHis)) - exGraphLo;
      bool doWeightedX = false;
      if (rebin) {
        double yDum(0.);
        int iGood = rebinAxisPtr->GetPoint(iPointGraph,xGraph,yDum);
        if (iGood == -1) continue;
        exGraphHi = abs(rebinAxisPtr->GetErrorXhigh(iPointGraph));
        exGraphLo = abs(rebinAxisPtr->GetErrorXlow(iPointGraph));        
        iBinFirst = max(1,iBinFirst-1);
        iBinLast  = nBinHis;
	// Check if we should calculate the weighted x mean
	if (abs(exGraphHi - exGraphLo) > 1e-3*(exGraphHi + exGraphLo))
	  doWeightedX = true;
      } else {
        // For histograms, first bin is underflow, skip here (included in norm above)
        if (iPointGraph == 0) continue;
      }
      double yGraph  = 0.0;
      double xMean   = (doWeightedX) ? 0.0 : xGraph;
      double err2lo  = 0.0;
      double err2hi  = 0.0;
      double eySysHi = 0.0;
      double eySysLo = 0.0;
      double sumW    = 0.0;
      // Histogram (TH1)
      // Bin # 0         : underflow
      // Bin # [1,nbins] : contens
      // Bin # nbins + 1 : overflow
      for (int iBinHis = iBinFirst; iBinHis <= iBinLast; ++iBinHis) {
        double xHis    = hisPtrSav[0]->GetBinCenter(iBinHis);
        double exHisLo = xHis - hisPtrSav[0]->GetBinLowEdge(iBinHis);
        double exHisHi = hisPtrSav[0]->GetBinWidth(iBinHis) - exHisLo;
        if (xHis + exHisHi < xGraph - exGraphLo) {
          iBinFirst = iBinHis; 
          continue;
        }
        if (xHis - exHisLo > xGraph + exGraphHi) { 
          break; 
        }
        double w = 1.0 ;
        if ( (exHisHi+exHisLo) > 0) 
          w = (min(xHis + exHisHi,xGraph + exGraphHi) 
               - max(xHis - exHisLo, xGraph - exGraphLo))
            / (exHisHi + exHisLo);        
        if (w <= 0.001) continue;
        double yHis = normFactor * hisPtrSav[0]->GetBinContent(iBinHis);
        // Skip zero bins (if zero skipping)
        if (zeroSkippingSav && yHis == 0.0) continue;
        sumW += w;
        yGraph += w * yHis;	
	if (doWeightedX) xMean += w * xHis * yHis;
        err2lo += pow(w * normFactor * hisPtrSav[0]->GetBinError(iBinHis),2.); 
        err2hi += pow(w * normFactor * hisPtrSav[0]->GetBinError(iBinHis),2.); 
        // Sum up weights for normalization
        sumWeights += w * yHis * (exHisHi+exHisLo);
        // Got uncertainty variations?
        if (hasOuter()) {
          double eyOutHi(0.), eyOutLo(0.);
          for (int iHis = 1; iHis < int(hisPtrSav.size()); ++iHis) {
            // Check if same bins, else give error and switch off uncertainties
            if (abs(hisPtrSav[iHis]->GetBinCenter(iBinHis) - xHis) > TINY || 
                abs(xHis - hisPtrSav[iHis]->GetBinLowEdge(iBinHis) - exHisLo) > TINY ||
                abs(hisPtrSav[iHis]->GetBinWidth(iBinHis) - (exHisLo+exHisHi)) > TINY) {
              cout<<"(VRgraph): incompatible bins for uncertainty histogram "
                  <<iHis<<" for "<<hisPtrSav[0]->GetName()<<endl;
              hasOuterSav = false;
              break;
            }
            // Skip if this variation was not filled
            if (hisPtrSav[iHis]->GetEntries() <= 0) continue;
            // Compute error bracket
            double yBand = normFactor * hisPtrSav[iHis]->GetBinContent(iBinHis);
            if (yBand > yHis) eyOutHi = max(eyOutHi,yBand-yHis);
            else eyOutLo = max(eyOutLo,yHis-yBand);            
          }
	  // Add theory/sys uncertainties linearly
	  // (E.g., prevents rebinning from artificially improving)
          eySysHi += w * eyOutHi;
          eySysLo += w * eyOutLo;
        }
      }
      // Don't store points that have zero weight.
      if (sumW == 0.0) continue;
      double y       = yGraph / sumW;
      double eyLo    = sigmaStat * sqrt(err2lo) / sumW;
      double eyHi    = sigmaStat * sqrt(err2hi) / sumW;
      double eyOutLo = sqrt(sigmaStat*sigmaStat*err2lo + pow(eySysLo,2.)) / sumW;
      double eyOutHi = sqrt(sigmaStat*sigmaStat*err2hi + pow(eySysHi,2.)) / sumW;
      if (doWeightedX) xMean /= yGraph;
      // Check for first non-zero graph point
      if (iPointGraph == firstPointGraph && y == 0.0) {
        ++firstPointGraph;
        continue;
      }
      else if (y == 0. && eyLo == 0. && eyHi == 0.) continue;
      // Count up graph point number from first non-overflow bin
      // and only add points inside range
      else if (xGraph < xHisEnd) { 
        ++nPoints;
        // Apply scatter in x (so as not to plot in exactly same point)
	double xGraphNow = xGraph;
	if (doWeightedX) {
	  xGraphNow = xMean;
	} else {
	  if (scatterX > 0.) xGraphNow += scatterX * exGraphHi;
	  else xGraphNow -= scatterX * exGraphLo;
	}
	double exGraphLoNow = xGraphNow-(xGraph-exGraphLo);
        double exGraphHiNow = (xGraph+exGraphHi) - xGraphNow;
        SetPoint(nPoints,xGraphNow,y);
        SetPointError(nPoints,exGraphLoNow,exGraphHiNow,eyLo,eyHi);
        outer.SetPoint(nPoints,xGraphNow,y);          
        outer.SetPointError(nPoints,exGraphLoNow,exGraphHiNow,
                            eyOutLo,eyOutHi);          
      }
    }
    
    // Check if normalizing to something
    if (normalizeIn && sumWeights != 0.0 && normFacIn != 0.0) {
      for (int iPoint = 0; iPoint <= nPoints; ++iPoint) {
        double yRaw, eyRawHi, eyRawLo, x;
        GetPoint(iPoint,x,yRaw);
        eyRawHi = GetErrorYhigh(iPoint);
        eyRawLo = GetErrorYlow(iPoint);
        SetPoint(iPoint,x,normFacIn*yRaw/sumWeights);
        SetPointEYhigh(iPoint,normFacIn*eyRawHi/sumWeights);
        SetPointEYlow(iPoint,normFacIn*eyRawLo/sumWeights);
        outer.GetPoint(iPoint,x,yRaw);
        eyRawHi = outer.GetErrorYhigh(iPoint);
        eyRawLo = outer.GetErrorYlow(iPoint);
        outer.SetPoint(iPoint,x,normFacIn*yRaw/sumWeights);
        outer.SetPointEYhigh(iPoint,normFacIn*eyRawHi/sumWeights);
        outer.SetPointEYlow(iPoint,normFacIn*eyRawLo/sumWeights);
      }	
    }
    
    return true;
  }      
  
  // Copy color scheme from another VRgraph
  void VRgraph::setColorScheme(VRgraph* graph) {
    SetMarkerSize(graph->GetMarkerSize());
    SetMarkerStyle(graph->GetMarkerStyle());
    SetMarkerColor(graph->GetMarkerColor());
    SetLineStyle(graph->GetLineStyle());
    SetLineColor(graph->GetLineColor());
    SetLineWidth(graph->GetLineWidth());
    SetFillStyle(graph->GetFillStyle());
    SetFillColor(graph->GetFillColor());
    outer.SetMarkerSize(graph->getOuter()->GetMarkerSize());
    outer.SetMarkerStyle(graph->getOuter()->GetMarkerStyle());
    outer.SetMarkerColor(graph->getOuter()->GetMarkerColor());
    outer.SetLineStyle(graph->getOuter()->GetLineStyle());
    outer.SetLineColorAlpha(graph->getOuter()->GetLineColor(),0.88);
    outer.SetLineWidth(graph->getOuter()->GetLineWidth());
    outer.SetFillStyle(graph->getOuter()->GetFillStyle());    
    outer.SetFillColor(graph->getOuter()->GetFillColor());
  }

  //-----------------------------------------------------------------
  // setColorScheme():
  // Set colors according to predefined color schemes, defined here

  bool VRgraph::setColorScheme(int iScheme) {
    // Define the VinciaRoot colour schemes
    gROOT->GetColor(21)->SetRGB(0., 0., 1.);
    gROOT->GetColor(22)->SetRGB(0., 0., 1.);
    gROOT->GetColor(22)->SetAlpha(0.5);
    gROOT->GetColor(23)->SetRGB(0., 0., 1.);
    gROOT->GetColor(23)->SetAlpha(0.33);
    // 2
    gROOT->GetColor(31)->SetRGB(1., 0., 0.);
    gROOT->GetColor(32)->SetRGB(1., 0., 0.);
    gROOT->GetColor(32)->SetAlpha(0.5);
    gROOT->GetColor(33)->SetRGB(1., 0., 0.1);
    gROOT->GetColor(33)->SetAlpha(0.33);
    // 3
    gROOT->GetColor(41)->SetRGB(0.16, 0.2, 0.16);
    gROOT->GetColor(42)->SetRGB(0.16, 0.2, 0.16);
    gROOT->GetColor(42)->SetAlpha(0.5);
    gROOT->GetColor(43)->SetRGB(0.16, 0.2, 0.16);
    gROOT->GetColor(43)->SetAlpha(0.33);
    // 4
    gROOT->GetColor(51)->SetRGB(0.9, 0.45, 0.);
    gROOT->GetColor(52)->SetRGB(0.9, 0.45, 0.);
    gROOT->GetColor(52)->SetAlpha(0.5);
    gROOT->GetColor(53)->SetRGB(0.9, 0.45, 0.);
    gROOT->GetColor(53)->SetAlpha(0.33);
    // 5
    gROOT->GetColor(61)->SetRGB(0.7, 0., 0.7);
    gROOT->GetColor(62)->SetRGB(0.7, 0., 0.7);
    gROOT->GetColor(62)->SetAlpha(0.5);
    gROOT->GetColor(63)->SetRGB(0.7, 0., 0.7);
    gROOT->GetColor(63)->SetAlpha(0.33);
    // 6
    gROOT->GetColor(71)->SetRGB(0., 0.36, 0.8);
    gROOT->GetColor(72)->SetRGB(0., 0.36, 0.8);
    gROOT->GetColor(72)->SetAlpha(0.5);
    gROOT->GetColor(73)->SetRGB(0., 0.36, 0.8);
    gROOT->GetColor(73)->SetAlpha(0.33);
    // 7
    gROOT->GetColor(81)->SetRGB(1.0, 0.35, 0.5);
    gROOT->GetColor(82)->SetRGB(1.0, 0.35, 0.5);
    gROOT->GetColor(82)->SetAlpha(0.5);
    gROOT->GetColor(83)->SetRGB(1.0, 0.35, 0.5);
    gROOT->GetColor(83)->SetAlpha(0.33);
    // 8
    gROOT->GetColor(91)->SetRGB(0., 0.4, 0.4);
    gROOT->GetColor(92)->SetRGB(0., 0.4, 0.4);
    gROOT->GetColor(92)->SetAlpha(0.5);
    gROOT->GetColor(93)->SetRGB(0., 0.4, 0.4);
    gROOT->GetColor(93)->SetAlpha(0.33);
    
    Int_t lineCol, statCol, totCol;
    Int_t lineStyle, markCol, markStyle;
    Width_t lineWidth(1);
    double markSize(0.4);
    bool schemeExists = true;
    // Data : one-sigma black-and-green scheme
    if (iScheme == 0) {
      markCol   =  1;
      markStyle = 21;
      lineStyle =  1;
      lineCol   =  1;
      statCol   = kGreen - 9;
      totCol    = kGreen - 4;
      markSize  = 0.55;
      outer.SetLineColor(lineCol);
      outer.SetLineWidth(lineWidth);
      outer.SetLineStyle(1);
    } 
    // Data : two-sigma black-and-yellow scheme
    else if (iScheme == -1) {
      markCol   =  1;
      markStyle = 21;
      lineStyle =  1;
      lineCol   =  1;
      statCol   = 90;
      totCol    = 91;
      markSize  = 0.55;
      outer.SetLineColor(lineCol);
      outer.SetLineWidth(lineWidth);
      outer.SetLineStyle(1);
    } 
    // Data : two-sigma grey-and-yellow scheme
    else if (iScheme == -2) {
      markCol   = 16;
      markStyle = 21;
      lineStyle =  0;
      lineCol   = 16;
      // No distinction between stat and sys cols for 2-sigma band
      // (Earlier used kYellow for and kYellow-9 for the total)
      statCol   = kYellow-7;
      totCol    = kYellow-7;
      markSize  = 0.55;
      outer.SetLineColor(lineCol);
      outer.SetLineWidth(lineWidth);
      outer.SetLineStyle(0);
    } 
    // 1
    else if (iScheme == 1) {
      lineStyle =  1; 
      lineCol   = 21;
      statCol   = 22;
      totCol    = 23;
      markStyle = 20;  
      markSize  = 0.35; 
      markCol   = lineCol;
      lineWidth = 1;
    } 
    // 2
    else if (iScheme == 2) {
      lineStyle =  2; 
      lineCol   = 31;
      statCol   = 32;
      totCol    = 33;
      markCol   = lineCol;
      markStyle = 33;
      markSize  = 0.6;
      lineWidth = 1;
    } 
    // 3
    else if (iScheme == 3) {
      lineStyle =  4;
      lineCol   = 41;
      statCol   = 42;
      totCol    = 43;
      markCol   = lineCol;
      markStyle =  4;
      lineWidth = 1;
      markSize  = 0.52;
    }
    // 4 
    else if (iScheme == 4) {
      lineStyle =  3;
      lineCol   = 51;
      statCol   = 52;
      totCol    = 53;
      markCol   = lineCol;
      markStyle = 25;
      lineWidth = 1;
      markSize  = 0.65;
    }
    // 5
    else if (iScheme == 5) {
      lineStyle =  1;
      lineCol   =  61;
      statCol   =  62;
      totCol    =  63;
      markCol   = lineCol;
      markStyle = 28;
      lineWidth = 1;
      markSize  = 0.85;
    }
    // 6
    else if (iScheme == 6) {
      lineStyle =  5;
      lineCol   =  71;
      statCol   =  72;
      totCol    =  73;
      markCol   = lineCol;
      markStyle =  5;
      lineWidth = 1;
      markSize  = 0.82;
    }
    // 7
    else if (iScheme == 7) {
      lineStyle =  4; 
      lineCol   = 81;
      statCol   = 82;
      totCol    = 83;
      markCol   = lineCol;
      markStyle = 26;
      lineWidth = 1;
      markSize  = 0.59;
    } 
    // Fallback: grey
    else {
      lineStyle =  3;
      lineCol   =  15;
      statCol   =  16;
      totCol    =  17;
      markCol   = lineCol;
      markStyle = 34;
      lineWidth = 1;
      markSize  = 0.64;
      schemeExists = false;
    }    
    if (iScheme >= 1) {
      SetLineColorAlpha(lineCol,0.88);
    }
    else {
      SetLineColor(lineCol);
    }
    SetMarkerSize(markSize);
    SetMarkerStyle(markStyle);
    SetMarkerColor(markCol);
    SetLineStyle(lineStyle);
    SetLineWidth(lineWidth);
    SetFillColor(statCol);
    outer.SetFillColor(totCol);
    return schemeExists;
  }

  //=====================================================================
  // VRgraphData Member function definitions 

  //-----------------------------------------------------------------
  // update():
  // Update graph
  // Normalize to normFacIn
  // Multiply stat *and* sys errors by sigmaTot (eg for 2-sigma bands)

  bool VRgraphData::update(bool normalizeIn, double normFacIn, 
                           double sigmaTot, double) {
    if (dataFilePtr == 0) {
      cout<<"Error: cannot update dataGraph without dataFilePtr"<<endl;
      return false;
    }
    sumWeights   = dataFilePtr->getSumWeights();
    if (sumWeights == 0.0) sumWeights = 1.0;
    int nBin    = dataFilePtr->size();    
    for (int iBin = 0; iBin < nBin; ++iBin) {
      double x    = dataFilePtr->getXC(iBin);
      double exLo = x - dataFilePtr->getXL(iBin);
      double exHi = dataFilePtr->getXR(iBin) - x;
      double y    = dataFilePtr->getY(iBin);
      double eyTotHi = dataFilePtr->getUtotHi(iBin);
      double eyTotLo = dataFilePtr->getUtotLo(iBin);
      double eyStatHi = dataFilePtr->getUstatHi(iBin);
      double eyStatLo = dataFilePtr->getUstatLo(iBin);        
      double eySysHi  = sqrt(max(0.0,pow(eyTotHi,2)-pow(eyStatHi,2)));
      double eySysLo  = sqrt(max(0.0,pow(eyTotLo,2)-pow(eyStatLo,2)));
      // Scale errors by sigmaTot (to show nSigma contours)      
      eyStatHi *= sigmaTot;
      eyStatLo *= sigmaTot;
      eySysHi  *= sigmaTot;
      eySysLo  *= sigmaTot;
      eyTotHi = sqrt(pow(eyStatHi,2) + pow(eySysHi,2));
      eyTotLo = sqrt(pow(eyStatLo,2) + pow(eySysLo,2));
      SetPoint(iBin,x,y);
      SetPointError(iBin,exLo,exHi,eyStatLo,eyStatHi);
      outer.SetPoint(iBin,x,y);          
      outer.SetPointError(iBin,exLo,exHi,eyTotLo,eyTotHi);          
      // For files containing data from multiple experiments, allow to use
      // different symbols for each experiment. 
      if (dataFilePtr->hasPointNames()) {
        // Check if this label already exists
        string thisLab = dataFilePtr->getPointName(iBin);
        if (auxGraphs.find(thisLab) == auxGraphs.end()) {
          auxGraphs[thisLab] = VRgraph(0);
          auxGraphs[thisLab].setColorScheme(0);
          int iMarker[14] =
            {21, 20, 29, 22, 23, 25, 24, 26, 27, 28, 30, 32, 31, 21};
          int markerStyle = iMarker[min((int)auxGraphs.size()-1,13)];
          auxGraphs[thisLab].SetMarkerStyle(markerStyle);
          auxGraphs[thisLab].SetMarkerSize(0.95);
        }        
      }
    } 
    
    // Check if normalizing to something
    if (normalizeIn && sumWeights != 0.0 && normFacIn != 0.0) {
      for (int iBin = 0; iBin < nBin; ++iBin) {
        double yRaw, eyRawHi, eyRawLo, x;
        GetPoint(iBin,x,yRaw);
        eyRawHi = GetErrorYhigh(iBin);
        eyRawLo = GetErrorYlow(iBin);
        SetPoint(iBin,x,normFacIn*yRaw/sumWeights);
        SetPointEYhigh(iBin,normFacIn*eyRawHi/sumWeights);
        SetPointEYlow(iBin,normFacIn*eyRawLo/sumWeights);
        eyRawHi = outer.GetErrorYhigh(iBin);
        eyRawLo = outer.GetErrorYlow(iBin);
        outer.SetPoint(iBin,x,normFacIn*yRaw/sumWeights);
        outer.SetPointEYhigh(iBin,normFacIn*eyRawHi/sumWeights);
        outer.SetPointEYlow(iBin,normFacIn*eyRawLo/sumWeights);
      }
    }

    // For files containing data from multiple experiments, fill one
    // auxiliary graph for each set of distinct points.    
    // Only need to do this at first update
    if (dataFilePtr->hasPointNames() && firstUpdate) {
      for (int iBin = 0; iBin < nBin; ++iBin) {
        double x, y;
        GetPoint(iBin,x,y);
        string thisLab = dataFilePtr->getPointName(iBin);
        int iPoint = auxGraphs[thisLab].GetN();
        auxGraphs[thisLab].SetPoint(iPoint,x,y);
        auxGraphs[thisLab].SetPointEYhigh(iPoint,0.);
        auxGraphs[thisLab].SetPointEYlow(iPoint,0.);
      }      
    }

    firstUpdate = false;
    return true;
  }
  
  //=====================================================================
  // VRgraphHis Member function definitions 

  //-----------------------------------------------------------------
  // update():
  // Update graph

  bool VRgraphHis::update(bool normalizeIn, double normFacIn, 
                           double sigmaStat, double) {
    if (hisFilePtr == 0) {
      cout<<"Error: cannot update dataGraph without hisFilePtr"<<endl;
      return false;
    }

    // Initialize weight sum
    sumWeights   = hisFilePtr->getSumWeights();
    if (sumWeights == 0.0) sumWeights = 1.0;
    
    // End of histogram
    int nBinHis    = hisFilePtr->size();
    double xHisEnd = hisFilePtr->getXR(nBinHis-1);

    // Allow optional rebinning according to a "rebinAxis"
    int nPointGraph     = nBinHis;
    int firstPointGraph = 0;
    if (rebin) {
      if (rebinAxisPtr != 0) {
        nPointGraph     = rebinAxisPtr->GetN();
        firstPointGraph = 0;
      } else {
        rebin = false;
      }
    }

    int nPoints  =  -1;
    double integral  = 0.;
    // Do inner and outer loop; 
    // outer: loop over graph bins (according to rebinner or hist, as desired)
    // inner: loop over histogram bins (starting where previous left off)
    // Note: check for half-bins or smaller. Not reliable. Probably only write 
    // that once (initialize to false, then write true, or use Pythia's INFO)
    int iBinFirst    = min(0,nBinHis-1);
    int iBinLast     = iBinFirst;
    for (int iPointGraph = 0; iPointGraph <= nPointGraph; ++iPointGraph) {
      // Start out with assumption of no rebinning: his points = graph points
      double xGraph    = hisFilePtr->getXC(min(iPointGraph,nBinHis-1));
      double exGraphLo = xGraph - hisFilePtr->getXL(min(iPointGraph,nBinHis-1));
      double exGraphHi = hisFilePtr->getXR(min(iPointGraph,nBinHis-1)) - xGraph;
      bool doWeightedX = false;
      if (rebin) {
        double yDum(0.);
        int iGood = rebinAxisPtr->GetPoint(iPointGraph,xGraph,yDum);
        if (iGood == -1) continue;
        exGraphHi = abs(rebinAxisPtr->GetErrorXhigh(iPointGraph));
        exGraphLo = abs(rebinAxisPtr->GetErrorXlow(iPointGraph));        
        iBinFirst = max(0,iBinFirst-1);
        iBinLast  = nBinHis-1;
	// Check if we should calculate the weighted x mean
	if (abs(exGraphHi - exGraphLo) > 1e-3*abs(exGraphHi + exGraphLo))
	  doWeightedX = true;
      } else {
        iBinFirst = iPointGraph;
        iBinLast  = iPointGraph;
        // For histogram files, first bin is ok, no need to skip for underflow
      }
      double yGraph  = 0.0;
      double xMean   = (doWeightedX) ? 0.0 : xGraph;
      double err2lo  = 0.0;
      double err2hi  = 0.0;
      double eySysHi = 0.0;
      double eySysLo = 0.0;
      double sumW    = 0.0;      
      // Histogram (TH1)
      // Bin # 0         : underflow
      // Bin # [1,nbins] : contens
      // Bin # nbins + 1 : overflow
      for (int iBinHis = iBinFirst; iBinHis <= iBinLast; ++iBinHis) {
        double xHis    = hisFilePtr->getXC(iBinHis);
        double exHisLo = xHis - hisFilePtr->getXL(iBinHis);
        double exHisHi = hisFilePtr->getXR(iBinHis) - xHis;
        if (xHis + exHisHi < xGraph - exGraphLo) {
          iBinFirst = iBinHis;
          continue;
        }
        if (xHis - exHisLo > xGraph + exGraphHi) {
          break;
        }
        double w = 1.0 ;
        if ( (exHisHi+exHisLo) > 0) 
          w = (min(xHis + exHisHi,xGraph + exGraphHi) 
               - max(xHis - exHisLo, xGraph - exGraphLo))
            / (exHisHi + exHisLo);        
        if (w <= 0.0) continue;
        double yHisBin = hisFilePtr->getY(iBinHis);
        if (zeroSkippingSav && yHisBin == 0.0) continue;
        // 
        double eyTotHiBin  = hisFilePtr->getUtotHi(iBinHis);
        double eyTotLoBin  = hisFilePtr->getUtotLo(iBinHis);
        double eyStatLoBin = hisFilePtr->getUstatLo(iBinHis);
        double eyStatHiBin = hisFilePtr->getUstatHi(iBinHis);
        double eySysHiBin  = sqrt(max(0.0,pow(eyTotHiBin,2.)-pow(eyStatHiBin,2.)));
        double eySysLoBin  = sqrt(max(0.0,pow(eyTotLoBin,2.)-pow(eyStatLoBin,2.)));
        // Don't plot points that correspond to unfilled histogram entries
        // (but include them in rebin averages if not doing zeroSkipping)
        if (!rebin && yHisBin == 0 && eyTotHiBin == 0. && eyTotLoBin == 0.) continue;
        sumW += w;
        yGraph += w * yHisBin;
	if (doWeightedX) xMean += xHis * w * yHisBin;
        err2lo += pow(w * eyStatLoBin,2.);
        err2hi += pow(w * eyStatHiBin,2.); 
        // Got uncertainty variations? Add linearly when combining bins
	// (e.g., to prevent rebinning from artificially improving)
        if (hasOuter()) {
          eySysHi += w * eySysHiBin;
          eySysLo += w * eySysLoBin;
        }
      }
      // Don't store points that have zero weights.
      if (sumW == 0.0) continue;
      double y       = yGraph / sumW;
      double eyLo    = sigmaStat * sqrt(err2lo) / sumW;
      double eyHi    = sigmaStat * sqrt(err2hi) / sumW;
      double eyOutLo = sqrt(sigmaStat*sigmaStat*err2lo + pow(eySysLo,2.)) / sumW;
      double eyOutHi = sqrt(sigmaStat*sigmaStat*err2hi + pow(eySysHi,2.)) / sumW;
      if (doWeightedX) xMean /= yGraph;
      // Check for first non-zero graph point
      if (iPointGraph == firstPointGraph && y == 0.0) {
        ++firstPointGraph;
        continue;
      }
      else if (y == 0. && eyLo == 0. && eyHi == 0.) continue;
      // Count up graph point number from first non-overflow bin
      // and only add points inside range
      else if (xGraph < xHisEnd) { 
        ++nPoints;
        SetPoint(nPoints,xMean,y);
	double exMeanLo = xMean - (xGraph-exGraphLo);
	double exMeanHi = (xGraph + exGraphHi) - xMean;
        SetPointError(nPoints,exMeanLo,exMeanHi,eyLo,eyHi);
        outer.SetPoint(nPoints,xMean,y);          
        outer.SetPointError(nPoints,exMeanLo,exMeanHi,eyOutLo,eyOutHi);
        integral += y * (abs(exGraphLo) + abs(exGraphHi));
      }
    }

    // Update sum of weights to include integral over rebinned region only
    // (optionally include overflow bin, eg for distributions where data
    //  normalisation includes tail but histogram only shows peak)
    if (rebin) {
      bool includeOverflow = true;        
      sumWeights = integral;
      if (includeOverflow) { 
        double xHisOver = hisFilePtr->getXL(nBinHis-1);
        double dXover   = xHisEnd - xHisOver;
        double yOver    = hisFilePtr->getY(nBinHis-1);
        sumWeights += yOver * dXover;
      }
    }
    
    // Check if normalizing to something
    if (normalizeIn && sumWeights != 0.0 && normFacIn != 0.0) {
      for (int iBin = 0; iBin <= nPoints; ++iBin) {
        double yRaw, eyRawHi, eyRawLo, x;
        GetPoint(iBin,x,yRaw);
        eyRawHi = GetErrorYhigh(iBin);
        eyRawLo = GetErrorYlow(iBin);
        SetPoint(iBin,x,normFacIn*yRaw/sumWeights);
        SetPointEYhigh(iBin,normFacIn*eyRawHi/sumWeights);
        SetPointEYlow(iBin,normFacIn*eyRawLo/sumWeights);
        eyRawHi = outer.GetErrorYhigh(iBin);
        eyRawLo = outer.GetErrorYlow(iBin);
        outer.SetPoint(iBin,x,normFacIn*yRaw/sumWeights);
        outer.SetPointEYhigh(iBin,normFacIn*eyRawHi/sumWeights);
        outer.SetPointEYlow(iBin,normFacIn*eyRawLo/sumWeights);
      }
    }
    return true;
  }
  
  //=====================================================================
  // VRgraphNorm Member function definitions 

  //-----------------------------------------------------------------
  // update(): 
  // Update graph

  bool VRgraphNorm::update(bool, double sigmaIn, 
                           double ratioMin, double ratioMax) {
    
    // Sanityauxg checks
    if (!hasNumerator || !hasDenominator) return false;
    if (numeratorPtr->GetN() <= 0) return false; 
    if (denominatorPtr->GetN() <= 0) return false; 

    // 
    hasOuterSav = numeratorPtr->hasOuter();
    
    // Zero sums for chi2 calculations
    double sumChi2th5       = 0.0;
    double sumSigma2chi2th5 = 0.0;
    int    nPointsChi2      = 0;

    // Compute normalized histogram (using binning from denominator)
    int nBin    = denominatorPtr->GetN();
    int nPoints = 0;
    int iFirst  = 0;
    int iLast   = nBin -1;
    // Count number of denominator bins that are actually processed
    // (e.g., data file can have more bins, but maybe only some are plotted)
    int nDenom  = 0;
    // Start numerator loop off at zero, then update for each iteration
    // (to avoid starting over from beginning for each numerator loop)
    int hBinStart = 0;
    for (int iBin = iFirst; iBin <= iLast; ++iBin) {
      
      // Extract bin to use for normalization
      double xC, yDenom, exLo, exHi, xL, xR;
      denominatorPtr->GetPoint(iBin,xC,yDenom);        
      // Sanity check
      if (yDenom == 0.0) continue;
      double eyDenomNhi = denominatorPtr->getOuter()->GetErrorYhigh(iBin)/yDenom;
      double eyDenomNlo = denominatorPtr->getOuter()->GetErrorYlow(iBin)/yDenom;
      exLo = denominatorPtr->GetErrorXlow(iBin);
      exHi = denominatorPtr->GetErrorXhigh(iBin);
      xL = xC - exLo;
      xR = xC + exHi;
      // Ratios cannot take weighted xMean into account; redefine to center
      xC = (xL + xR)/2.;
      exLo = xC - xL;
      exHi = xR - xC;
      // Loop over numerator bins and rebin like denominator
      double sumw      = 0.0;
      double sumy      = 0.0;
      // Statistical MC generation errors
      double sumErrHi2 = 0.0;
      double sumErrLo2 = 0.0;
      // Note: up/down variation curves must be added linearly!
      double sumUpper  = 0.0;
      double sumLower  = 0.0;
      VRgraph* h=numeratorPtr;
      TGraphAsymmErrors* hU=numeratorPtr->getOuter();
      int nh = h->GetN();
      // Check plot range
      double xMin, xMax, yDum;
      h->GetPoint(0,xMin,yDum);
      xMin -= h->GetErrorXlow(0);
      h->GetPoint(nh-1,xMax,yDum);
      xMax +=  h->GetErrorXhigh(nh-1);
      // Write upper bound in numberically stable way
      if ( (xL - xMax)/(abs(xL)+abs(xMax)) >= -TINY ) continue;
      if ( xR <= xMin ) continue;
      // Update number of denominator bins that have been processed
      nDenom++;
      // Loop over bins
      double xRprev = xL;
      double bwPrev = 1.;      
      for (int hBin=hBinStart; hBin<nh; ++hBin){
        // Get central values for this numerator bin
        double xNum, yNum, xNumU, yNumU;
        h->GetPoint(hBin,xNum,yNum);
        hU->GetPoint(hBin,xNumU,yNumU);       
        double xLnum = xNum - h->GetErrorXlow(hBin);
        double xRnum = xNum + h->GetErrorXhigh(hBin);
        double bw = xRnum-xLnum;
        // Check if not reached denominator bin left edge
        if (xRnum <= xL) continue;
        // Update starting bin for next numerator loop
        hBinStart = max(0,hBin-1);
        // Check if already passed denominator bin right edge
        if (xLnum >= xR) break;    
        // Sanity check
        if (bw <= 0.0) continue;
        double w = 1.0;
        // Uncertainties
        double uStatHi = h->GetErrorYhigh(hBin);
        double uStatLo = h->GetErrorYlow(hBin);
        double ySysHi  = yNumU + sqrt(max(0.0,pow(hU->GetErrorYhigh(hBin),2) 
                                          - pow(uStatHi,2)));
        double ySysLo  = yNumU - sqrt(max(0.0,pow(hU->GetErrorYlow(hBin),2) 
                                          - pow(uStatLo,2)));
        // Subtract partial bin overlap
        w -= max(0.,xL-xLnum)/bw;
        w -= max(0.,xRnum-xR)/bw;
        w = max(w,0.0);
        sumw += w;
        // Add skipped empty bins (on the left)
        sumw += max(0.0,(xLnum-xRprev))/bw;
        xRprev = xRnum;
        bwPrev = bw;
        // Sum
        sumy      += w*yNum;
        sumErrHi2 += w*pow(uStatHi,2);
        sumErrLo2 += w*pow(uStatLo,2);
        sumUpper  += w*ySysHi;
        sumLower  += w*ySysLo;
      }  	  
      // Sanity check
      if (sumw <= 0.) continue;
      // Add skipped empty bins between last and right edge
      sumw += max(0.0,xR-xRprev)/bwPrev;
      
      // Compute average values and total stat error
      double y           = sumy/sumw;          
      double eyStatHi    = sigmaIn * sqrt(sumErrHi2)/sumw;
      double eyStatLo    = sigmaIn * sqrt(sumErrLo2)/sumw;
      double yVarUpper   = sumUpper/sumw;
      double yVarLower   = sumLower/sumw;
      double eyVarHi     = max(0.0,yVarUpper-y);
      double eyVarLo     = max(0.0,y-yVarLower);
      
      // Normalize by data
      Double_t yN        = y/yDenom;
      Double_t eyStatHiN = eyStatHi/yDenom;          
      Double_t eyStatLoN = eyStatLo/yDenom;          
      // Add stat error in quadrature for normalized total uncertainty
      Double_t eyTotHiN  = sqrt(pow(eyVarHi,2.) + pow(eyStatHi,2.))/yDenom;
      Double_t eyTotLoN  = sqrt(pow(eyVarLo,2.) + pow(eyStatLo,2.))/yDenom;   
      
      // Compute chi2 with 5% blanket uncertainty
      // Chi2 is for central MC prediction (with current statistics)
      // Uncertainty on Chi2 is calculated from current MC stat uncertainty.
      // The +/- value used to compute the uncertainty on the chi2 value
      // is multiplied by sigChi2. A conservative default could be 1.645 sigma 
      // (corresponding to 90% confidence). For the time being, 1 sigma is used,
      // since that is how people usually interpret a +/- uncertainty. 
      // This number could in principle be made user-modifiable.
      double sigChi2 = 1.0;
      if (yN != 0.0 && max(eyTotHiN,eyTotLoN) > 0.) {
        ++nPointsChi2;
        double chi2th5 = ( yN > 1.0 ) 
          ? pow(yN - 1.0,2.)/(pow(eyDenomNhi,2.)+pow(0.05*yN,2.))
          : pow(yN - 1.0,2.)/(pow(eyDenomNlo,2.)+pow(0.05*yN,2.));
        double sigma2chi2th5 = ( yN > 1.0 ) 
          ? chi2th5 * 4.*pow(sigChi2*eyStatLoN,2.) 
          / (pow(eyDenomNhi,2.)+pow(0.05,2.))
          : chi2th5 * 4.*pow(sigChi2*eyStatHiN,2.) 
          / (pow(eyDenomNlo,2.)+pow(0.05,2.));
        sumChi2th5       += chi2th5;
        sumSigma2chi2th5 += sigma2chi2th5;
      }

      // Limit ratios to max ratioMin/ratioMax (to clip plot)
      eyStatHiN = (yN + eyStatHiN > ratioMax)? max(0.0,ratioMax-yN) : eyStatHiN;
      eyStatLoN = (yN - eyStatLoN < ratioMin)? max(0.0,yN-ratioMin) : eyStatLoN;
      eyTotHiN  = (yN + eyTotHiN > ratioMax) ? max(0.0,ratioMax-yN) : eyTotHiN;
      eyTotLoN  = (yN - eyTotLoN < ratioMin) ? max(0.0,yN-ratioMin) : eyTotLoN;
      
      // Do not plot x errors on ratio plots (except for data)
      double exOuterLo = exLo;
      double exOuterHi = exHi;
      if (numeratorPtr != denominatorPtr) {
	exLo = 0.0;
	exHi = 0.0;
      }

      // Set point value and errors;
      ++nPoints;
      SetPoint(nPoints-1,xC,yN);
      SetPointError(nPoints-1,exLo,exHi,eyStatLoN,eyStatHiN);
      outer.SetPoint(nPoints-1,xC,yN);          
      outer.SetPointError(nPoints-1,exOuterLo,exOuterHi,eyTotLoN,eyTotHiN);          

    }

    // Save reduced chi2 value and relative uncertainy from MC stats
    // (calculation of chi2 only meaningful if all bins are populated)
    if (nPoints >= 1 && nPointsChi2 == nDenom) {
      chi2sav       = sumChi2th5/nPointsChi2;
      chi2errSav    = sqrt(sumSigma2chi2th5)/nPointsChi2; 
    } else {
      chi2sav       = -1.0;
      chi2errSav    =  1.0;
    }

    // Return
    return true;
  }
    
  //=====================================================================
  // VRgraphVar Member function definitions 

  //-----------------------------------------------------------------
  // update():
  // Update graph

  bool VRgraphVar::update(bool, double, double, double) {
    
    // Sanity checks
    if (referencePtr->GetN() <= 0) return false; 
    
    // Compute difference histogram (using binning from reference)
    // and normalize by reference
    TH1D* hRef  = referencePtr->getTH1DPtr(0);
    TH1D* hVar  = referencePtr->getTH1DPtr(iVarSav);
    int nBin    = hRef->GetNbinsX();
    int nPoints = -1;
    int iFirst  = 1;
    int iLast   = nBin;
    // Loop over histogram bins
    for (int iBin = iFirst; iBin <= iLast; ++iBin) {
      double xRef = hRef->GetBinCenter(iBin);
      double yRef = hRef->GetBinContent(iBin);
      //double eRef = hRef->GetBinError(iBin);
      double yVar = hVar->GetBinContent(iBin);
      //double eVar = hVar->GetBinError(iBin);
      double yMin  = min(yRef,yVar);
      double yMax  = max(yRef,yVar);
      double eMin = 0.0;
      double eMax = 0.0;
      for (int iVar=1; iVar<=int(referencePtr->nOuter()); ++iVar) {
        double yTmp = referencePtr->getTH1DPtr(iVar)->GetBinContent(iBin);
        if (yTmp >= yMax) {
	  yMax = yTmp;
	  eMax = referencePtr->getTH1DPtr(iVar)->GetBinError(iBin);
	}
        else if (yTmp <= yMin) {
	  yMin = yTmp;
	  eMin = referencePtr->getTH1DPtr(iVar)->GetBinError(iBin);
	}
      }
      // Sanity check (require non-zero total uncertainty)
      if (yMax - yMin == 0.0) continue;
      
      // Is this an up or down variation?
      double yFrac = 0.;
      double eyFracHi = 0.;
      double eyFracLo = 0.;
      if (yVar >= yRef) {
        // Sanity check (require non-zero total up variation)
        if (yMax - yRef <= 0.0) continue;
        yFrac = (yVar - yRef)/(yMax - yRef);          
	// Optionally assign errors to the errors
        //eyFracHi = sqrt(pow(eVar,2)+pow(eRef,2))/(yMax - yRef);
        //if (yFrac + eyFracHi > 1.0) eyFracHi = 1.0 - yFrac;
	// Optionally scale result by statistical precision
	//yFrac *= eRef/eMax ;
	// Make the band filled, from 0 to yFrac.
        eyFracLo = abs(yFrac);
      } else {
        if (yRef - yMin <= 0.0) continue;
        yFrac = (yVar - yRef)/(yRef - yMin);
        //eyFracLo = sqrt(pow(eVar,2)+pow(eRef,2))/(yRef - yMin);
        //if (yFrac - eyFracLo < -1.0) eyFracLo = -1.0 - yFrac;
	// Optionally scale result by statistical precision
	// yFrac /= eRef/eMin ;
	// Make the band filled, from yFrac to 0.
        eyFracHi = abs(yFrac);
      }
      
      // Set point value and errors;
      ++nPoints;
      SetPoint(nPoints,xRef,yFrac);
      SetPointError(nPoints,0.,0.,eyFracLo,eyFracHi);
    }
    return true;
  }
  
  //-----------------------------------------------------------------
  // setStyle(): 
  // Set line style for each variation, according to predefined schemes
  // defined here

  void VRgraphVar::setStyle(int iVar) {
    if (iVar == 0) {
      SetLineStyle(1);
      SetLineWidth(1.0);        
      SetFillColor(kBlack);
    }
    else if (iVar == 1) {
      SetLineStyle(1);
      SetLineWidth(1.0);        
      SetLineColor(kBlue-10);
      SetFillColor(kBlue-10);
    }
    else if (iVar == 2) {
      SetLineStyle(1);
      SetLineWidth(1.0);        
      SetLineColor(kRed-10);
      SetFillColor(kRed-10);
    }
    else if (iVar == 3) {
      SetLineStyle(1);
      SetLineWidth(2.0);        
      SetLineColor(kBlue-9);
      SetFillColor(kBlue-9);
      SetFillStyle(3305);
    }
    else if (iVar == 4) {
      SetLineStyle(1);
      SetLineWidth(2.0);        
      SetLineColor(kRed-8);
      SetFillColor(kRed-8);
      SetFillStyle(3305);
    }
    else if (iVar == 5) {
      SetLineStyle(1);
      SetLineWidth(2.0);        
      SetLineColor(kBlue-7);
      SetFillColor(kBlue-7);
      SetFillStyle(0);
    }
    else if (iVar == 6) {
      SetLineStyle(1);
      SetLineWidth(2.0);        
      SetLineColor(kRed-6);
      SetFillColor(kRed-6);
      SetFillStyle(0);
    }
    else if (iVar == 7) {
      SetLineStyle(2);
      SetLineWidth(1);        
      SetLineColor(kGray+2);
      SetFillStyle(0);
    }
    else if (iVar == 8) {
      SetLineStyle(1);
      SetLineWidth(1);        
      SetLineColor(kGray+2);
      SetFillStyle(0);
    }
    else if (iVar == 9) {
      SetLineStyle(3);
      SetLineWidth(1.0);        
      SetLineColor(kBlue+2);
      SetFillColor(kBlue+2);
      SetFillStyle(0);
    }
    else if (iVar == 10) {
      SetLineStyle(3);
      SetLineWidth(2.0);        
      SetLineColor(kRed+2);
      SetFillColor(kRed+2);
      SetFillStyle(0);
    }
  }

} // End namespace Vincia 

