// TimeAntShower.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2017 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the TimeAntShower class, 
// as follows:
// 1) inherited member functions from the Pythia 8 TimeShower classs
// 2) non-inherited member functions for the TimeAntShower class

// Include the Vincia headers 
#include "Vincia/TimeAntShower.h"
#include "Vincia/SpaceAntShower.h"

namespace Vincia {

//**********************************************************************

// The TimeAntShower class and its members

//**********************************************************************

// BEGIN INHERITED FUNCTIONS FOR INTERFACE WITH PYTHIA8
//  : init     : global shower initialization function
//  : shower   : perform shower off partonic system
//  : prepare  : prepare partonic system for showering
//  : update   : update the partonic system
//  : rescatterUpdate : purely dummy
//  : rescatterPropagateRecoil : purely dummy
//  : pTnext   : generate next scale in downwards evolution
//  : branch   : perform a branching, store results in event record

//*********

// Initialize Shower

void TimeAntShower::init( BeamParticle* beamAPtrIn, 
  BeamParticle* beamBPtrIn) {

  // Check if already initialized 
  if (isInit)
    if (settingsPtr->word("Merging:Process").compare("void") == 0)
      return;

  // Event counters and debugging
  nAccepted       = -1; // Set to -1, so prepare works properly
  nSelected       = -1;
  nVetoUserHooks  = 0;
  nFailHadLevel   = 0;
  nCallPythiaNext = 0;

  // Verbose level and version number
  verbose         = settingsPtr->mode("Vincia:verbose");
  version         = settingsPtr->parm("Vincia:versionNumber");

  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::init","begin --------------");

  // Store input pointers for future use. 
  beamAPtr     = beamAPtrIn;
  beamBPtr     = beamBPtrIn;
  sbbSav       = m2(beamAPtr->p(), beamBPtr->p());

  // Possibility to allow user veto of emission step.
  hasUserHooks      = (userHooksPtr != 0);
  canVetoEmission   = (hasUserHooks && userHooksPtr->canVetoFSREmission());

  // Shower parameters
  evolutionType     = settingsPtr->mode("Vincia:evolutionType");
  // Number of active quark flavours
  nGluonToQuarkF    = settingsPtr->mode("Vincia:nGluonToQuarkF");

  // Perturbative cutoff
  cutoffType         = settingsPtr->mode("Vincia:cutoffTypeF");
  if (cutoffType == 0) cutoffType = evolutionType;
  cutoffScale        = settingsPtr->parm("Vincia:cutoffScaleFF");

  // Normalization factors for types 1 and 2 evolution variables
  evNormFac = 1.0;
  if ( evolutionType == 1 )
    evNormFac =  settingsPtr->parm("Vincia:pTnormalization");
  else if ( evolutionType == 2 )
    evNormFac = settingsPtr->parm("Vincia:mDnormalization");
  normFacQ2max = (evolutionType == 1 ? evNormFac/4.0 : evNormFac/2.0);

  // Number of flavours to be treated as massless
  // (Can be made user-specifiable in future if desired)
  nFlavZeroMassKin   = 5;
  nFlavZeroMassDyn   = settingsPtr->mode("Vincia:nFlavZeroMassF");
  //massCorrectionEmit = settingsPtr->mode("Vincia:massCorrectionEmitF");
  //massCorrectionSplit= settingsPtr->mode("Vincia:massCorrectionSplitF");

  // Mass windows : specifies thresholds for each flavour.
  mt          = vinComPtr->mt;
  mb          = vinComPtr->mb;
  mc          = vinComPtr->mc;
  ms          = vinComPtr->ms;
  // Derived scales
  mtb         = sqrt(mt*mb);

  // Set shower alphaS pointer
  useCMW                = settingsPtr->flag("Vincia:useCMW");
  alphaSptr             = &vinComPtr->alphaStrong;
  if (useCMW) alphaSptr = &vinComPtr->alphaStrongCMW;

  // AlphaS parameters
  alphaSvalue        = settingsPtr->parm("Vincia:alphaSvalue");
  alphaSorder        = settingsPtr->mode("Vincia:alphaSorder");
  aSkMu2             = pow2(settingsPtr->parm("Vincia:alphaSkMuF"));
  aSkMu2Split        = pow2(settingsPtr->parm("Vincia:alphaSkMuSplitF"));
  alphaSmax          = settingsPtr->parm("Vincia:alphaSmax"); 
  alphaSmuFreeze     = settingsPtr->parm("Vincia:alphaSmuFreeze");
  mu2freeze          = pow2(alphaSmuFreeze);

  // Smallest allowed scale for running alphaS
  alphaSmuMin        = max( alphaSmuFreeze, 1.05 * alphaSptr->Lambda3() );
  mu2min             = pow2(alphaSmuMin);

  // Check largest numerical value of alphaS we can have
  if (alphaSorder >= 1) alphaSmax = min(alphaSmax, alphaSptr->alphaS(mu2min));

  // Initialize Trial Generators
  // Note: additional trial generators can be added and initialized by 
  // adding them in the header and adding an additional push_back here.
  trialGenPtrs[trialSoft.name()] = &trialSoft;
  trialGenPtrs[trialCollI.name()] = &trialCollI;
  trialGenPtrs[trialCollK.name()] = &trialCollK;
  trialGenPtrs[trialSplitI.name()] = &trialSplitI;
  trialGenPtrs[trialSplitK.name()] = &trialSplitK;
  map<string,TrialGeneratorFSR*>::iterator it;
  for (it=trialGenPtrs.begin(); it!=trialGenPtrs.end(); ++it) {
    it->second->initPtr(rndmPtr,settingsPtr);      
    it->second->init(evolutionType,evNormFac);
  }

  // Uncertainty Bands and Weights
  uncertaintyBands   = settingsPtr->flag("Vincia:uncertaintyBands");
  allowReweighting   = settingsPtr->flag("Vincia:allowReweighting");

  // Global flag for helicity dependence
  helicityShower     = settingsPtr->flag("Vincia:helicityShower");

  // local flag for polarization and mass. Default values
  isMassless = true;    
  usePolarizationNow = helicityShower;

  // Bias settings
  biasInHard         = settingsPtr->flag("Vincia:biasInHardProcess");
  biasInResDec       = settingsPtr->flag("Vincia:biasInResonanceDecays");
  biasInMPI          = settingsPtr->flag("Vincia:biasInMPIshowers");
  biasAll            = settingsPtr->parm("Vincia:biasAll");
  biasBottom         = settingsPtr->parm("Vincia:biasBottom");
  biasCharm          = settingsPtr->parm("Vincia:biasCharm");
  biasCutoff         = settingsPtr->parm("Vincia:biasCutoff");
  
  // Hyperjet compatibility mode on/off
  hyperjet           = settingsPtr->flag("Vincia:hyperjet");

  // Colour reconnections
  CRmode        = settingsPtr->mode("Vincia:CRmode");
  if (CRmode > 0) { // TODO no CR at the moment
    printOut("TimeAntShower::init","warning; forcing CRmode == 0 "
      "since disabled at the moment.");
    settingsPtr->mode("Vincia:CRmode",0);
    CRmode = 0;
  }
  CRjunctions   = settingsPtr->flag("Vincia:CRjunctions");

  // Matching settings
  matchingLO         = settingsPtr->mode("Vincia:matchingLO");
  meCorrections      = (matchingLO >= 1);
  meCorrectionsNow   = meCorrections;
  matchingFullColor  = settingsPtr->flag("Vincia:matchingFullColor");
  matchingRegOrder   = settingsPtr->mode("Vincia:matchingRegOrder");
  // Matching scale can only be imposed at highest matched multiplicity
  // or not at all. (Matching assumes all previous multiplicities were
  // matched over all of phase space.)
  if (matchingRegOrder <= matchingLO-1) {
    printOut("TimeAntShower::init","warning; forcing matchingRegOrder >= "
      "matchingLO");
    matchingRegOrder = max(matchingLO,matchingRegOrder);
  }
  matchingRegShape   = settingsPtr->mode("Vincia:matchingRegShape");
  matchingScaleIsAbs = settingsPtr->flag("Vincia:matchingRegScaleIsAbsolute");
  matchingScale      = settingsPtr->parm("Vincia:matchingRegScale");
  matchingScaleRatio = settingsPtr->parm("Vincia:matchingRegScaleRatio");
  matchingIRcutoff   = settingsPtr->parm("Vincia:matchingIRcutoff");
  Hinterference      = settingsPtr->flag("Vincia:Hinterference");
  useMG4             = settingsPtr->flag("Vincia:useMG4");
  useMG5             = settingsPtr->flag("Vincia:useMG5");
  useMHV             = settingsPtr->flag("Vincia:useMHVamplitudes");

  // Initialize headroom factors (optimized dynamically later?)
  optimizeHeadroom = settingsPtr->flag("Vincia:MCoptimization");
  iRegDependence   = false;
  if (uncertaintyBands) {
    optimizeHeadroom = false;
    if (helicityShower) initHeadroom(16.0, 16.0, 50.0, 50.0);
    else initHeadroom(8.0, 8.0, 20.0, 20.0);
  } else initHeadroom(1.0, 1.0, 1.0, 1.0);
  pAcceptFuss      = settingsPtr->parm("Vincia:PacceptFuss");
  pAcceptPT0       = settingsPtr->parm("Vincia:PacceptPT0");
  nReportMCviol    = 5;
  nReportedMCviol  = 0;

  // Resize Paccept to the maximum number of elements
  Paccept.resize(max(weightsPtr->nWeights(),1));
  PME = 0.0;

  // Hard system(s) and MEclass
  hardSys.clear();
  hardSys.resize(0);
  MEclass.clear();
  MEclass.resize(0);

  // Statistics
  nTrialsSum         = 0;
  nTrialsVeto        = 0;
  nTrialsHull        = 0;
  nTrialsHullMassive = 0;
  nTrialsCutM        = 0;
  nTrialsCutQ        = 0;
  nClosePSforHQ      = 0;
  nMCviols.clear();
  nMCviols.resize(3);
  maxMCviols.clear();
  maxMCviols.resize(3);
  nNegPas.clear();
  nNegPas.resize(3);
  minNegPas.clear();
  minNegPas.resize(3);
  for (int i=0; i<3; i++) {
    nMCviols[i]   = 0;
    maxMCviols[i] = 0.0;
    nNegPas[i]    = 0;
    minNegPas[i]  = 0.0;
  }

  // Internal Histograms
  if (verbose >= 2) initVinciaHistos();

  // Initialize factorization scale and parameters for shower starting scale
  q2fac          = 0.0;
  qFac           = 0.0;
  qMaxMatch      = settingsPtr->mode("Vincia:QmaxMatch");
  q2maxFudge     = pow2(settingsPtr->parm("Vincia:QmaxFudge"));
  q2maxFudgeMPI  = pow2(settingsPtr->parm("Vincia:QmaxFudgeMPI"));

  // Parameters for merging
  string mProc   = settingsPtr->word("Merging:Process");
  doMergeSav     = !(mProc.compare("void") == 0);
  incomingType   = 0;
  if (doMergeSav) {
         if (mProc.find("e+e->")   != string::npos) incomingType = 1;
    else if (mProc.find("ta+ta->") != string::npos) incomingType = 2;
    else if (mProc.find("mu+mu->") != string::npos) incomingType = 3;
    else if (mProc.find("pp>")     != string::npos) incomingType = 10;
  } else {
    int idBeamA = abs(settingsPtr->mode("Beams:idA"));
    int idBeamB = abs(settingsPtr->mode("Beams:idB"));
         if ( (idBeamA == 11  ) && (idBeamB == 11  ) ) incomingType = 1;
    else if ( (idBeamA == 13  ) && (idBeamB == 13  ) ) incomingType = 2;
    else if ( (idBeamA == 15  ) && (idBeamB == 15  ) ) incomingType = 3;
    else if ( (idBeamA == 2212) && (idBeamB == 2212) ) incomingType = 10;
  }

  // Reset saved particle state
  indexSav.resize(0);
  partsSav.resize(0);    

  // Initialize USR antennae. Not done before to make sure we catch if
  // something was changed in Settings before.
  if (verbose >= 6)
    cout<<" TimeAntShower(): initializing antennaSet"<<endl;
  antSetPtr->init();

  // Print VINCIA header and list of parameters
  if (verbose >= 1 && !headerIsPrinted) header();

  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::init","end --------------");

  // Is initialized now
  isInit = true;

}

//*********
  
// Limition of first emission.

bool TimeAntShower::limitPTmax(Event& event, double, double) {
  // We handle scale internally, unless merging is applied.
  if (!doMergeSav) return false;
  // Leptonic initial states
  if ( event[1].isLepton() && event[2].isLepton() ) {
    int nFSparton = 0;
    for (int i=1; i<(int)event.size(); i++) if (event[i].isFinal()) {
      int id = event[i].idAbs();
      if ( (id == 21) || (id < 6) ) nFSparton++;
    }
    if (nFSparton == 2) return true;
  }
  // Hadronic initial states
  if ( event[1].isHadron() && event[2].isHadron() ) {
    int  nFSparton = 0;
    bool isQCDnow  = true;
    for (int i=3; i<(int)event.size(); i++) {
      int id = event[i].idAbs();
      if ( (id != 21) && (id > 5) ) isQCDnow = false;
      else if (event[i].isFinal()) nFSparton++; 
    } 
    if (isQCDnow && (nFSparton == 2)) return true;
    if (nFSparton == 0) return true;
  }
  return false;
}

//*********
  
// Top-level routine to do a full time-like shower in resonance decay.
  
int TimeAntShower::shower(int iBeg, int iEnd, Event& event, double pTmax, 
  int nBranchMax) {
    
  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::shower","begin --------------");
    
  // Add new system, automatically with two empty beam slots.
  int iSys = partonSystemsPtr->addSys();
    
  // Verbose output
  if (verbose >= 8) 
    printOut("TimeAntShower::shower","preparing to shower. System no. "
      +num2str(iSys));
    
  // Loop over allowed range to find all final-state particles.
  Vec4 pSum;
  for (int i = iBeg; i <= iEnd; ++i) if (event[i].isFinal()) {
      partonSystemsPtr->addOut( iSys, i);
      pSum += event[i].p();
    }
  partonSystemsPtr->setSHat( iSys, pSum.m2Calc() );
    
  // Let prepare routine do the setup.    
  prepare(iSys, event, false);
    
  // Begin evolution down in pT from hard pT scale. 
  int nBranchNow = 0;
  do {
    double pTtimes = pTnext(event, pTmax, 0.);
      
    // Do a final-state emission (if allowed).
    if (pTtimes > 0.) {
      if (branch(event)) ++nBranchNow; 
      pTmax = pTtimes;
    }
      
    // Keep on evolving until nothing is left to be done.
    else pTmax = 0.;
  } while (pTmax > 0. && (nBranchMax <= 0 || nBranchNow < nBranchMax)); 
    
  // Return number of emissions that were performed.
  return nBranchNow;
    
}
  
//*********
  
// Prepare system for evolution.
// Note: this function so far assumes that the system is input
// in colour order. A generalization would be needed to organize
// systems that are not in colour order.

void TimeAntShower::prepare(int iSys, Event& event, bool limitPTmaxIn) {
    
  // Verbose output
  if (verbose >= 6)
    printOut("TimeAntShower::prepare","begin --------------");

  // If verbose, list event before showering
  if (verbose >= 4) {
    if (verbose >= 5) event.list();
    printOut("TimeAntShower::prepare","preparing system no. "
      +num2str(iSys)+" for showering. Size = "
      +num2str(partonSystemsPtr->sizeOut(iSys))+
      " limitPTmax = "+bool2str(limitPTmaxIn));
  }
    
  // Statistics (zero info on qBranch scales)
  if (iSys < 4) {
    for (int j=0; j<11; ++j) {
      qBranch[iSys][j]=0.0;
      pTphys[iSys][j]=0.0;
    }
  }
    
  // Counter for accepted events
  nAccepted = max(nAccepted, infoPtr->nAccepted());
    
  // First time prepare is called for an event
  bool firstCallInEvent = (iSys == 0);
  // getCounter(3) number of times a Pythia::next() call has been begun.
  int nCallPythiaNextNow = infoPtr->getCounter(3);
  if (nCallPythiaNextNow > nCallPythiaNext) {
    nCallPythiaNext  = nCallPythiaNextNow;
    nVetoUserHooks   = 0;
    nFailHadLevel    = 0;
    firstCallInEvent = true;
  }
  // Last event got accepted.
  long nSelectedNow = infoPtr->nSelected();
  if (nSelectedNow > nSelected) {
    nSelected        = nSelectedNow;
    nVetoUserHooks   = 0;
    nFailHadLevel    = 0;
    firstCallInEvent = true;
  }
  // getCounter(10) number of times the selection of a new hard process has
  // been begun. = 1 if no user hooks veto, > 1 if user hooks veto.
  int nVetoUserHooksNow = (infoPtr->getCounter(10)-1);
  if (nVetoUserHooksNow > nVetoUserHooks) {
    nVetoUserHooks   = nVetoUserHooksNow;
    firstCallInEvent = true; 
  }
  // getCounter(14) number of times the loop over parton- and hadron-level processing 
  // has begun for a hard process. = 1 if everything after user hooks veto is
  // succesfull, > 1 eg if hadron level fails.
  int nFailHadLevelNow = (infoPtr->getCounter(14)-1);
  if (nFailHadLevelNow > nFailHadLevel) {
    nFailHadLevel    = nFailHadLevelNow;
    firstCallInEvent = true;
  }
    
  // Resetting for first time in new event
  if (firstCallInEvent) {
    //if (iSys == 0) {
    // Reset counters
    vinComPtr->resetCounters();

    // Reset the weights in new events
    weightsPtr->resetWeights(infoPtr->nAccepted());

    // Reset all vectors for the first time we are called
    hardSys.clear();          hardSys.resize(0);
    isResonance.clear();      isResonance.resize(0);
    MEclass.clear();          MEclass.resize(0);
    nQLneeded.clear();        nQLneeded.resize(0);
    Hint.clear();             Hint.resize(0);
    additionalJets.clear();   additionalJets.resize(0);
    indexSav.clear();         indexSav.resize(0);
    partsSav.clear();         partsSav.resize(0);
    polarizedSys.clear();     polarizedSys.resize(0);
    branchelementals.clear(); branchelementals.resize(0);
    Q2hat.clear();            Q2hat.resize(0);
    nBranch.clear();          nBranch.resize(0);
    nBranchFSR.clear();       nBranchFSR.resize(0);
    mSystem.clear();          mSystem.resize(0);
    nG.clear();               nG.resize(0);
    nQQ.clear();              nQQ.resize(0);
    savedBranchings.clear();
  }
  
  // Sanity check: at least two particles in system
  int sizeSystem = partonSystemsPtr->sizeAll(iSys);
  if (sizeSystem <= 1) return;

  // Resize to make sure we can access the iSys'th element
  int currentSize = branchelementals.size();
  if ( currentSize < (iSys + 1) ) {
    hardSys.resize(iSys + 1);
    MEclass.resize(iSys + 1);
    nQLneeded.resize(iSys + 1);
    Hint.resize(iSys + 1);
    additionalJets.resize(iSys + 1);
    indexSav.resize(iSys + 1);
    partsSav.resize(iSys + 1);
    polarizedSys.resize(iSys + 1);
    branchelementals.resize(iSys + 1);
    Q2hat.resize(iSys + 1);
    nBranch.resize(iSys + 1);
    nBranchFSR.resize(iSys + 1);
    mSystem.resize(iSys + 1);
    isResonance.resize(iSys + 1);
    nG.resize(iSys + 1);
    nQQ.resize(iSys + 1);
  }
    
  // Check if antennae in previous systems are still around. E.g., if this is a
  // hadron->partons decay, the hard-process antennae are no longer present.
  for (int iSysPrev=iSys-1; iSysPrev>=0; --iSysPrev) {
    vector<bool> iErase;
    for (int iAnt=(int)branchelementals[iSysPrev].size()-1; iAnt>=0; --iAnt) {
      int i1 = branchelementals[iSysPrev][iAnt].iOld1;
      int i2 = branchelementals[iSysPrev][iAnt].iOld2;
      // Check if the endpoints have decayed
      if (event[i1].status() < 0 || event[i2].status() < 0)
        iErase.push_back(iAnt);
    }
    // Erase any antennae whose endpoints were marked as decayed
    for (int iAnt=0; iAnt<(int)iErase.size(); ++iAnt) {
      branchelementals[iSysPrev].erase(branchelementals[iSysPrev].begin() + iErase[iAnt]);
    }
  }
    
  // We don't have a starting scale for this system yet
  Q2hat[iSys] = 0.0;
  // After prepare we always have zero branchings
  nBranch[iSys] = 0;
  nBranchFSR[iSys] = 0;
  // Initialize polarization with false or ISR value if possible
  polarizedSys[iSys] = ( ((int)ISRshowerPtr->polarizedSys.size() > iSys) ?
    ISRshowerPtr->polarizedSys[iSys] : false );
  
  if (iSys >= (int)ISRshowerPtr->polarizedSys.size())
    ISRshowerPtr->polarizedSys.resize(iSys+1);
  ISRshowerPtr->polarizedSys[iSys] = usePolarizationNow;
    
  // Note that we have not yet binned the first branching scale
  if (iSys == 0) firstQBranchBinned = false;
    
  // Vector to find total invariant mass of final state system
  Vec4 pSum;

  if (verbose >= 9) cout<<" Making colour maps "<<endl;
    
  // Find and save all colors and anticolors
  // Not only FF, to find the neighbours
  map<int,int> indexOfAcol;
  map<int,int> indexOfCol;
  vector< pair<int,int> > antFF;
    
  // Loop over event record. Find indices of FS particles with color lines
  // Make new copies so any modifications by VINCIA 
  // (eg polarization assignment) are clearly recorded
  int iAncestor  = 0;
  for (int i = 0; i < sizeSystem; ++i) {    
    int i1 = partonSystemsPtr->getAll( iSys, i);
    // Skip if not present in final state
    if ( i1 <= 0 ) continue;
    // Save to colour maps (taking neg-index sextets into account)
    int col  = event[i1].col();
    int acol = event[i1].acol();
    if (!event[i1].isFinal()) { // Switch colours for initial partons
      col  = acol;
      acol = event[i1].col();
    }
    // Save colours
    if (col > 0) indexOfCol[col] = i1;
    else if (col < 0) indexOfAcol[-col] = i1;
    if (acol > 0) indexOfAcol[acol] = i1;
    else if (acol < 0) indexOfCol[-acol] = i1;
    // Check for ancestor only for final state particles
    if (event[i1].isFinal()) {
      // Check if all the partons have a single common mother
      if (iAncestor >= 0) {
        int iMot1 = event[i1].mother1();
        int iMot2 = event[i1].mother2();
        if (iMot2 == 0) iMot2 = iMot1;
        if (iAncestor == 0) iAncestor = iMot1;
        // If more than one ancestor found, set iAncestor = -1
        if (iMot1 != iAncestor || iMot2 != iMot1) iAncestor = -1; 
      }
      // Sum total FS momentum
      pSum += event[i1].p();
      // Update list of final-final connections
      if (col > 0 && indexOfAcol.count(col) == 1) {
        if (event[indexOfAcol[col]].isFinal())
          antFF.push_back( make_pair(i1,indexOfAcol[col]) );
      }
      if (acol < 0 && indexOfAcol.count(-acol) == 1) {
        if (event[indexOfAcol[-acol]].isFinal())
          antFF.push_back( make_pair(i1,indexOfAcol[-acol]) );
      }
      if (acol > 0 && indexOfCol.count(acol) == 1) {
        if (event[indexOfCol[acol]].isFinal())
          antFF.push_back( make_pair(indexOfCol[acol], i1) );
      }
      if (col < 0 && indexOfCol.count(-col) == 1) {
        if (event[indexOfCol[-col]].isFinal())
          antFF.push_back( make_pair(indexOfCol[-col], i1) );
      }
    }
    // Add to list of partons
    indexSav[iSys].push_back( i1 );
    partsSav[iSys].push_back( event[i1] );
  }

  // Sanity check: If common ancestor found, check that it is a resonance
  if (iAncestor > 0) {
    if ( !event[iAncestor].isResonance() ) iAncestor=-1;
    else { // Add resonance at the beginning
      indexSav[iSys].insert( indexSav[iSys].begin(), iAncestor );
      partsSav[iSys].insert( partsSav[iSys].begin(), event[iAncestor] );
    }
  }
  isResonance[iSys] = (iAncestor > 0);

  if (verbose >= 9) cout<<" Checking process class for MECs"<<endl;
    
  // Check the process class for MECs and whether system is "hard"
  bool hardSysNow = false;
  int nPartsTot   = (int)partsSav[iSys].size();
  int nColIn = 0, nColOut = 0, nGluTot = 0, nBosIn = 0, nBosOut = 0;
  int nZandW = 0, iHiggs = -1;
  // Also check if at least one QCD parton in final state (used when setting
  // upper limit for evolution)
  bool hasFSpartons = false;
  for (int ip=0; ip<nPartsTot; ip++) {
    // Is this a coloured object?
    int colSum = partsSav[iSys][ip].col()+partsSav[iSys][ip].acol();
    if (colSum > 0) partsSav[iSys][ip].isFinal() ? nColOut++ : nColIn++;
    // Is this a boson we have in the MECs?
    else if ( (partsSav[iSys][ip].idAbs() > 22) && 
      (partsSav[iSys][ip].idAbs() < 26) ) {
      partsSav[iSys][ip].isFinal() ? nBosOut++ : nBosIn++;
      partsSav[iSys][ip].idAbs() == 25 ? (iHiggs = ip) : nZandW++;
    }
    // Is this a gluon?
    if (partsSav[iSys][ip].id() == 21) nGluTot++;
    // If status code is 21,22,23 we are in the hard system
    if ( abs(partsSav[iSys][ip].statusAbs()-22) < 2 ) hardSysNow = true;
    // If parton is final-state and can be produced by shower? 
    if ( partsSav[iSys][ip].isFinal() ) { 
      int idAbs = partsSav[iSys][ip].idAbs();
      if ( (idAbs >= 1 && idAbs <= 5) || idAbs == 21) hasFSpartons = true;
    }
  }
  // Ask ISR if the system is hard, if so adjust hardSysNow
  // NOTE this needs to be done if ISR made q's massless
  if (ISRshowerPtr->isSysHard(iSys)) hardSysNow = true;
  int nColTot = nColIn + nColOut;
  int nBosTot = nBosIn + nBosOut;
  // Find process class for MG4 MECs, adjust number of branchings (used for 
  // finding QCD order), adjust number of additional jets present
  int MEclassNow   = -1;
  meCorrectionsNow = meCorrections;
  if ( (nColTot == nPartsTot) && (nColOut > 0) && (nColIn > 0) ) {
    MEclassNow     = pureQCD;
    nBranch[iSys]  = nPartsTot-4;
  } else if ( (nBosTot == 1) && (nBosTot+nColTot == nPartsTot) ) {
    MEclassNow     = (nBosOut == 1 ? singleBosonProd : singleBosonDecay);
    nBranch[iSys]  = nColTot-2;
  } else meCorrectionsNow = false;
  // For merging only allow MECs in ordered events
  if (doMergeSav) {
    if (vinMergingHooksPtr->doOrderedMECs()) meCorrectionsNow = meCorrections;
    else meCorrectionsNow = false;
  }
  // Set the properties needed for MECs
  hardSys[iSys]        = hardSysNow;
  MEclass[iSys]        = MEclassNow;
  nQLneeded[iSys]      = nZandW;
  additionalJets[iSys] = nBranch[iSys];
  // For Higgs check if we should do interference
  if (iHiggs > -1 && !Hinterference) {
    Particle higgs = partsSav[iSys][iHiggs];
    int idSum = -999;
    if (MEclassNow == singleBosonDecay)
      idSum = event[higgs.daughter1()].id() + event[higgs.daughter2()].id();
    else if (MEclassNow == singleBosonProd)
      idSum = event[higgs.mother1()].id() + event[higgs.mother2()].id();
    if (idSum > -999) Hint[iSys] = ( (idSum == 42) ? 2 : 1 );
    else Hint[iSys] = 0;
  } else Hint[iSys] = 0;

  // Factorization scale
  if (hardSysNow) {
    q2fac = infoPtr->Q2Fac();
    qFac  = sqrt(q2fac);
  }

  // Store total invariant mass of the final state
  mSystem[iSys]  = m(pSum);

  // Check if 1->n or 2->n
  int nIn = 0;
  if (isResonance[iSys]) nIn = 1;
  else if (partonSystemsPtr->hasInAB(iSys)) nIn = 2;

  // Set FSR starting scale of this system
  if (isResonance[iSys]) {
    if (verbose >= 9) cout<<" Setting FSR starting scale for resonance decay"<<endl;
    Q2hat[iSys] = pow2(mSystem[iSys]);
  } else if (hardSysNow) {
    if (verbose >= 9) cout<<" Setting FSR starting scale for hard system"<<endl;
    // Set the initial value of the starting scale: factorisation scale or
    // phase-space maximum according to option
    if (qMaxMatch == 1 || (qMaxMatch == 0 && hasFSpartons) ) {
      // Wimpy: enhanceq2fac * factorisation scale from PYTHIA
      Q2hat[iSys] = q2maxFudge * q2fac;
    } else {
      // Power: Phase-space maximum 
      Q2hat[iSys] = m2( beamAPtr->p(), beamBPtr->p());
    }
    // TMP!
    // Could add an option to let VINCIA calculate its own preferred scale,
    // possibly including a PDF ratio to re-normalise the cross section
    // to use that value.
  } else if (nIn == 2) {
    if (verbose >= 9) cout<<" Setting properties of MPI system"<<endl;
    // Set starting scale for MPI systems: min of incoming parton scales.
    // Find positions of incoming colliding partons.
    int in1 = partonSystemsPtr->getInA(iSys);
    int in2 = partonSystemsPtr->getInB(iSys);
    Q2hat[iSys] = q2maxFudgeMPI * pow2(min(event[in1].scale(),event[in2].scale()));
  } else {
    // Assume hadron -> partons decay. Starting scale = mSystem. 
    Q2hat[iSys] = pow2(mSystem[iSys]);
  }

  // Verbose output
  if (verbose >= 9) cout<<" Q2fac from infoPtr = "<<q2fac<<" -> QF = "<<qFac
                        <<"; Q2start for this system = "<<Q2hat[iSys]
                        <<" -> Qstart = "<<sqrt(Q2hat[iSys])<<endl;

  // Sanity check: can only shower systems with more than 0 FF connection
  if (antFF.size() <= 0) return;

  // VINCIA starts by creating a copy of the particles it will work with
  // Colour and polarization information may be changed or added,
  // respectively, and masses may be set to zero for partons VINCIA
  // wants to treat as massless.
  treatAsMassless = true;
  int nISmassless = 0;
  int nISmassive  = 0;
  int nFS         = 0;
  vector<int> jMakeMassless;
  for (int i=0; i<(int)indexSav[iSys].size(); i++) {
    int nr = indexSav[iSys][i];
    if ( event[nr].m2Calc() > TINYMASS ) {
      // For resonance decays allow massive final state
      if (isResonance[iSys]) {
        if (event[nr].idAbs() <= nFlavZeroMassKin)
          jMakeMassless.push_back(i);
        // Ignore resonance itself
        else if (i>nIn) treatAsMassless = false;
      }
      // For everything else map everyone with colors to massless
      else if (abs(event[nr].col())+abs(event[nr].acol()) > 0)
        jMakeMassless.push_back(i);
      if (!event[nr].isFinal()) nISmassive++;
    }
    else if (!event[nr].isFinal()) nISmassless++;
    if (event[nr].isFinal()) nFS++;
  }
  // Special case, 2->n, with only massless incoming guys
  bool doSpecial    = ( (nISmassive == 0) && (nISmassless = 2) );
  int  nToMassless  = (int)jMakeMassless.size();
  bool madeMassless = (jMakeMassless.size() > 0);
    
  // Sanitize input. 
  // Bottom and lighter quarks are always bookkept as massless by Vincia
  // (Top quarks may be made massless as well, for ultra-high energies)
  while (jMakeMassless.size() > 0) {
    // Special case, 2->n, with only massless incoming guys and either
    // 2 partons to make massless or >1 FS guys with 1 to make massless
    if ( doSpecial && ( (nToMassless == 2) || 
        (nToMassless == 1 && nFS > 1) ) ) {
      int i1 = jMakeMassless[0];
      int i2 = jMakeMassless[nToMassless-1];
      if (i2 == i1) { // Need massive partner
        for (int ip=0; ip<(int)partsSav[iSys].size(); ip++) {
          if (partsSav[iSys][ip].isFinal() && (ip != i1)) i2 = ip;
        }
      }
      // Go to rest frame of pair
      Vec4 p1 = partsSav[iSys][i1].p();
      Vec4 p2 = partsSav[iSys][i2].p();
      RotBstMatrix M;
      M.toCMframe(p1,p2);
      p1.rotbst(M);
      p2.rotbst(M);
      double eCMs = p1.e() + p2.e();
      // Check if both to be massless
      if (nToMassless == 2 || partsSav[iSys][i2].m2() <= TINY) {
        p1 = Vec4(0.0, 0.0,  eCMs/2.0, eCMs/2.0);
        p2 = Vec4(0.0, 0.0, -eCMs/2.0, eCMs/2.0);
        // PYTHIA stores mass as "5th" component, to preserve precision
        partsSav[iSys][i1].m(0.0);
        partsSav[iSys][i2].m(0.0);
      }
      // Else just make i1 massless
      else {
        double e1 = (eCMs - partsSav[iSys][i2].m2()/eCMs)/2.;
        double e2 = eCMs - e1;
        p1 = Vec4(0.0, 0.0,  e1, e1);
        p2 = Vec4(0.0, 0.0, -e1, e2);
        // PYTHIA stores mass as "5th" component, to preserve precision
        partsSav[iSys][i1].m(0.0);
      }
      M.invert();
      p1.rotbst(M);
      p2.rotbst(M);
      partsSav[iSys][i1].p(p1);
      partsSav[iSys][i2].p(p2);
      jMakeMassless.clear();
      jMakeMassless.resize(0);
      continue;
    }
    // Special case, 2->n, with only massless incoming guys
    // use mapToMassless
    if (doSpecial) {
      vector<Vec4> momenta;
      for (int i=0; i<(int)jMakeMassless.size(); i++)
        momenta.push_back(partsSav[iSys][jMakeMassless[i]].p());
      // Mapping
      vinCluPtr->mapToMassless(momenta, 0);
      // Set new momenta
      for (int i=0; i<(int)jMakeMassless.size(); i++) {
        partsSav[iSys][jMakeMassless[i]].p(momenta[i]);
        // PYTHIA stores mass as "5th" component, to preserve precision
        partsSav[iSys][jMakeMassless[i]].m(0.0);
      }
      jMakeMassless.clear();
      jMakeMassless.resize(0);
      continue;
    }
    // Start with last parton
    // (-> easy to use pop_back to erase it when we are done with it)
    int i1 = jMakeMassless[jMakeMassless.size()-1];
    Particle& parton1 = partsSav[iSys][i1];
    // If this particle was already made massless, erase and continue
    if (parton1.m2() <= TINY) {
      jMakeMassless.pop_back();
      continue;
    }
    // Just make this parton massless
    if (parton1.idAbs() > 6) {
      double facInv = parton1.pAbs()/parton1.e();
      if (facInv > 0.0) parton1.rescale3(1./facInv);
      // PYTHIA stores mass as "5th" component, to preserve precision
      partsSav[iSys][i1].m(0.0);
      partsSav[iSys][i1].p(parton1.p());
    } else { // Find matching partner
      bool isIn1  = (!parton1.isFinal());
      bool isAnti = (parton1.id() < 0);
      int col1 = (isAnti ? parton1.acol() : parton1.col());
      for (int i2=0; i2<(int)partsSav[iSys].size(); ++i2) {
        // Skip self
        if (i2 == i1) continue;
        Particle& parton2 = partsSav[iSys][i2];
        bool isIn2  = (!parton2.isFinal());
        // Look for opposite-colour 
        int col2 = ( ((isIn1 && isIn2) || (!isIn1 && !isIn2)) ? 
          (isAnti ? parton2.col() : parton2.acol()) :
          (isAnti ? parton2.acol() : parton2.col()) );
        if (col1 == col2) {
          // Go to rest frame of parton pair
          Vec4 p1 = parton1.p();
          Vec4 p2 = parton2.p();
          RotBstMatrix M;
          M.toCMframe(p1,p2);
          p1.rotbst(M);
          p2.rotbst(M);
          double eCMs = p1.e() + p2.e();
          // Check if both to be massless
          if (parton2.m2() <= TINY) {
            p1 = Vec4(0.0, 0.0,  eCMs/2.0, eCMs/2.0);
            p2 = Vec4(0.0, 0.0, -eCMs/2.0, eCMs/2.0);
            // PYTHIA stores mass as "5th" component, to preserve precision
            partsSav[iSys][i1].m(0.0);
            partsSav[iSys][i2].m(0.0);
          }
          // Else just make this parton massless
          else {
            double e1 = (eCMs - parton2.m2()/eCMs)/2.;
            double e2 = eCMs - e1;
            p1 = Vec4(0.0, 0.0,  e1, e1);
            p2 = Vec4(0.0, 0.0, -e1, e2);
            // PYTHIA stores mass as "5th" component, to preserve precision
            partsSav[iSys][i1].m(0.0);
          }
          M.invert();
          p1.rotbst(M);
          p2.rotbst(M);
          partsSav[iSys][i1].p(p1);
          partsSav[iSys][i2].p(p2);
          break;
        }
      }
    }
    // Done with this parton
    jMakeMassless.pop_back();
  }
    
  // Verbose output
  if (verbose >= 6) {
    if (treatAsMassless) 
      printOut("TimeAntShower::prepare","Will treat input particles as relativistic");
    else 
      printOut("TimeAntShower::prepare","Will treat input particles as massive");
  }
    
  // Polarization: check if VINCIA should (attempt to) assign helicities
  bool doPolarize = false;
  usePolarizationNow = (treatAsMassless && helicityShower);
  if (usePolarizationNow) {
    // Check whether event is already polarized, allow bosons to be unpolarized
    bool isPolarized = true;
    for (int i=0; i<(int)partsSav[iSys].size(); ++i) 
      if ( !(partsSav[iSys][i].idAbs() > 21 && partsSav[iSys][i].idAbs() < 38)
        && partsSav[iSys][i].pol() == 9 ) isPolarized = false;
    // If not, see if we can polarize it
    if (!isPolarized) {
      matrixElementCorrectionsPtr->prepMECs(MEclassNow, nZandW, Hint[iSys]);
      usePolarizationNow = 
        matrixElementCorrectionsPtr->polarize(nIn,partsSav[iSys]);
      if (usePolarizationNow) doPolarize = true;
    }
  }
  // Set system flag and update general flag
  polarizedSys[iSys] = usePolarizationNow;
  if ((int)ISRshowerPtr->polarizedSys.size() > iSys)
    ISRshowerPtr->polarizedSys[iSys] = usePolarizationNow;
  usePolarizationNow = (treatAsMassless && helicityShower);
  if (verbose >= 7) printOut("TimeAntShower::prepare",
    "doPolarize = "+bool2str(doPolarize));
    
  // Colorise: assign each dipole a random color index between 1 and 9.
  // (For use in colour reconnections)
  map<int,int> colorMap;
  bool doColorize = (CRmode >= 1);
  if (doColorize) {
    // TODO once this is used we need to make sure that Pythia knows how
    // many colTags we used
    int nextTag = event.nextColTag();
    doColorize = reconnectionsPtr->getColourMap(colorMap,partsSav[iSys],nextTag);
  }
  if (verbose >= 7) printOut("TimeAntShower::prepare",
    "doColorize = "+bool2str(doColorize));
    
  // Create new (polarized, massless, color-indexed) copies of
  // original parton system;
  // only if we changed anything to avoid with changing the
  // system after ISR already did
  if (madeMassless || doPolarize || doColorize) {
    map<int,int> iChanged;
    indexOfCol.clear();
    indexOfAcol.clear();
    for (int i=0; i<(int)partsSav[iSys].size(); i++) {
      int iOld = indexSav[iSys][i];
      int iNew = iOld;
      if (event[iOld].isFinal()) iNew = event.copy(iOld,51);
      else {
        iNew = event.copy(iOld,-event[iOld].statusAbs());
        // Everyone who as iOld has daughter needs to get
        // iNew as daughter now
        for (int ip=0; ip<(int)event.size(); ip++) if (ip < iOld) {
            if (event[ip].daughter1() == iOld) {
              // iNew > daughter2, that's fine
              event[ip].daughters(iNew, event[ip].daughter2());
            }
            if (event[ip].daughter2() == iOld) {
              // iNew < daughter2, need swap
              event[ip].daughters(iNew, event[ip].daughter1());
            }
          }
      }
      iChanged[iOld] = iNew;
      // Update 4-momentum and mass (eg if made massless)
      event[iNew].p(partsSav[iSys][i].p());
      event[iNew].m(partsSav[iSys][i].m());
      // Set polarizations and colors if new ones defined
      if (polarizedSys[iSys]) event[iNew].pol(partsSav[iSys][i].pol());
      if (doColorize) {
        event[iNew].col(colorMap[event[iNew].col()]);
        event[iNew].acol(colorMap[event[iNew].acol()]);
      }
      // Update properties in indexSav[iSys] and partsSav[iSys]
      indexSav[iSys][i] = iNew;
      partsSav[iSys][i] = event[iNew];
      // Update the list of systems.
      partonSystemsPtr->replace(iSys, iOld, iNew); 
      // Update color maps
      int col  = event[iNew].col();
      int acol = event[iNew].acol();
      if (!event[iNew].isFinal()) { // Switch colours for initial partons
        col  = acol;
        acol = event[iNew].col();
      }
      if (col  != 0) indexOfCol[col]   = iNew;
      if (acol != 0) indexOfAcol[acol] = iNew;
    }
    // Let ISR know we did changes to the system
    ISRshowerPtr->updateAfterFSR(iSys, event, iChanged);
  }
    
  if (verbose >= 6) {
    printOut("TimeAntShower::prepare", "saved the following "
      "particles: nr id col acol p");
    for (int i=0; i<(int)partsSav[iSys].size(); i++) {
      cout << "  " << setw(3) << indexSav[iSys][i];
      cout << " " << setw(2) << partsSav[iSys][i].id();
      cout << " " << setw(3) << (int)partsSav[iSys][i].pol();
      cout << " " << setw(3) << partsSav[iSys][i].col();
      cout << " " << setw(3) << partsSav[iSys][i].acol();
      cout << " " << partsSav[iSys][i].p();
    }
  }
    
  // Now loop over colored particles to create branchelementals (=antennae)
  for (map<int,int>::iterator it=indexOfCol.begin(); 
       it != indexOfCol.end(); ++it) {
    int col  = it->first;
    int i1   = it->second;        // iOld1/iNew1 is colour
    int i2   = indexOfAcol[col];  // iOld2/iNew3 is anticolour
    if (col == 0 || i1 == 0 || i2 == 0) continue;
    // Don't include not FF antennae
    if (!event[i1].isFinal() || !event[i2].isFinal() ) continue;
    if (verbose >= 7) {
      printOut("TimeAntShower::prepare",
        "Creating antenna between "+num2str(i1,4)+", "
        +num2str(i2,4)+": col = "+num2str(col,5));
    }
      
    // Store trial QCD antenna and add trial generators depending on type
    BranchelementalFSR trial(event,i1,i2);
    trial.resetTrialGenerators(trialGenPtrs, false);

    // Save system and ancestor
    trial.system    = iSys;
    trial.iAncestor = iAncestor;
    
    trial.renewTrial();
    trial.mCalc();
      
    // Find neigbor on i1 side
    if (event[i1].acol() != 0) {
      int colL = event[i1].acol();
      trial.iL = indexOfCol[colL];
    }
      
    // Find neigbor on i2 side
    if (event[i2].col() != 0) {
      int colR = event[i2].col();
      trial.iR = indexOfAcol[colR];        
    }
      
    // Save branchelemental
    branchelementals[iSys].push_back(trial);
  }
    
  // Count up number of gluons and quark pairs
  nG[iSys]  = 0;
  nQQ[iSys] = 0;
  for (int i=0; i<(int)partsSav[iSys].size(); i++) {
    if (partsSav[iSys][i].id()==21) nG[iSys]++;
    else if (abs(partsSav[iSys][i].id()) < 7) nQQ[iSys]++;
  }
  // Half the quarks to get quark pairs
  nQQ[iSys] = nQQ[iSys]/2;

  // Sanity check
  if (branchelementals[iSys].size() <= 0) {
    if (verbose >= 5) printOut("TimeAntShower::prepare",
      "did not find any antennae: exiting.");
    return; 
  }
      
  //If verbose output, list dipoles
  if (verbose >= 5) list(); 
    
  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::prepare","finished --------------");

}
  
//*********
  
// Update list of branchelementals after ISR radiation
  
void TimeAntShower::update( int iSys, Event& event, bool) {

  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::update","begin --------------");
    
  // If verbose, list event before showering
  if (verbose >= 5) event.list();
    
  // Count number of branches
  nBranch[iSys]++;
    
  // Particles in the list are already updated by ISR
  // Find and save all colors and anticolors
  // Not only FF, to find the neighbours
  map<int,int> indexOfAcol;
  map<int,int> indexOfCol;
  vector< pair<int,int> > antFF;
  int sizeSystem = partonSystemsPtr->sizeAll(iSys);
  for (int i = 0; i < sizeSystem; ++i) {   
    int i1 = partonSystemsPtr->getAll( iSys, i);
    if ( i1 <= 0 ) continue;
    // Save to colour maps
    int col  = event[i1].col();
    int acol = event[i1].acol();
    if (!event[i1].isFinal()) { // Switch colours for initial partons
      col  = acol;
      acol = event[i1].col();
    }
    // Save colours
    if (col > 0) indexOfCol[col] = i1;
    else if (col < 0) indexOfAcol[-col] = i1;
    if (acol > 0) indexOfAcol[acol] = i1;
    else if (acol < 0) indexOfCol[-acol] = i1;
    if (event[i1].isFinal()) {
      // Update list of final-final connections
      if (col > 0 && indexOfAcol.count(col) == 1) {
        if (event[indexOfAcol[col]].isFinal())
          antFF.push_back( make_pair(i1,indexOfAcol[col]) );
      }
      if (acol < 0 && indexOfAcol.count(-acol) == 1) {
        if (event[indexOfAcol[-acol]].isFinal())
          antFF.push_back( make_pair(i1,indexOfAcol[-acol]) );
      }
      if (acol > 0 && indexOfCol.count(acol) == 1) {
        if (event[indexOfCol[acol]].isFinal())
          antFF.push_back( make_pair(indexOfCol[acol], i1) );
      }
      if (col < 0 && indexOfCol.count(-col) == 1) {
        if (event[indexOfCol[-col]].isFinal())
          antFF.push_back( make_pair(indexOfCol[-col], i1) );
      }
    }
  }

  // Sanity check: can only shower systems with more than 1 FF connection
  if (antFF.size() <= 0) return;
    
  // Count up number of gluons and quark pairs
  nG[iSys]  = 0;
  nQQ[iSys] = 0;
  for (int i=0; i<(int)partsSav[iSys].size(); i++) {
    if (partsSav[iSys][i].id()==21) nG[iSys]++;
    else if (abs(partsSav[iSys][i].id()) < 7) nQQ[iSys]++;
  }
  // Half the quarks to get quark pairs
  nQQ[iSys] = nQQ[iSys]/2;
    
  if (verbose >= 6) {
    printOut("TimeAntShower::update", "saved the following "
      "particles: nr id col acol p");
    for (int i=0; i<(int)partsSav[iSys].size(); i++) {
      cout << "  " << setw(3) << indexSav[iSys][i];
      cout << " " << setw(2) << partsSav[iSys][i].id();
      cout << " " << setw(3) << partsSav[iSys][i].col();
      cout << " " << setw(3) << partsSav[iSys][i].acol();
      cout << " " << partsSav[iSys][i].p();
    }
  }
    
  // Loop over the antennae, we need to reset neighbours and check
  // antenna partons
  for (int i=0; i<(int)branchelementals[iSys].size(); i++) {
    BranchelementalFSR* antPtr = &branchelementals[iSys][i];
    bool changed   = false;
    // Copy old ids
    int id1        = antPtr->old1.id();
    int id2        = antPtr->old2.id();
    // Antenna partons
    if (!event[antPtr->iOld1].isFinal()) {
      // Renew trials if partons participated in a branching in IF
      // Status 44 = outgoing shifted by a branching
      if (event[antPtr->iOld1].status() != 44) antPtr->renewTrial();
      // iOld1 is always the one with the colour
      antPtr->iOld1 = indexOfCol[antPtr->old1.col()];
      changed = true;
    }
    if (!event[antPtr->iOld2].isFinal()) {
      if (event[antPtr->iOld2].status() != 44) antPtr->renewTrial();
      // iOld1 is always the one with the anticolour
      antPtr->iOld2 = indexOfAcol[antPtr->old2.acol()];
      changed = true;
    }
    antPtr->old1 = event[antPtr->iOld1];
    antPtr->old2 = event[antPtr->iOld2];
    // Neighbours
    if (antPtr->old1.id() == 21)
      antPtr->iL = indexOfCol[antPtr->old1.acol()];
    else antPtr->iL = 0;
    if (antPtr->old2.id() == 21)
      antPtr->iR = indexOfAcol[antPtr->old2.col()];
    else antPtr->iR = 0;
    // Recalculate mass and check we need to reset trial generators
    if (changed) {
      antPtr->mCalc();
      if (id1 != antPtr->old1.id() || id2 != antPtr->old2.id())
        antPtr->resetTrialGenerators(trialGenPtrs, false);
    }
    // Remove the antennae out of the list
    // This way we can check later if ISR added a new FF antenna
    // iOld1 / iOld2 is colour / anticolour
    pair<int,int> pairNow = make_pair(antPtr->iOld1,antPtr->iOld2);
    vector< pair<int,int> >::iterator iter;
    iter = find (antFF.begin(), antFF.end(), pairNow);
    if (iter != antFF.end()) antFF.erase(iter);
  }
    
  // Is there a FF connection left?
  for (int i=0; i<(int)antFF.size(); i++) {
    int i1   = antFF[i].first;   // iOld1/iNew1 is colour
    int i2   = antFF[i].second;  // iOld2/iNew3 is anticolour
    // Don't include not FF antennae
    if (!event[i1].isFinal() || !event[i2].isFinal() ) continue;
    if (verbose >= 7) {
      printOut("TimeAntShower::update",
        "Creating antenna between "+num2str(i1,4)+", "
        +num2str(i2,4)+": col = "+num2str(event[i1].col(),5));
    }
    // Store trial QCD antenna and add trial generators depending on type
    BranchelementalFSR trial(event,i1,i2);
    trial.resetTrialGenerators(trialGenPtrs, false);
    // Find neigbor on i1 side
    if (event[i1].acol() != 0) {
      int colL = event[i1].acol();
      trial.iL = indexOfCol[colL];
    }
    // Find neigbor on i2 side
    if (event[i2].col() != 0) {
      int colR = event[i2].col();
      trial.iR = indexOfAcol[colR];        
    }
    // Save system
    trial.system = iSys;
    // This an antenna created by ISR, hence not a resonance decay
    trial.iAncestor = 0;
    branchelementals[iSys].push_back(trial); 
  }
    
  for (int i=0; i<(int)branchelementals[iSys].size(); i++) {
    BranchelementalFSR* trial = &branchelementals[iSys][i];
    // Update saved antenna mass
    trial->mCalc();
    // If we went from matched to non-matched order renew all trials
    if (meCorrectionsNow && (nBranch[iSys]==matchingLO) && hardSys[iSys]) {
      trial->renewTrial();
    }
  }
    
  // Sanity check
  if (branchelementals[iSys].size() <= 0) {
    if (verbose >= 5) printOut("TimeAntShower::update",
      "did not find any antennae: exiting.");
    return; 
  }
      
  //If verbose output, list dipoles
  if (verbose >= 5) list(true);
  
  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::update","finished --------------");

}

//*********
  
// Select next pT in downwards evolution of the existing dipoles.
  
double TimeAntShower::pTnext(Event& event, double pTevolBegAll,
  double pTevolEndAll, bool /*isFirstTrial*/, bool /*doTrialIn*/) {

  // Sanity checks
  if (branchelementals.size() <= 0) return 0.0;
  if (!settingsPtr->flag("Vincia:FSR")) return 0.0;

  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::pTnext","begin --------------");

  // Verbose output
  if (verbose >= 4)
    printOut("TimeAntShower::pTnext",
      "(re)starting evolution between pTevolBegAll = " + num2str(pTevolBegAll) +
      " and pTevolEndAll = " + num2str(pTevolEndAll));

  // Checks: qEnd < all cutoffs?, no ants?
  bool qEndsmallerCutoff = true, zeroAnts = true;

  // Denote VINCIA scales by "q", PYTHIA ones by "pTevol". 
  double qOld    = pTevolBegAll;    
  double qEndAll = pTevolEndAll;

  // Initialize winner scale
  double qWin = 0.0;
  winnerPtr   = NULL;
  iSysWin     = -1;

  // Verbose output
  if (verbose >= 6) list(true);

  // TMP! saved branchings disabled
  /*
  // Sanitise list of saved branchings
  for (int i=(int)savedBranchings.size()-1; i>=0; --i) {
    int indx    = savedBranchings[i].getTrialIndex();
    // Sanitise list
    if (indx < 0) {
      savedBranchings.erase(savedBranchings.begin() + i);
      continue;
    }
    double qSav = savedBranchings[i].scaleSav[indx];
    // Sanitise list
    if (qSav < TINY) {
      savedBranchings.erase(savedBranchings.begin() + i);
      continue;
    }
  }

  // Initialise from saved branchings (trials generated earlier which
  // were postponed to occur at a lower scale)
  for (int i=0; i<(int)savedBranchings.size(); ++i) {
    int indx    = savedBranchings[i].getTrialIndex();
    double qSav = savedBranchings[i].scaleSav[indx];
    if (qSav > qEndAll && qSav > qWin) {
      winnerPtr  = &savedBranchings[i];
      qWin       = qSav;
      iSysWin    = winnerPtr->system;
    }
  }
  */

  // Loop over the systems we have ----------- BEGIN
  int nrSystems = (int)branchelementals.size();
  for (int iSys = 0; iSys < nrSystems; iSys++) {

    int nrAnts = (int)branchelementals[iSys].size();
    if (nrAnts <= 0) continue;
    zeroAnts = false;

    // Max Q scale for this system
    double qMax = ( doMergeSav ? qOld : min( qOld, sqrt(Q2hat[iSys]) ) );

    // Loop over antennae ----------- BEGIN
    for (int iAnt = 0; iAnt < nrAnts; ++iAnt) {

      // Shorthand for current dipole-antenna
      BranchelementalFSR* trialPtr = &branchelementals[iSys][iAnt];

      // Verbose output
      if (verbose >= 6) {
        printOut("TimeAntShower::pTnext",
          "* processing antenna " + num2str(iAnt,4)
          + " col = " + num2str(trialPtr->old1.col())
          + " nTrialGens = " + num2str(trialPtr->nTrialGenerators(),2));
      }

      // Generate new trial branchings, starting from qMax
      double qBegin = qMax;

      // All evolution normalizations are limited by invariant mass of dipole
      double sAnt    = trialPtr->s();
      double mAnt    = trialPtr->m();
      qBegin = min(qBegin,mAnt);

      // Lowest evolution boundary: impose hadronization scale.
      double qEnd = qEndAll;

      // Translate the cutoff scale to a lower bound on the evolution variable
      double qCutoff = cutoffScale;
      if (cutoffType == 0 || cutoffType == evolutionType) {
        qCutoff = sqrt(evNormFac) * cutoffScale;
      } 
      // mD evolution with a cutoff in pT: larger mD hull
      else if (cutoffType == 1 && evolutionType == 2) {
        if (4*pow2(cutoffScale) > sAnt) qCutoff = qBegin;
        else
          qCutoff = sqrt( evNormFac * sAnt 
            * (1.0 - sqrt(1.0 - 4.0*pow2(cutoffScale)/sAnt)));
      } 
      // pT evolution with a cutoff in mD: larger pT hull
      else if (cutoffType == 2 && evolutionType == 1) {
        qCutoff = sqrt( evNormFac * pow(cutoffScale,4.) / sAnt);
      }
      // Cutoff for g->qq splittings
      double qCutoffSplit = 0.0;
      if (cutoffType == 1  || (cutoffType == 0 && evolutionType == 1)) {
        if (4*pow2(cutoffScale) > sAnt) qCutoffSplit = max(mAnt,qBegin);
        else {
          qCutoffSplit = mAnt*sqrt(0.5*(1.0 - 
              sqrt(1.0 - 4.0*pow2(cutoffScale)/sAnt)));
        }
      } else if (cutoffType == 2 || (cutoffType == 0 && evolutionType == 2)) {
        qCutoffSplit = qCutoff;
      }

      qEnd = max(qEndAll, min(qCutoff, qCutoffSplit));

      // No trial generators for this antenna
      if (trialPtr->nTrialGenerators() == 0) continue;

      // Loop over trial functions. Find and save max trial scale.
      double qTrialMax    = 0.0;
      int    indxMax      = -1;

      // Loop over trial generators ----------- BEGIN
      for (int indx = 0; indx < (int)trialPtr->nTrialGenerators(); ++indx) {

        // Pointer to trial generator for this trial:
        TrialGeneratorFSR* trialGenPtr = trialPtr->trialGenPtrSav[indx];
        string trialGenName            = trialGenPtr->name();
        int iAntPhys                   = trialPtr->getPhysIndex(indx);

        double qBeginNow = qBegin;

        // Phase space limit can never be exceeded
        qBeginNow = min(qBeginNow,mAnt);

        // If restart scale < scale we already found continue
        if (qBeginNow < qWin || qBeginNow < qTrialMax) continue;

        // Determine what the hadronization cutoff is for this trial
        double qCutoffNow = qCutoff;
        if (iAntPhys == iQGsplit || iAntPhys == iGGsplit)
          qCutoffNow = qCutoffSplit;

        if (qEndAll > qCutoffNow) qEndsmallerCutoff = false;

        // Check if any phase space still open
        if ( (qBeginNow <= qCutoffNow) || (qBeginNow <= qEnd) ) {
          // Verbose output
          if (verbose >= 9)
            printOut("TimeAntShower::pTnext",
              "skipping this trial since qBeginNow = " + num2str(qBeginNow) +
              " qCutoffNow = " + num2str(qCutoffNow) +
              " qEnd = " + num2str(qEnd));
          continue;
        }

        double qTrial = qBeginNow;

        bool needNewTrial = true;
        // TMP! Saved trials disabled
        /*
        // Check if we have a saved trial, unless we have external merging
        if (!doMergeSav && trialPtr->hasTrial(indx)) {
          if (trialPtr->getTrialScale(indx) < (qBeginNow+TINY)) {
            qTrial = trialPtr->getTrialScale(indx);   
            // Verbose output
            if (verbose >= 9)
              printOut("TimeAntShower::pTnext",
                "Trial " + num2str(indx) + " " + trialGenName +
                " Retrieving saved qTrial = " + num2str(qTrial));
            needNewTrial = false;
          }
        }
        */

        // Generate new q value / no saved trial ----------- BEGIN
        if (needNewTrial || doMergeSav) {

          // Phase space mass correction for flavours bookkept as massive
          //    Ecm / sqrt( lambda( Ecm^2, m_parent1^2, m_parent2^2 ) )
          // with Lambda the Kallen function, to account for the
          // normalisation to the two-particle phase space volume
          double psMassCorr = 1.0;
	  int id1 = trialPtr->old1.id();
	  int id2 = trialPtr->old2.id();
	  double s1 = (id1 <= nFlavZeroMassKin) ? 0.0 : trialPtr->old1.m2();
	  double s2 = (id2 <= nFlavZeroMassKin) ? 0.0 : trialPtr->old2.m2();
	  // Compute phase-space correction
	  if (s1 > TINY || s2 > TINY) {
	    psMassCorr = sAnt/sqrt(pow2(sAnt - s1 - s2) - 4*s1*s2);
	    if (verbose >= 8 && psMassCorr > 1.0)
	      printOut("TimeAntShower::pTnext",
		       "massive phase space volume factor is " + num2str(psMassCorr) );
	  }

	  // Impose evolution windows
          bool acceptRegion = false;
          int iRegion = getRegion(qTrial);

          // Go through regions ----------- BEGIN
          while (!acceptRegion) {

            // Set overestimated Z range for trial generation
            double qMinNow = getQmin(iRegion);
            double yMinNow = pow2(qMinNow)/sAnt;
            double zMinNow = trialGenPtr->getZmin(yMinNow);
            double zMaxNow = trialGenPtr->getZmax(yMinNow);

            bool gluonSplit = (iAntPhys == iQGsplit || iAntPhys == iGGsplit);

            // Set color factor for trial
            double colFac = getAnt(iAntPhys)->getChargeFac();
            int nF        = getNf(iRegion);
            if (gluonSplit) colFac *= min(nF,nGluonToQuarkF);

            // Set headroom factor
            double headroomFac = getHeadroom( ( gluonSplit ? 1 : 0 ), iRegion,
              nG[iSys], nQQ[iSys], iSys, qMinNow );

            // Normalization factor (headroom * phaseSpaceMassCorrection)
            double normFac = headroomFac * psMassCorr;            

            // Verbose output
            if (verbose >= 9)
              printOut("TimeAntShower::pTnext", "Trial " + num2str(indx)
                + " " + trialGenName + ": qBeginNow = " + num2str(qTrial)
                + " (evolution window " + num2str(iRegion) + ", qMin = "
                + num2str(qMinNow) + ", nF = " + num2str(nF) + ") qCutoffNow = "
                + num2str(qCutoffNow) + " qEnd = " + num2str(qEnd));

            // Effective renormalization-scale prefactor. 
            double kR = sqrt( gluonSplit ? aSkMu2Split : aSkMu2 );

            // Check if we can use a running alphaS
            bool runAlpha = false;
            if (alphaSorder >= 1 && evolutionType == 1) {
              runAlpha = true;
              // If we are close to lambdaQCD, use constant instead
              if (qMinNow < 2.5*alphaSptr->Lambda3()/kR) runAlpha = false;
            }

	    // Biasing
	    bool   doBias  = false;
	    double biasFac = 1.0;
	    if ( qTrial > biasCutoff ) {
	      if (hardSys[iSys] && biasInHard) doBias = true;
	      else if (isResonance[iSys] && biasInResDec) doBias = true;
	      else if (!hardSys[iSys] && !isResonance[iSys] &&
		partonSystemsPtr->hasInAB(iSys) && biasInMPI) doBias = true;
	      if (doBias) {
		biasFac *= biasAll;
		// At the trial level, all gluon splittings enhanced by
		// max(biasCharm, biasBottom)
		if (gluonSplit && min(nF,nGluonToQuarkF) >= 4)
		  biasFac *= max(biasCharm,biasBottom);
	      }
	    }

            // 0) Sanity check for zero branching probability
            if (colFac < TINY || headroomFac < TINY) {
              // Save old scale
              double qTmp = qTrial;
              qTrial = 0.0;              
              trialPtr->saveTrial(indx,qTmp,qTrial,0.,0.,0.,0.);
            }

	    // 1) AlphaS running inside trial integral
            // Only used for pT-evolution with muR propto pT 
            // and for mQQ-evolution with muR propto mQQ
            else if (runAlpha) { 
              // Save old scale
              double qTmp = qTrial;
              // One-loop beta function (two-loop imposed by veto, below)
              double b0 = (33.0 - 2.0*nF) / (12.0 * M_PI);
              // Use 3-flavour alphaS for overestimate
              double lambdaEff = alphaSptr->Lambda3();
              // Generate new q value, with alphaS running inside trial integral
              double q2trial = trialGenPtr->genQ2run(pow2(qTrial),sAnt,
		zMinNow,zMaxNow,colFac,b0,kR,lambdaEff,normFac,biasFac);
              qTrial = sqrt(q2trial);
              // Save trial information
              double evNormNow = evNormFac;
              if (iAntPhys == iQGsplit || iAntPhys == iGGsplit) evNormNow = 1.0;
              double muEff    = kR*qTrial/sqrt(evNormNow)/lambdaEff;
              double alphaEff = 1./b0/log(pow2(muEff));
              trialPtr->saveTrial(indx,qTmp,qTrial,zMinNow,zMaxNow,colFac,
		alphaEff,headroomFac,biasFac);
            }

            // 2) AlphaS outside trial integral
            else {
              // Save old scale
              double qTmp = qTrial;

              // 0) constant alphaS
              double facAlphaS = alphaSvalue;

              // 1) running with muR not proportional to evolution variable
              if (alphaSorder >= 1) {
                facAlphaS = alphaSmax;
                // Special case for mD evolution with muR proportional to pT
                if (evolutionType == 2) {
                  // Lowest pT scale that can be reached in this mD window
                  double pT2min = pow(qMinNow,4.)/sAnt/pow2(evNormFac);
                  double mu2 = max(mu2min,pow2(kR)*pT2min);
                  // alphaS for overestimate
                  facAlphaS = alphaSptr->alphaS(mu2);
                  facAlphaS = min(alphaSmax,facAlphaS);
                  // Special for 2-loop running below 1 GeV: extra headroom
                  if (alphaSorder >= 2 && mu2 < 1.) facAlphaS *= 1.1;
                }
              }

              // Generate new q value, with alphaS outside trial integral
              double q2trial = trialGenPtr->genQ2(pow2(qTrial),
		sAnt,zMinNow,zMaxNow,colFac,facAlphaS,normFac,biasFac); 
              qTrial = sqrt(q2trial);
              // Save trial information
              trialPtr->saveTrial(indx,qTmp,qTrial,zMinNow,zMaxNow,colFac,
		facAlphaS,headroomFac,biasFac);
            } 

            // Check evolution window boundaries 
            if (qTrial > qMinNow) acceptRegion = true;
            else if (qMinNow < qWin || qMinNow < qTrialMax) {
              if (verbose >= 8)
                printOut("TimeAntShower::pTnext",
                  "stopping evolution, already found scale that is bigger");
              acceptRegion = true;
              trialPtr->renewTrial(indx);
              // TMP! need to set restart scales properly once saved trials
              // will be used!
              qTrial = 0.0;
            } else if (iRegion == 0) { 
              acceptRegion = true;
              qTrial       = 0.0;
            } else { 
              qTrial = qMinNow;
              iRegion--;
            }

          } 
          // Go through regions ----------- END

        }
        // Generate new q value / no saved trial ----------- END

        if (qTrial > qTrialMax) {
          qTrialMax = qTrial;
          indxMax   = indx;
        }

      }
      // Loop over trial generators ----------- END

      // Did I win?
      if (qTrialMax >= qWin || qWin <= 0.0) {      
        winnerPtr          = trialPtr;
        qWin               = qTrialMax;
        iSysWin            = iSys;
      } 

      // Verbose output
      if (verbose >= 7) {
        printOut("TimeAntShower::pTnext", "qTrialMax = " + num2str(qTrialMax)
          + (indxMax >= 0 ? " ("+trialPtr->trialGenPtrSav[indxMax]->name()+")"
          : "") + (iAnt+1 == nrAnts ? "; final qWin = " : "; current qWin = ")
          + num2str(qWin) + " i1 i2 = " + num2str(winnerPtr->iOld1) + " "
          + num2str(winnerPtr->iOld2) + " in system " + num2str(iSysWin,4));
      }
      
    }
    // Loop over antennae ----------- END

  }
  // Loop over the systems we have ----------- END

  // If non-zero branching scale found: continue
  if (qWin > qEndAll) {
    // Verbose output  
    if (verbose >= 5) {
      int indxWin = winnerPtr->getTrialIndex();
      if (verbose >= 7)  list();
      printOut("TimeAntShower::pTnext", "=== Winner at scale qWin = "
        + num2str(qWin)
        + " trial type " + winnerPtr->trialGenPtrSav[indxWin]->name()
        + " col = " + num2str(winnerPtr->old1.col(),5)
        + " in System " + num2str(iSysWin,4));
    }
  }

  else {
    qWin = 0.0;
    // Check if qEnd < all cutoffs
    if (qEndsmallerCutoff && !zeroAnts) {
      // Verbose output
      if (verbose >= 5) {
        printOut("TimeAntShower::pTnext","=== All trials now below cutoff "
          "qEndAll = "+num2str(qEndAll,3)+".");
        if (verbose >= 6) {
          printOut("TimeAntShower::pTnext","Final configuration was:");
          event.list();
        }
      }

      // FSR is done (at least for this system), set the weights
      weightsPtr->doWeighting();

      // If we have not yet binned a branching scale, bin 0 as the first
      // branching scale
      if (verbose >= 2 && !firstQBranchBinned) {
        if (vinciaHistos.find("1stBranchingQE/eCM") != vinciaHistos.end())
          vinciaHistos["1stBranchingQE/eCM"].fill(0);
        firstQBranchBinned = true;
      }

      // Optionally, before passing the event back to Pythia, apply
      // non-perturbative colour reconnections
      if (CRmode >= 1) {
        // Find branchelementals of hard system
        int iSysHard = -1;
        for (int i=0; i<(int)hardSys.size(); i++)
          if (hardSys[i]) iSysHard = i;
        if (iSysHard > -1) {
          // First try junction-string reconnections, if switched on
          if (CRjunctions) {
            int nReconnectionsJun 
              = reconnectionsPtr->doJunctions(event, branchelementals[iSysHard]);
            if (verbose >= 4 && nReconnectionsJun >= 1) 
              printOut("pTnext()","Junction(s) created before HAD");
          }
          int measureType    = 1;
          int nReconnections = 
            reconnectionsPtr->doReconnect(event, branchelementals[iSysHard],
              0.0, measureType);
          if (verbose >= 3 && nReconnections >= 1) 
            printOut("pTnext()","Antenna reconnection(s) created before HAD");
        }
      }
    }
  }
    
  // End of pTnext: finalize and return to PYTHIA

  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::pTnext","finished --------------");

  // Done! 
  return qWin;

}
  
//*********
  
// Perform a branching
  
bool TimeAntShower::branch(Event& event, bool isInterleaved) {
    
  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::branch","begin --------------");
       
  // The winning branchelemental is stored in winnerPtr
  // Extract current QE, trial index, etc.
  int iTrial           = winnerPtr->getTrialIndex();
  double qNew          = winnerPtr->scaleSav[iTrial];
  // Sanity check
  if (qNew <= 0.0) return false;
  int iAntPhys         = winnerPtr->getPhysIndex(iTrial);
  bool isSwapped       = winnerPtr->getIsSwapped(iTrial);
  iSysSel              = winnerPtr->system;
  // Is this a saved (postponed) branching?
  bool isPostponed     = winnerPtr->isPostponed;
  if (isPostponed) cout<<" isPostponed ==============================="<<endl;
  // Verbose output
  if (verbose >= 5)
    printOut("TimeAntShower::branch","* Processing Branching at scale Q = " 
      +num2str(qNew)+"   isInterleaved = "+bool2str(isInterleaved));
    
  //------------------------------------------------------------------------
  // GENERATE FULL TRIAL KINEMATICS (AND REJECT IF OUTSIDE PHASE SPACE)
  // (if not just restoring a saved branching)
  if ( !isPostponed ) {

    // Count up global number of attempted trials
    ++nTrialsSum;

    // Generate kinematics
    if ( !generateKinematics(winnerPtr)) {
      // Mark this trial as "used", will need to generate a new one.
      winnerPtr->renewTrial(iTrial);
      return false;     
    }
  }
  else {
    // Saved branching: mark as used
    winnerPtr->scaleSav[iTrial] = -1.0;
  }

  //------------------------------------------------------------------------
  // ASSIGN COLOUR FLOW
  // NOTE: Using lastColTag here, if branching accepted we tell Pythia
  // NOTE: redone for saved branchings; could save some time by saving it
  //       Also relies on nothing physically changing so all scales are the same
  //       If a stochastic choice of LC colour flow is made, then we must save and
  //       reuse the original choice. May need to just repaint with updated colour
  //       line numbers in case the previous ones were already used.
  bool usedColTag = assignColourFlow(event, winnerPtr);
    
  //------------------------------------------------------------------------
  // UPDATE LIST OF PARTONS
  vector<Particle> partsNow = partsSav[iSysSel];
  vector<int> antPos; 
  antPos.resize(3);
  // Replace the old ones...
  for (int i = 0; i < (int)indexSav[iSysSel].size(); ++i) { 
    if (indexSav[iSysSel][i] == winnerPtr->iOld1) {
      partsNow[i] = winnerPtr->new1;
      antPos[0]   = i;
    }
    else if (indexSav[iSysSel][i] == winnerPtr->iOld2) {
      partsNow[i] = winnerPtr->new3;
      antPos[1]   = i;
    }
  }
  // ... and add the new one.
  partsNow.push_back(winnerPtr->new2);
  antPos[2] = (int)partsNow.size()-1;
  
  //------------------------------------------------------------------------
  // PHASE SPACE CLOSED FOR GETTING RID OF HEAVY QUARKS?
  if (!ISRshowerPtr->checkHeavyQuarkPhaseSpace(partsNow,iSysSel)) {
    if (verbose >= 6)
      printOut("TimeAntShower::branch", "branching rejected because phase "
        "space after branching does not allow forced splittings");
    // Mark this trial as "used", will need to generate a new one.
    winnerPtr->renewTrial(iTrial);
    ++nClosePSforHQ;
    return false;
  }
    
  //------------------------------------------------------------------------
  // VETO STEP: DECIDE WETHER TO ACCEPT OR REJECT BRANCHING
  if (not acceptTrial(event, winnerPtr, iSysSel, partsNow, antPos)) {
    // Mark this trial as "used", will need to generate a new one.
    winnerPtr->renewTrial(iTrial);
    return false;
  }
    
  // Flag to tell whether a general update of antenna parents is necessary
  bool updateMothers = false;
    
  //------------------------------------------------------------------------
  // PUT NEW PARTICLES INTO EVENT RECORD
  // Store a copy of event, to be used if branching vetoed by userHooks
  Event evtOld   = event;
  int evtSizeOld = evtOld.size();
  int iOld1      = winnerPtr->iOld1;
  int iOld2      = winnerPtr->iOld2;
  winnerPtr->new1.scale(qNew);
  winnerPtr->new2.scale(qNew);
  winnerPtr->new3.scale(qNew);
  int iNew1      = event.append(winnerPtr->new1);
  int iNew2      = event.append(winnerPtr->new2);
  int iNew3      = event.append(winnerPtr->new3);
  // Update event pointers if necessary
  event.restorePtrs();
  // Check if we went from polarized to unpolarized state
  // If so, depolarize parton state
  // (A more complete alternative here would be to create depolarized 
  // copies of all partons and then update everything, but deemed
  // unnecessary for now.)
  if ( event[iOld1].pol() != 9 && event[iOld2].pol() != 9 && 
    (event[iNew1].pol() == 9 || event[iNew2].pol() == 9 ||
      event[iNew3].pol() == 9)) {
    if (verbose >= 5) 
      printOut("TimeAntShower::branch","Depolarizing parton state");
    // Depolarize parton state (except the branching mothers, 
    // which will be replaced by unpolarized daughters) 
    int sizeSystem = partonSystemsPtr->sizeOut(iSysSel);
    for (int i = 0; i < sizeSystem; ++i) {    
      int i1 = partonSystemsPtr->getOut( iSysSel, i);
      // Skip if not present in final state
      if ( i1 <= 0 || !event[i1].isFinal()) continue;
      // Skip if mother parton to be replaced by daughter
      if ( i1 == winnerPtr->iOld1 || i1 == winnerPtr->iOld2) continue;
      // Else depolarize
      if ( event[i1].pol() != 9 ) event[i1].pol(9);
      // Tell that antenna mothers should be updated in update loop below
      updateMothers = true;
    }
    // Also make sure all daughters are depolarized as well
    winnerPtr->new1.pol(9);
    winnerPtr->new2.pol(9);
    winnerPtr->new3.pol(9);
  }
    
  //------------------------------------------------------------------------
  // SET MOTHERS AND DAUGHTERS
  // Mark original dipole partons as branched
  // iOld1=I -> iNew1=i
  //         -> iNew2=j
  // iOld2=K -> iNew3=k
  event[iOld1].statusNeg();
  event[iOld2].statusNeg();
  // Gluon emission
  if (iAntPhys == iQQemit || iAntPhys == iQGemit || iAntPhys == iGGemit) {
    event[iNew1].mothers(iOld1,0);
    event[iNew2].mothers(iOld1,iOld2);
    event[iNew3].mothers(iOld2,0);
    event[iOld1].daughters(iNew2,iNew1); 
    event[iOld2].daughters(iNew3,iNew2);
  }
  // Gluon splitting in old1
  else if (isSwapped) {
    event[iNew1].mothers(iOld1,0);
    event[iNew2].mothers(iOld1,0);
    event[iNew3].mothers(iOld2,iOld2);
    event[iOld1].daughters(iNew2,iNew1); 
    event[iOld2].daughters(iNew3,iNew3);
  }
  // Gluon splitting in old2
  else {
    event[iNew1].mothers(iOld1,iOld1);
    event[iNew2].mothers(iOld2,0);
    event[iNew3].mothers(iOld2,0);
    event[iOld1].daughters(iNew1,iNew1); 
    event[iOld2].daughters(iNew3,iNew2);
  }
  event[iNew1].daughters(0,0); 
  event[iNew2].daughters(0,0);
  event[iNew3].daughters(0,0);
    
  //------------------------------------------------------------------------
  // VETO BY USERHOOKS
  // Possibility to allow user veto of emission step.
  if (canVetoEmission)
    if (userHooksPtr->doVetoFSREmission(evtSizeOld, event, iSysSel)) {
      // Restore old event
      event = evtOld;
      return false;
    }

  //------------------------------------------------------------------------
  // EVERYTHING ACCEPTED

  // If diagnostic histograms are on, write in the branching scale
  if (verbose >= 2 && !firstQBranchBinned) {
    if (vinciaHistos.find("1stBranchingQE/eCM") != vinciaHistos.end())
      vinciaHistos["1stBranchingQE/eCM"].fill( qNew/event[0].e() );
    firstQBranchBinned = true;
  }
    
  // Update the list of systems.
  partonSystemsPtr->replace(iSysSel, iOld1, iNew1);  
  partonSystemsPtr->addOut(iSysSel, iNew2);
  partonSystemsPtr->replace(iSysSel, iOld2, iNew3); 
    
  // Update number of gluons and quark pairs
  if (winnerPtr->new2.id() == 21) ++nG[iSysSel];
  else {
    nG[iSysSel] -= 1;
    ++nQQ[iSysSel];
  }
    
  //------------------------------------------------------------------------
  // UPDATE LIST OF PARTONS AND PARTON NUMBERS
  // Save the first particle (might be ancestor)
  int iAncestor     = indexSav[iSysSel][0];
  Particle ancestor = partsSav[iSysSel][0];
  indexSav[iSysSel].clear(); indexSav[iSysSel].resize(0);
  partsSav[iSysSel].clear(); partsSav[iSysSel].resize(0);
  if (winnerPtr->iAncestor > 0) {
    indexSav[iSysSel].push_back( iAncestor );
    partsSav[iSysSel].push_back( ancestor );
  }
  int sizeSystem = partonSystemsPtr->sizeAll(iSysSel);
  for (int i = 0; i < sizeSystem; ++i) {    
    int i1 = partonSystemsPtr->getAll( iSysSel, i);
    if ( i1 <= 0 ) continue;
    indexSav[iSysSel].push_back( i1 );
    partsSav[iSysSel].push_back( event[i1] );
  }
  if (verbose >= 6) {
    printOut("TimeAntShower::branch", "updated system, now we have "
      "the following particles: nr id col acol p");
    for (int i=0; i<(int)partsSav[iSysSel].size(); i++) {
      cout << "  " << setw(3) << indexSav[iSysSel][i];
      cout << " " << setw(2) << partsSav[iSysSel][i].id();
      cout << " " << setw(3) << partsSav[iSysSel][i].col();
      cout << " " << setw(3) << partsSav[iSysSel][i].acol();
      cout << " " << partsSav[iSysSel][i].p();
    }
  }
    
  // Communicate changed lists to ISR shower
  ISRshowerPtr->partsSav = partsSav;
  ISRshowerPtr->indexSav = indexSav;

  // Update old and add new dipoles
  if (verbose >= 5)
    printOut("TimeAntShower::branch",
      "=== Branching Accepted. Updating dipole-antenna(e)");
    
  //************************************************************************
  // Updates of the branched antennae
  //************************************************************************

  // If we are restoring a saved (postponed) branching from the list of
  // savedBranchings, mark it for removal from the saved list, and replace
  // in the main list of antennae
  if (isPostponed) {
    winnerPtr->scaleSav[iTrial] = -1.0;
    // Find the (first) element in the main list which matches
    for (int iAnt=0; iAnt<(int)branchelementals[iSysSel].size(); iAnt++) {
      BranchelementalFSR* antPtr = &branchelementals[iSysSel][iAnt];
      if (antPtr->iOld1 == iOld1 && antPtr->iOld2 == iOld2) {
        branchelementals[iSysSel][iAnt] = *winnerPtr;
        winnerPtr = &branchelementals[iSysSel][iAnt];
        break;
      }
    }
  }
    
  // Store old neighbours
  int iL  = winnerPtr->iL;
  int iR  = winnerPtr->iR;
    
  if (winnerPtr->new1.colType() > 0) {        
    // Normal case, where iNew1-iNew2 has an LC color connection
    winnerPtr->reset(event,iNew1,iNew2);
    winnerPtr->old1 = event[iNew1];
    winnerPtr->old2 = event[iNew2];    

    // Update trial generators
    winnerPtr->resetTrialGenerators(trialGenPtrs, false);
      
    // Update neighbours
    winnerPtr->iL = iL;
    winnerPtr->iR = iNew3;
      
    // If second parton is octet, insert new antenna (iNew2,iNew3)
    if (event[iNew2].colType() > 0) {
      // Define new dipole-antenna
      BranchelementalFSR trial(event,iNew2,iNew3);
      trial.system = iSysSel;
      trial.old1   = event[iNew2];
      trial.old2   = event[iNew3];
        
      // Set trial generators
      trial.resetTrialGenerators(trialGenPtrs, false);
        
      // Set neighbors
      trial.iL     = iNew1;
      trial.iR     = iR;
        
      // Keep family name for matching
      trial.iAncestor = winnerPtr->iAncestor;
        
      // Add dipole-antenna to list, insert after iNew1-Inew2 to keep order
      for (int iAnt=0; iAnt<int(branchelementals[iSysSel].size()); iAnt++) {          
        if (&branchelementals[iSysSel][iAnt] == winnerPtr) {
          branchelementals[iSysSel].insert(branchelementals[iSysSel].begin()+iAnt+1,trial); 
          break;
        }
      }
    }
  } else {
    // Special case for qbar q-x final state, where only second pair is dipole
    winnerPtr->reset(event,iNew2,iNew3);
    winnerPtr->old1 = event[iNew2];
    winnerPtr->old2 = event[iNew3];
      
    // Update trial generators
    winnerPtr->resetTrialGenerators(trialGenPtrs, false);
      
    // Update neighbors
    winnerPtr->iL = iNew1;
    winnerPtr->iR = iR;
  }
    
  //************************************************************************
  // Updates of the other antennae
  //************************************************************************
    
  //------------------------------------------------------------------------
  // Updates
  for (int iAnt = 0; iAnt < (int)branchelementals[iSysSel].size(); ++iAnt) {
    BranchelementalFSR* antPtr = &branchelementals[iSysSel][iAnt];
      
    // Update mother partons in case a depolarization took place
    if (updateMothers) {
      antPtr->old1 = event[antPtr->iOld1];
      antPtr->old2 = event[antPtr->iOld2];
    }
      
    // The post-branching branchelementals
    if (antPtr->iOld1 == iNew1 || antPtr->iOld1 == iNew2) {
      // No update necessary for this pair (handled above)
    }
      
    else {
      // Neighbour update 
      if (antPtr->iL == iOld2) antPtr->iL = iNew3;
      if (antPtr->iR == iOld1) antPtr->iR = iNew1;
        
      if (antPtr->iOld1 == iOld2) {
        antPtr->iOld1 = iNew3;
        antPtr->old1  = event[iNew3];
        // Update neighbors
        antPtr->iL    = iNew2;
        antPtr->renewTrial();
        // Update trial generators
        antPtr->resetTrialGenerators(trialGenPtrs, false);
      }
        
      if (antPtr->iOld2 == iOld1) {
        antPtr->iOld2 = iNew1;
        antPtr->old2  = event[iNew1];
        // Update neighbors
        antPtr->iR    = iNew2;
        antPtr->renewTrial();
        // Update trial generators
        antPtr->resetTrialGenerators(trialGenPtrs, false);
      }
    }
      
    // Update saved antenna mass
    antPtr->mCalc();
      
  }
    
  if (verbose >= 5) {
    event.list();
    list();
    if (verbose >= 7) {
      cout<<" Masses (mCalc): "<<scientific
          <<event[iNew1].m()<<" ("<<event[iNew1].mCalc()<<")  "
          <<event[iNew2].m()<<" ("<<event[iNew2].mCalc()<<")  "
          <<event[iNew3].m()<<" ("<<event[iNew3].mCalc()<<")  "<<endl;
    }
  }
    
  // Count the number of branchings in the system
  nBranch[iSysSel]++;
  nBranchFSR[iSysSel]++;
      
  // Let Pythia know how many color tags we used
  if (usedColTag) {
    event.nextColTag();
    if (event[iNew2].id() == 21) {
      int lastTag = event.lastColTag();
      int colMax  = max(event[iNew2].col(),event[iNew2].acol());
      while (colMax > lastTag) lastTag = event.nextColTag();
    }
  }
    
  // Check the event after each branching
  vinComPtr->showerChecks(event, false);
    
  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::branch","finished --------------");
    
  // Statistics for first 10 branchings in first 4 systems
  if (iSysSel < 4 && nBranchFSR[iSysSel] < 11) {
    qBranch[iSysSel][nBranchFSR[iSysSel]]=qNew;
    pTphys[iSysSel][nBranchFSR[iSysSel]]=event[iNew2].pT();
  }

  // Done. 
  return true;
    
}
  
//*********
  
// Update list of branchelementals after MPI rescattering 
  
void TimeAntShower::rescatterUpdate( int iSys, Event& event) {
  if (verbose >= 1)
    printErr("TimeAntShower::rescatterUpdate","system "+num2str(iSys)
      +" (function not implemented)");
  if (verbose >= 5) event.list();
}
  
  
//**********************************************************************
  
// END OF INHERITED FUNCTIONS FOR INTERFACE TO PYTHIA8
// BEGIN VINCIA INTERNAL FUNCTIONS
  
//**********************************************************************
  
//*********
  
// Update antennae if ISR changed system in prepare
  
void TimeAntShower::updateAfterISR(int iSys, Event& event, map<int,int> iChanged) {
    
  // Don't need to do anything
  if ((int)branchelementals.size() <= iSys ) return;
  if (branchelementals[iSys].size() <= 0) return;
    
  // Update the particles in my list
  for (int i=0; i<(int)indexSav[iSys].size(); i++) {
    int iOld = indexSav[iSys][i];
    indexSav[iSys][i] = iChanged[iOld];
    partsSav[iSys][i] = event[iChanged[iOld]];
  }
    
  // Loop over the antennae to reset antenna partons and neighbours
  for (int i=0; i<(int)branchelementals[iSys].size(); i++) {
    BranchelementalFSR* antPtr = &branchelementals[iSys][i];
    int iOld1 = antPtr->iOld1;
    int iOld2 = antPtr->iOld2;
    antPtr->iOld1 = iChanged[iOld1];
    antPtr->iOld2 = iChanged[iOld2];
    antPtr->old1  = event[iChanged[iOld1]];
    antPtr->old2  = event[iChanged[iOld2]];
    // Recalculate the mass
    antPtr->mCalc();
    // Trialgenerators are still the same
    // Neighbours
    if (antPtr->iL != 0) antPtr->iL = iChanged[antPtr->iL];
    if (antPtr->iR != 0) antPtr->iR = iChanged[antPtr->iR];
    // Update daughter partons
    iOld1 = antPtr->iOld1;
    iOld2 = antPtr->iOld2;
    antPtr->new1 = Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
    antPtr->new2 = Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
    antPtr->new3 = Particle(0,51,iOld1,iOld2,0,0,0,0,0.); 
    antPtr->new1.setEvtPtr(&event);
    antPtr->new2.setEvtPtr(&event);
    antPtr->new3.setEvtPtr(&event);
  }
    
}
  
//*********
  
// Initialize internal and diagnostics histograms
  
void TimeAntShower::initVinciaHistos() {
    
  // Verbose output
  if (verbose >= 7) {
    printOut("TimeAntShower::initVinciaHistos","begin --------------");
  }
    
  vinciaHistos.clear();
    
  // Don't book diagnostics histograms if verbose output not switched on.
  if (verbose <= 1) return;
    
  // Global
  int nBin;
  double binWid, wid, c, xMin, xMax;
    
  // MATCHING WEIGHTS
  nBin   = 61;
  binWid = 0.1;
  wid    = nBin * binWid;
  c      = 0.0;
  xMin   = c - wid/2.;
  xMax   = c + wid/2.;
  Hist HPME2LogUSR("Matching weight",nBin,xMin,xMax);
    
  // Histos for Paccept : gluon emission
  for (int nGtmp = 0; nGtmp <= 9; ++nGtmp) {
    for (int nQQtmp = 0; nQQtmp <= 5; ++nQQtmp) {
      string baseState = num2str(2*nQQtmp,1)+"q"+num2str(nGtmp,1)+"g";
      for (int iBranchType = 1; iBranchType <= 3; ++iBranchType) {
        string state = baseState;
        if (iBranchType == 3) state +="SplitB";
        else if (iBranchType == 2) state +="SplitA";
        else state +="Emit";
        string hisTitle="Log10(ME/AntPhys):" + state;
        vinciaHistos[hisTitle]=HPME2LogUSR;
        hisTitle="Log10(ME/AntTrial):" + state;
        vinciaHistos[hisTitle]=HPME2LogUSR;
      }
    }
  }

  // Limit highest multiplicity histograms by matching or verbose
  int nMultHi = 2 + matchingLO;
  if (verbose >= 7) nMultHi = max(nMultHi,10);
  if (verbose >= 8) nMultHi = max(nMultHi,100); 

  // Histos for pTevol scales and alphaS
  nBin   = 100;
  binWid = 0.1;
  wid    = nBin * binWid;
  xMax   = 0.;
  xMin   = xMax - wid;
  Hist HqLn("Ln(q2/sSystem) distribution (USR)",nBin,xMin,xMax);
  Hist HalphaS("alphaS distribution (USR)", nBin, 0., 2.);
  Hist HalphaSratio("alphaS/alphaTrial distribution (USR)", nBin, 0., 2.);
  for (int nGtmp = 0; nGtmp <= nMultHi; ++nGtmp) {
    for (int nQQtmp = 0; 2*nQQtmp + nGtmp <= nMultHi; ++nQQtmp) {
      string baseState = num2str(2*nQQtmp,1)+"q"+num2str(nGtmp,1)+"g";
      for (int iBranchType = 1; iBranchType <= 3; ++iBranchType) {
        string state = baseState;
        if (iBranchType == 3) state +="SplitB";
        else if (iBranchType == 2) state +="SplitA";
        else state +="Emit";
        string hisTitle="Ln(q2trial/sSystem):" + state;
        vinciaHistos[hisTitle]       = HqLn;
        hisTitle="Ln(q2/sSystem):" + state;
        vinciaHistos[hisTitle]       = HqLn;
        string hisTitleAlphaS = "alphaS:" + state;
        vinciaHistos[hisTitleAlphaS] = HalphaS;
        string hisTitleAlphaSratio = "alphaSratio:" + state;
        vinciaHistos[hisTitleAlphaSratio] = HalphaSratio;
      }
    }
  }
    
  // Histogram of QE/eCMHad for the first branching
  Hist hQ2E("QE/eCM",1000,0.0,1.0);
  vinciaHistos["1stBranchingQE/eCM"] = hQ2E;
}
    
// ******
// 
// generateKinematics(): generates the post-branching momenta 
//               assumes branching scale trialPtr->qNew already generated
//     Checks that require knowing only the local (2->3) branching 
//     kinematics (without referring to the full event) belong here.
// Steps:
//     *) Generate second branching invariant, given qNew
//     *) Set flavours and masses
//     *) Check local hadronization veto inside trial antenna
//        and construct kinematics in LAB frame
//        Construct new partons
// Return true if physical kinematics could be generated, otherwise false

bool TimeAntShower::generateKinematics(BranchelementalFSR* trialPtr) {
    
  // Verbose output
  if (verbose >= 7)
    printOut("TimeAntShower::generateKinematics","begin --------------");

  //************************************************************************
  // Basic info about trial function and scale
  //************************************************************************

  int iTrial     = trialPtr->getTrialIndex();
  if (iTrial < 0) return false;
  int iAntPhys   = trialPtr->getPhysIndex(iTrial);
  bool isSwapped = trialPtr->getIsSwapped(iTrial);

  double qNew  = trialPtr->getTrialScale(iTrial);
  double sAnt  = trialPtr->s();

  // Consistency check: antenna mass and branching scale > TINY
  if ( sAnt < TINY) {
    if (verbose >= 3 && sAnt < TINY) {
      printOut("TimeAntShower::generateKinematics",
        "Warning: s ="+num2str(sAnt)+"  Q2 = "+num2str(pow2(qNew)));
    }
    // Count up local and global failures due to hadronization and return
    trialPtr->nHadr++;
    ++nTrialsCutM;
    return false;
  }
    
  //************************************************************************
  // Generate zeta variable, then work out y12, y23
  //************************************************************************
    
  double y12, y23;
  bool pass = trialPtr->genTrialInvariants(y12, y23, iTrial);
  // Check phase-space hull
  if (! pass) {
    if (verbose >= 6) printOut("TimeAntShower::generateKinematics","failed "
      "genTrialInvariants, returning.");
    // Count up local and global failure rate due to hull and return false
    trialPtr->nHull++;
    ++nTrialsHull;
    return false;
  }

  //************************************************************************
  // Set flavors and masses
  //************************************************************************

  // Gluon emission: inherit parent flavors and add middle gluon 
  int splitFlavour = 21;
  if (iAntPhys == iQQemit || iAntPhys == iQGemit || iAntPhys == iGGemit) {
    trialPtr->new1.id(trialPtr->old1.id());
    trialPtr->new2.id(21);
    trialPtr->new3.id(trialPtr->old2.id());
    trialPtr->new1.m(trialPtr->old1.m());
    trialPtr->new2.m(0.0);
    trialPtr->new3.m(trialPtr->old2.m());
  }

  // Gluon splitting:
  else if ( iAntPhys == iQGsplit || iAntPhys == iGGsplit) {

    // Set flavor of splitting
    double nF = trialPtr->getColFac(iTrial);
    splitFlavour = int(rndmPtr->flat() * nF) + 1;

    // QG->QQbar'Q' and GG->GQbarQ
    if ( !isSwapped ) {
      trialPtr->new1.id(trialPtr->old1.id());
      trialPtr->new2.id(-splitFlavour);
      trialPtr->new3.id( splitFlavour);
      trialPtr->new1.m(trialPtr->old1.m());
      if (splitFlavour <= nFlavZeroMassKin) {
        trialPtr->new2.m(0.);
        trialPtr->new3.m(0.);
      } else {
        trialPtr->new2.m(trialPtr->new2.m0());
        trialPtr->new3.m(trialPtr->new3.m0());
      }
    } 
    // GQbar->Qbar'Q'Qbar and GG->QbarQG
    else {
      trialPtr->new1.id(-splitFlavour);
      trialPtr->new2.id( splitFlavour);
      trialPtr->new3.id(trialPtr->old2.id());
      if (splitFlavour <= nFlavZeroMassKin) {
        trialPtr->new1.m(0.);
        trialPtr->new2.m(0.);
      } else {
        trialPtr->new1.m(trialPtr->new1.m0());
        trialPtr->new2.m(trialPtr->new2.m0());
      }
      trialPtr->new3.m(trialPtr->old2.m());
    } 
      
  }
    
  // Check phase-space boundaries
  bool insidePS = true;
  if (y12 + y23 > 1.) insidePS = false;

  // Simple phase-space checks
  // Use kinematic masses to determine energies of the partons that
  // will actually be created in the event record.
  // Assume yij are dot products here, normalised by antenna mass.
  // (For partons bookkept as massless, m1 and m3 are zero here.)
  double m1 = trialPtr->new1.m();
  double m2 = trialPtr->new2.m();
  double m3 = trialPtr->new3.m();
  double mAnt = sqrt(sAnt);
  double e1chk = sAnt*(1-y23+pow2(m1))/(2.*mAnt);
  double e3chk = sAnt*(1-y12+pow2(m3))/(2.*mAnt);
  double e2chk = mAnt - e1chk - e3chk;
  double m2ij = y12 * sAnt + pow2(m1) + pow2(m2);
  double m2jk = y23 * sAnt + pow2(m2) + pow2(m3);
  double m2ik = (1-y12-y23) * sAnt - pow2(m2);
  // Check if E > m would be satisfied if using real masses
  if ( m1 < TINY && trialPtr->new1.idAbs() > nFlavZeroMassDyn)
    m1 = trialPtr->new1.m0();
  if ( m2 < TINY && trialPtr->new2.idAbs() > nFlavZeroMassDyn)
    m2 = trialPtr->new2.m0();
  if ( m3 < TINY && trialPtr->new3.idAbs() > nFlavZeroMassDyn)
    m3 = trialPtr->new3.m0();
  if (e1chk < m1 || e2chk < m2 || e3chk < m3) insidePS = false;
  // Check invariant masses above thresholds if using real masses
  if (m2ij < pow2(m1+m2) || m2jk < pow2(m2+m3) || m2ik < pow2(m1+m3))
    insidePS = false;

  // Check phase space veto
  if (not insidePS) {
    // Count up local and global failure rate due to hull and return false
    trialPtr->nHull++;
    ++nTrialsHull;
    return false;
  }
    

  //************************************************************************
  // Check local hadronization veto(s) : inside this Branchelemental
  //************************************************************************

  // Require 12 and 23 invariants at least above lowest physical meson mass 
  // (consistent with length > thickness assumption of string model)
  // Note: means we check m(qq) for g->qq splitting as well, but that is
  // presumably not a significant effect.

  // 12
  double mMin12 = vinComPtr->mHadMin(trialPtr->new1.id(), trialPtr->new2.id()); 
  if ( m2ij < pow2(1.01*mMin12) ) {
    if (verbose >= 5) printOut("TimeAntShower::generateKinematics",
      "=== Branching Vetoed. m12 < 1.01*mMes.");
    // Count up local and global failures due to hadronization and return
    trialPtr->nHadr++;
    ++nTrialsCutM;
    return false;
  }
  double mMin32 = vinComPtr->mHadMin(trialPtr->new3.id(), trialPtr->new2.id());
  if ( m2jk < pow2(1.01*mMin32) ) {
    if (verbose >= 5) printOut("TimeAntShower::generateKinematics",
      "=== Branching Vetoed. m23 < 1.01*mMes.");
    // Count up local and global failures due to hadronization and return
    trialPtr->nHadr++;
    ++nTrialsCutM;
    return false;
  }
  
  // Check ordinary hadronization cutoff: Ariadne pT or alternatives 
  double q2forHad = m2ij * m2jk / sAnt;
  // Alternative: Cutoff in Dipole dot product
  if (cutoffType == 2) q2forHad = min(y12,y23) * sAnt;
  // Check cutoffScale
  if (q2forHad < pow2(cutoffScale)) {
    if (verbose >= 9) {
      printOut("TimeAntShower::generateKinematics",
        "qHad = " + num2str(sqrt(q2forHad), 0)
        + " < qCut = " + num2str(cutoffScale, 0) + ", veto.");
    }
    // Count up local and global failures due to hadronization and return
    trialPtr->nHadr++;
    ++nTrialsCutQ;
    return false;
  }
    
  //***********************************************************************
  // Generate full kinematics for this trial branching
  //***********************************************************************
    
  // Define pointer to Global or Sector antenna, depending on shower setting
  // At this point, we only really need the kinematics map from antPtr
  AntennaFSR* antPtr = getAnt(iAntPhys);

  // Generate random (uniform) phi angle
  double phi = 2 * M_PI * rndmPtr->flat();
    
  // Generate branching kinematics, starting from dipole-antenna parents
  vector<Vec4> pOld, pNew;
  pOld.push_back(trialPtr->old1.p());
  pOld.push_back(trialPtr->old2.p());
  int kineMapType  = antPtr->getKineMapType();
  vinCluPtr->map2to3(pNew, pOld, kineMapType, y12, y23, phi);
    
  // Save momenta
  trialPtr->new1.p(pNew[0]);
  trialPtr->new2.p(pNew[1]);
  trialPtr->new3.p(pNew[2]);
    
  // Set default polarizations: unpolarized (may be assigned later)
  trialPtr->new1.pol(9); 
  trialPtr->new2.pol(9); 
  trialPtr->new3.pol(9); 
    
  // Note: colour flow is not assigned here, since this requires 
  // knowledge of which is the next global colour tag available in
  // the event.
    
  // Set mothers
  trialPtr->new1.mothers(trialPtr->iOld1,trialPtr->iOld2);
  trialPtr->new2.mothers(trialPtr->iOld1,trialPtr->iOld2);
  trialPtr->new3.mothers(trialPtr->iOld1,trialPtr->iOld2);
        
  // All clear
  return true;
}

// ******
// 
// acceptTrial : main function to accept or reject a trial branching 
//               after full branching kinematics have been constructed.
//     Checks and corrections that require knowing the full event kinematics
//     belong here.
//     *) check sector ordering (if doing sector showers) 
//     *) Optional: check global hadronization vetos
//     *) Optional: apply beyond-LL (beyond 2->3) improvements 
//     *) Assign helicities 
//     *) Optional: match to fixed-order QCD matrix elements (from MadGraph)
//     *) Optional: post-processing of uncertainty variations
//     *) Accept or Reject Trial Emission
// 

bool TimeAntShower::acceptTrial(Event& event, BranchelementalFSR* trialPtr,
  int iSysBranch, vector<Particle> partsNow, vector<int> antPos) {

  // Verbose output
  if (verbose >= 7)
    printOut("TimeAntShower::acceptTrial","begin --------------");

  // Basic info about trial function and scale
  int iTrial     = trialPtr->getTrialIndex();
  if (iTrial < 0) return false;
  double qNew    = trialPtr->getTrialScale(iTrial);
  double q2new   = pow2(qNew);
  int iAntPhys   = trialPtr->getPhysIndex(iTrial);
  bool isSwapped = trialPtr->getIsSwapped(iTrial);

  //************************************************************************
  // Check global hadronization veto(s) (requiring full event)
  // Note: here only check neighbor invariants. Branching invariants were
  // already checked above. 
  // Note 2: for neighbors, we are only checking invariant mass > pion mass.
  // Could in principle be made more elaborate by checking cutoffScale
  // as well, given sufficient reason to do so. 
  //**********************************************************************

  // Require all neighbor dot products at least above pion mass
  // (consistent with length > thickness assumption of string model)
  double sAnt  = trialPtr->s();
  double m2min = sAnt;
  double m2pi  = pow2(particleDataPtr->m0(111));
  if (trialPtr->iL != 0)
    m2min = m2(event[trialPtr->iL],trialPtr->new1);
  if (trialPtr->iR != 0) 
    m2min = min(m2min,m2(event[trialPtr->iR],trialPtr->new3));
  if (m2min < 1.05*m2pi) {
    if (verbose >= 5) printOut("TimeAntShower::acceptTrial","=== Branching Vetoed."
      " m2neighbor < 1.05*m2pi.");
    trialPtr->nHadr++;
    ++nTrialsCutM;
    return false;
  }

  //**********************************************************************
  // Masses
  //**********************************************************************

  // Dot products
  double s12 = 2.0*dot4(trialPtr->new1,trialPtr->new2);
  double s23 = 2.0*dot4(trialPtr->new2,trialPtr->new3);
  double y12 = s12/sAnt;
  double y23 = s23/sAnt;

  //**********************************************************************
  // Helicity of mother partons 
  //**********************************************************************

  double hA = trialPtr->old1.pol();
  double hB = trialPtr->old2.pol();
  bool isPolarized = polarizedSys[iSysBranch] && (hA != 9 && hB != 9);

  //**********************************************************************
  // Compute spin-summed trial antenna sum for these variables
  // including colour and headroom factors
  //**********************************************************************

  // Sum over individual trial terms
  // Also, check whether all trial generators are phrased in terms
  // of the dot products (which is what we compute above) or whether 
  // eg the gluon-splitting ones are phrased in terms of the real 
  // invariant masses?
  double antTrialSum = 0.0;
  for (int iSum = 0; iSum < (int)trialPtr->nTrialGenerators(); ++iSum) {
    // Only include terms that correspond to the current physical antenna
    if (trialPtr->getPhysIndex(iSum) != iAntPhys ||
      trialPtr->getIsSwapped(iSum) != isSwapped ) continue; 
    // Evaluate trial term
    double antTrial    = trialPtr->trialGenPtrSav[iSum]->aTrial(y12,y23);
    // Get Headroom and Color factors
    double headroomFac = trialPtr->getHeadroomFac(iSum);
    double colFac      = trialPtr->getColFac(iSum);
    // (color factor for splitings is just nF, which is not included in
    // physical antenna normalization nor in matrix elements)
    if (iAntPhys == iQGsplit || iAntPhys == iGGsplit) 
      colFac = getAnt(iAntPhys)->getChargeFac();
    antTrial *= colFac * headroomFac;
    antTrialSum += antTrial;
  }

  //**********************************************************************
  // Biased kernels. 
  // Note: all trial functions for the same iAntPhys must use same biasFac
  //**********************************************************************
  double biasFac = trialPtr->getBiasFac(iTrial);
  // If bias was applied but branching is below biasing cutoff, do
  // accept/reject here with probability trial/bias-trial to get back
  // to unbiased trial probability. (Trials only biased for biasFac > 1.)
  if (biasFac > 1.0 && qNew <= biasCutoff) {
    if ( rndmPtr->flat() > 1./biasFac ) return false;
    biasFac = 1.0;
  }
  
  //**********************************************************************
  // If physical antenna function is mirror of current trial, translate to
  // swapped invariants for antenna-function evaluation
  //**********************************************************************

  double hAant  = hA;
  double hBant  = hB;
  double y12ant = y12;
  double y23ant = y23;
  if ( isSwapped ) {
    hAant  = hB;
    hBant  = hA;
    y12ant = y23;
    y23ant = y12;
  }

  //***********************************************************************
  // AlphaS
  // Impose default choice. Can differ slighly from trial even when running
  // inside trial integral, due to flavor thresholds. Here, alphaS(mu) is
  // returned directly, with the number of flavors active at mu, whereas 
  // the number of flavors in the trial integral is controlled by the value
  // of the trial scale.
  //***********************************************************************

  double alphaTrial = trialPtr->getAlphaTrial(iTrial);
  if (isnan(alphaTrial)) cout<<" alphaTRIAL IS NAN!!!"<<endl;
  if (alphaTrial != alphaTrial) cout<<" alphaTRIAL IS != alphaTRIAL"<<endl;
  // Scale is pT for gluon emission and mQQ for gluon splitting
  // (take into account that postponed branchings have a lower scale)
  double mu2     = min(q2new/evNormFac,s12*s23/sAnt);
  double kMu2Usr = aSkMu2;
  if (trialPtr->new2.colType() == -1) {
    mu2     = s23;
    kMu2Usr = aSkMu2Split;
  } else if (trialPtr->new2.colType() == 1) {
    mu2     = s12;
    kMu2Usr = aSkMu2Split;
  }
  double mu2Usr    = max(mu2min, mu2freeze + mu2*kMu2Usr);
  // alphaS value
  double alphaSusr = min(alphaSmax, alphaSptr->alphaS(mu2Usr));

  // Verbose output
  if (verbose >= 8)
    printOut("TimeAntShower::acceptTrial", "alphaTrial = "
      + num2str(alphaTrial) + ", alphaSusr = " + num2str(alphaSusr));

  //************************************************************************
  // Compute spin-summed physical antennae (spin selection below)
  // (note: antPhys normalised to be dimensionless = a * sAnt
  //************************************************************************

  // Define pointer antennae
  AntennaFSR* antPtr = getAnt(iAntPhys);

  // Compute spin-summed antenna function 
  if (verbose >= 8) 
    printOut("TimeAntShower::acceptTrial","Evaluating antenna function");

  double helSum;   
  // Unpolarized case: ignore parent helicities
  if (!isPolarized) {
    helSum = antPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
      9, 9, 9, 9, 9);
  }
  // Polarized case: use parent helicities
  else {
    helSum = antPtr->antennaFunction(y12ant, y23ant, 0., 0., 0.,
      hAant, hBant, 9, 9, 9);
  }

  // Verbose output
  if (verbose >= 8)
    printOut("TimeAntShower::acceptTrial","helSum  = "+num2str(helSum/sAnt));

  // Store (partial) helicity sum
  double antPhys = helSum;

  //************************************************************************ 
  // Antenna Mass Correction Terms
  //   Controlled by nFlavZeroMassDyn 
  //************************************************************************ 
  
  // Mass corrections to gluon emission (off)
  if ( iAntPhys == iQQemit || iAntPhys == iQGemit ) { 
    /*
    int id1 = trialPtr->new1.idAbs(); 
    double m1 = trialPtr->new1.m0();    
    if (id1 <= nFlavZeroMassDyn) m1 = 0.0; 
    int id3 = trialPtr->new3.idAbs(); 
    double m3 = trialPtr->new3.m0(); 
    if (id3 <= nFlavZeroMassDyn) m3 = 0.0;
    if (m1 > 0. || m3 > 0.) { 
      // Correct generated eikonal singularity structure
      // dot products 
      double s1 = pow2(m1); 
      double s3 = pow2(m3); 
      double d12 = y12 * sAnt;
      double d23 = y23 * sAnt; 
      antPhys -= (2*s1/pow2(d12) + 2*s3/pow2(d23)) * sAnt;
      if (antPhys < 0) {
	if (verbose >= 5) cout<<" PROBLEM : ANTENNA FUNCTION NEGATIVE "<<
			    "AFTER MASS CORRECTION "<<endl;
	++nTrialsCutM;
	return false;
      } 
    } 
    */
  }

  // Mass corrections to gluon splitting (positive) 
  else if ( (iAntPhys == iQGsplit || iAntPhys == iGGsplit)
	    && trialPtr->new2.id() > nFlavZeroMassDyn) { 
    double m2q  = pow2(trialPtr->new2.m0()); 
    double m2qq = y23ant * sAnt + 2*m2q;
    antPhys += m2q / pow2(m2qq) * sAnt;
  } 
  
  //************************************************************************
  // HYPERJET: special mode for processing events from HEJ 
  // => matching to soft-resummed matrix elements
  // Ignore otherwise
  //************************************************************************

  if ( hyperjet && (iAntPhys == iQQemit || iAntPhys == iQGemit
      || iAntPhys == iGGemit) ) {
    // If running in hyper-jet mode, subtract soft Eikonal for gluon emission
    // for emissions off the hard interaction
    if (verbose >= 8) 
      printOut("TimeAntShower::acceptTrial","HyperJet mode: Subtracting Eikonal");
    antPhys -= 2.0*(1.-y12-y23)/(y12*y23);
  }

  //************************************************************************
  // Apply color (charge) factor
  //************************************************************************

  antPhys *= antPtr->getChargeFac();

  //**********************************************************************
  // Compute accept probability : pAcc = Physical/Trial 
  //**********************************************************************

  Paccept[0] = antPhys/antTrialSum;  
  
  // Verbose output
  if (verbose >= 8 || ( verbose >= 6 && Paccept[0] > 1.0 ) ) {
    printOut("TimeAntShower::acceptTrial", " ");
    cout << "   AntTrial/s = " << setprecision(6) << antTrialSum/sAnt << endl;
    cout << "   AntPhys/s  = " << setprecision(6) << antPhys/sAnt << endl;
    cout << "   Paccept    = " << setprecision(6) << Paccept[0] << endl;
    cout << "   Biasing    = " << setprecision(6) << biasFac<<endl;
  }

  //***********************************************************************
  // Choose helicities for daughter particles 
  // (so far only for massless polarized shower)
  //***********************************************************************

  double aHel = 0.0;
  if ( isPolarized ) {
    // Sum over final-state helicities
    // (Slightly inefficient to first evaluate sum and then reevaluate
    // individual terms below, but perhaps acceptable for increased 
    // clarity. Some optimization is attempted by starting with the 
    // dominant (MHV) contributions below. Further optimization could
    // be achieved by implementing some form of caching of the individual
    // contributions.)
    // Generate random number
    double randHel = rndmPtr->flat() * helSum;
    // Select helicity
    int hi(0), hj(0), hk(0);
    for (hi = hAant; abs(hi) <= 1; hi -= 2*hAant) {
      for (hk = hBant; abs(hk) <= 1; hk -= 2*hBant) {
        for (hj = hAant; abs(hj) <= 1; hj -= 2*hAant) {
          aHel = antPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
            hAant, hBant ,hi, hj, hk);
          randHel -= aHel;
          if (verbose >= 8)
            printOut("TimeAntShower::acceptTrial", "antPhys("+
              num2str(int(hAant))+" "+num2str(int(hBant))+"  -> "+num2str(hi)+
              " "+num2str(hj)+" "+num2str(hk)+") = "+num2str(aHel/sAnt)+
              ", m(ar,rb) = "+num2str(sqrt(y12ant*sAnt))+", "+
              num2str(sqrt(y23ant*sAnt))+", isSwapped = "+bool2str(isSwapped)+
              "; sum = "+num2str(helSum/sAnt));
          if (randHel < 0.) break;
        }
        if (randHel < 0.) break;
      }
      if (randHel < 0.) break;
    }
    if (verbose >= 7) 
      printOut("TimeAntShower::acceptTrial", "selected"+num2str((int)(hAant))+
        " "+num2str(int(hBant))+"  -> "+num2str(hi)+" "+num2str(hj)+" "+
        num2str(hk)+", isSwapped = "+bool2str(isSwapped)+"\n");
    // Assign helicities (taking swapped invariants into account)
    if (!isSwapped) {
      //cout<<" Selected : " <<hA<<" "<<hB<<" > "<<hi<<" "<<hj<<" "<<hk<<endl;
      trialPtr->new1.pol(hi);
      trialPtr->new2.pol(hj);
      trialPtr->new3.pol(hk);
    } 
    else {
      //cout<<" Selected : " <<hB<<" "<<hA<<" > "<<hk<<" "<<hj<<" "<<hi<<endl;
      trialPtr->new1.pol(hk);
      trialPtr->new2.pol(hj);
      trialPtr->new3.pol(hi);        
    }
  }
  // If not polarized
  else {      
    trialPtr->new1.pol(9);
    trialPtr->new2.pol(9);
    trialPtr->new3.pol(9);
    polarizedSys[iSysBranch] = false;
  }

  // Store chosen polarizations also in vector representing system state
  partsNow[antPos[0]].pol(trialPtr->new1.pol());
  partsNow[antPos[2]].pol(trialPtr->new2.pol());
  partsNow[antPos[1]].pol(trialPtr->new3.pol());

  //**********************************************************************
  // Apply alphaS choices
  //**********************************************************************

  // Reweight central accept probability with user choice   
  if (verbose >= 3 && alphaSusr > alphaTrial)
    cout<<" alphaSusr = "<<alphaSusr
	<<" alphaTrial = "<<alphaTrial
	<<" -> ratio = "<<alphaSusr/alphaTrial<<endl;
  Paccept[0] *= alphaSusr / alphaTrial;

  //**********************************************************************
  // Matrix Element Corrections
  //**********************************************************************

  // QCD order as in Born*QCD
  int nQCDnow = nBranch[iSysWin]+1;
  // Only up to matchingLO
  bool doMEC = (meCorrectionsNow && nQCDnow <= matchingLO);
  // Only for hard process
  bool isHardNow = hardSys[iSysWin];
  if (!isHardNow) doMEC = false;
  // We are already below the cutoff (only checking for last order)
  if ( (nQCDnow == matchingLO) && (q2new < pow2(matchingIRcutoff)) ) 
    doMEC = false;

  // Current number of quarks and gluon
  int nGnow  = nG[iSysWin];
  int nQQnow = nQQ[iSysWin];
  if (trialPtr->new2.id() == 21) ++nGnow; 
  else {
    ++nQQnow;
    nGnow -= 1;
  }

  // Compute matching scale dampening factor (if imposed)
  double DeltaDamp = 1.0;
  double q2damp    = q2new;
  double q2match   = q2fac;
  if (doMEC && nQCDnow >= matchingRegOrder && matchingRegOrder != 0) {
    // If matching scale is abolute use given scale
    if (matchingScaleIsAbs) q2match = pow2(matchingScale);
    // If matching scale is relative use given ratio
    else q2match = pow2(matchingScaleRatio)*q2match;
    // MatchingRegShape = 0: Step Function
    if (matchingRegShape == 0) {
      if (q2damp < q2match) {
        DeltaDamp = 0.0;
        doMEC = false;
      }
      else DeltaDamp = 1.0;
    }
    // MatchingRegShape = 1: linear in log(Q) between 1/2 QM2 < Q2 < 2 QM2
    else {
      double log2R = log(q2damp/q2match)/log(2.);
      if (log2R < -1) {
        DeltaDamp = 0.0;
        doMEC = false;
      } else if (log2R > 1.) {
        DeltaDamp = 1.0;
      } else {            
        DeltaDamp = 0.5*log2R + 0.5 ;
      }
    }
  }

  // Set all relevant properties and check if we can do the correction
  if (doMEC) {
    matrixElementCorrectionsPtr->prepMECs(MEclass[iSysWin], nQLneeded[iSysWin],
      Hint[iSysWin]);
    doMEC = matrixElementCorrectionsPtr->canDoMEC(partsNow, isHardNow);
  }

  // Compute MEC and apply to accept probability
  if (doMEC) {
    // Copy event
    Event eventNew = event;
    if (incomingType > 9) {
      // LHC-like: only keep system and beams
      eventNew.remove(3,eventNew.size()-1);
      // Append the incoming partons, reset relations to beams
      for (int i=0; i<(int)partsNow.size(); i++) if (!partsNow[i].isFinal())
        eventNew.append(partsNow[i]);
      eventNew[1].daughters(3,0); eventNew[2].daughters(4,0);
      eventNew[3].mothers(1,0);   eventNew[4].mothers(2,0);
    } else eventNew.remove(5,eventNew.size()-1);
    // Now add all final state particles and initial bosons if present
    for (int i=0; i<(int)partsNow.size(); i++) {
      if (partsNow[i].isFinal()) eventNew.append(partsNow[i]);
      else if ( (partsNow[i].idAbs() > 22) && (partsNow[i].idAbs() < 26) )
        eventNew.append(partsNow[i]);
    }
    // If no merging cluster only shower jets
    int nClustStep = ( doMergeSav ? -1 : nQCDnow-additionalJets[iSysWin] );
    PME = matrixElementCorrectionsPtr->performMECs(eventNew, this,
      ISRshowerPtr, nClustStep);
    if (verbose >= 3 && PME > 1.) {
      cout<<" PME = "<<PME<<endl;
    }
    // Modify accept probabilitiy
    Paccept[0] *= 1.0 + DeltaDamp * (PME - 1.0);
    // Fill diagnostics histos?
    if (verbose >= 3 || (verbose >= 2 && doMEC)) {
      string state = num2str(2*nQQnow,1) + "q" + num2str(nGnow,1) + "g";   
      if (trialPtr->new2.colType() == 2) state += "Emit";
      else if (trialPtr->new2.colType() == -1) state += "SplitA";
      else if (trialPtr->new2.colType() == 1) state += "SplitB";
      string HPacc = "Log10(ME/AntTrial):" + state;
      if (vinciaHistos.find(HPacc) != vinciaHistos.end())
        vinciaHistos[HPacc].fill(log10(max(TINY,Paccept[0])));
      string HqTrial = "Ln(q2trial/sSystem):" + state; 
      if (vinciaHistos.find(HqTrial) != vinciaHistos.end())
        vinciaHistos[HqTrial].fill(log(max(TINY,pow2(qNew/mSystem[iSysWin]))));
    }
  }

  //**********************************************************************
  // Print MC violations
  //**********************************************************************

  if ( verbose >= 8 ) 
    printOut("TimeAntShower::acceptTrial"," Trial Paccept = "
      +num2str(Paccept[0]));

  bool violation  = (Paccept[0] > 1.0 + TINY);
  bool negPaccept = (Paccept[0] < 0.0);

  if (violation || negPaccept) {
    int indexAccProb;
    if (meCorrectionsNow && nQCDnow <= matchingLO && isHardNow)
      indexAccProb = ( doMEC ? 1 : 2 );
    else indexAccProb = 0;
    if (violation) {
      nMCviols[indexAccProb]++;
      maxMCviols[indexAccProb] = max(maxMCviols[indexAccProb],Paccept[0]);
      // Automatically increase headroom factor when violations encountered
      setHeadroom( (iAntPhys == iQGsplit || iAntPhys == iGGsplit) ? 1 : 0, -1, 
        nG[iSysWin], nQQ[iSysWin], iSysWin, qNew, min(1.2,Paccept[0]) );
    } else if (negPaccept) {
      nNegPas[indexAccProb]++;
      minNegPas[indexAccProb] = min(minNegPas[indexAccProb],Paccept[0]);
    }
    bool report = false;
    bool isLast = false;
    // Report violations if verbose >= 3 or verbose >= 1 && fussy or
    // verbose >= 1 && negative Paccept, for first nReportMCviol times
    if ( (verbose >= 1 && Paccept[0] > pAcceptFuss && qNew > pAcceptPT0) ||
      (verbose >= 1 && negPaccept) || verbose >= 3 ) {
      nReportedMCviol++;
      if (nReportedMCviol <= nReportMCviol) {
        report = true;
        if (nReportedMCviol == nReportMCviol) isLast = true;
      }
    }
    // Always report in debugging mode
    if (report || verbose >= 4) {
      printOut("TimeAntShower::acceptTrial",
        (doMEC ? " Paccept(shower and ME, order" : " Paccept(shower, order")
	       +num2str(nQCDnow,2)+") = "+num2str(Paccept[0])+" at q = "+num2str(qNew)
        +" for gluon "+(iAntPhys<iQGsplit ? "emission" :"splitting"));
      if (verbose >= 4 || Paccept[0]>100.0) {
        if (doMEC)
          cout <<" bare PME = "<<PME<<" qDamp = "<<sqrt(q2damp)<<" qDamp/qMatch = 1/"
               <<sqrt(q2match/q2damp)<<" DeltaDamp = "<<DeltaDamp<<endl;
        cout <<" IK -> ijk = "<<trialPtr->old1.id()<<" "<<trialPtr->old2.id()
             <<" -> "<<trialPtr->new1.id()<<" "<<trialPtr->new2.id()<<" "
             <<trialPtr->new3.id()<<" mij = "<<sqrt(m2(trialPtr->new1.p()+
             trialPtr->new2.p()))<<" mjk = "<<sqrt(m2(trialPtr->new2.p()+
             trialPtr->new3.p()))<<" mijk = "<<sqrt(sAnt)<<endl;
      }
      if (isLast && verbose < 4) cout <<" Further warnings suppressed!"<<endl;
      // Print more info for bad cases
      if (verbose >= 5) trialPtr->list();
      if ((Paccept[0]>2.0 && verbose>=4) || (Paccept[0]>10.0 && verbose>=3)) {
        cout << " Momentum Listing after the branching: " << endl;
        for (int i=0; i<(int)partsNow.size(); i++) {
          cout <<" id"<<i<<(i>9 ? " = " : "  = ")<<num2str(partsNow[i].id(),7);
          if (polarizedSys[iSysBranch])
            cout <<", pol = "<<num2str((int)partsNow[i].pol(),3);
          cout <<", p = "<<partsNow[i].p();
        }
        cout << " Momentum Listing before the branching: " << endl;
        for (int i=0; i<(int)partsSav[iSysWin].size(); i++) {
          cout <<" id"<<i<<(i>9 ? " = " : "  = ")
               <<num2str(partsSav[iSysWin][i].id(),7);
          if (polarizedSys[iSysBranch])
            cout <<", pol = "<<num2str((int)partsSav[iSysWin][i].pol(),3);
          cout <<", p = "<<partsSav[iSysWin][i].p();
        }
      }
    }
  }

  //**********************************************************************
  // MC reweighting and uncertainty bands
  //**********************************************************************

  pair<double,double> accRecWeight = weightsPtr->doMCreweighting(Paccept);
  double acceptWeight = accRecWeight.first;
  double rejectWeight = accRecWeight.second;

  if (uncertaintyBands) {
    weightsPtr->doVariationsFSR(Paccept, isHardNow, doMEC, trialPtr,
      iTrial, alphaSptr, alphaSusr, mu2*kMu2Usr, mu2min, mu2freeze, alphaSmax,
      isPolarized, antPhys/sAnt/antPtr->getChargeFac(), helSum/sAnt,
      aHel/sAnt);
  }  
  
  //**********************************************************************
  // Accept / Reject step
  // Bias factors < 1 (radiation inhibition) treated by modifying pAccept.
  //**********************************************************************

  if (rndmPtr->flat() > min(1.0,biasFac)*Paccept[0]) {
    if (verbose >= 5) 
      printOut("TimeAntShower::acceptTrial",
        "=== Branching Vetoed. wPhys/wTrial = " + num2str(Paccept[0])
	       + " * biasFac = "+num2str(biasFac));

    // In case of reweighting or uncertainties, scale the weight(s)
    if (allowReweighting) weightsPtr->scaleWeightAll(rejectWeight);
    if (uncertaintyBands) weightsPtr->scaleWeightVar(Paccept, false, isHardNow,
      "FSR", (iAntPhys == iQGsplit || iAntPhys == iGGsplit) ? 1 : 0);

    // Biasing (ignores allowReweighting flag since biasing by definition
    // must reweight)
    if (biasFac != 1.0) weightsPtr->scaleWeightBiasReject(Paccept[0],biasFac);
    
    // For each failed trial, attempt to (slowly) optimize headroomfactors
    if ( trialPtr->nVeto >= 1 && optimizeHeadroom )
      setHeadroom( (iAntPhys == iQGsplit || iAntPhys == iGGsplit) ? 1 : 0, 
        -1, nG[iSysWin], nQQ[iSysWin], iSysWin, qNew, 0.9999 );

    // Count up number of vetoed branchings
    trialPtr->nVeto++;
    ++nTrialsVeto;

    // Exit.
    return false;
  }

  // In case of reweighting or uncertainties, scale the weight(s)
  if (allowReweighting) weightsPtr->scaleWeightAll(acceptWeight);
  if (uncertaintyBands) weightsPtr->scaleWeightVar(Paccept, true, isHardNow,
      "FSR", (iAntPhys == iQGsplit || iAntPhys == iGGsplit) ? 1 : 0);

  // Biasing (ignores allowReweighting flag since biasing by definition
  // must reweight)
  if (biasFac != 1.0) weightsPtr->scaleWeightBiasAccept(biasFac);

  // Fill diagnostics histos?
  if (verbose >= 3 || (verbose >= 2 && doMEC)) {
    string state = num2str(2*nQQnow,1) + "q" + num2str(nGnow,1) + "g";
    if (trialPtr->new2.colType() == 2) state += "Emit";
    else if (trialPtr->new2.colType() == -1) state += "SplitA";
    else if (trialPtr->new2.colType() == 1) state += "SplitB";
    string HqPhys = "Ln(q2/sSystem):" + state;
    if (vinciaHistos.find(HqPhys) != vinciaHistos.end()) {
      vinciaHistos[HqPhys].fill(log(max(TINY,pow2(qNew/mSystem[iSysWin]))));
    }
    string HalphaS = "alphaS:" + state;
    string HalphaSratio = "alphaSratio:" + state;
    if (doMEC) {
      string HPacc = "Log10(ME/AntPhys):" + state;
      if (vinciaHistos.find(HPacc) != vinciaHistos.end()) {
        vinciaHistos[HPacc].fill(log10(max(TINY,PME)));
      }
    }
  }

  // Verbose output
  if (verbose >= 6) 
    printOut("TimeAntShower::acceptTrial","finished --------------");

  // All clear, yey
  return true;

}
  
// ******
// 
// assignColourFlow():
// 
// Paint a trialPtr with colour-flow information,
// according to the type of branching.
// Nearest-neighbours are never assigned the same "index" (for CR).

bool TimeAntShower::assignColourFlow(Event& event, BranchelementalFSR* trialPtr) {
  bool usedColTag = false;
  // Set colour flow, deleting branched dipoles and adding new ones.
  trialPtr->new1.acol(trialPtr->old1.acol());
  trialPtr->new1.col(0);
  trialPtr->new2.acol(0);
  trialPtr->new2.col(0);
  trialPtr->new3.acol(0);
  trialPtr->new3.col(trialPtr->old2.col());
  int colOld = trialPtr->old1.col();
  // Gluon emission:
  if (trialPtr->new2.id() == 21) {
    double s12 = dot4(trialPtr->new1.p(),trialPtr->new2.p());
    double s23 = dot4(trialPtr->new2.p(),trialPtr->new3.p());
    // Who inherits the colour?
    bool inh12 = reconnectionsPtr->inherit12(s12,s23);
    // NOTE: Use lastColTag here, if branching accepted we tell
    // Pythia that we used one
    int nTag   = event.lastColTag()+1;
    usedColTag = true;
    // Check other colourtag for gluons
    int colL   = ( (trialPtr->iL != 0) ? trialPtr->old1.acol() : 0 );
    int colR   = ( (trialPtr->iR != 0) ? trialPtr->old2.col()  : 0 );
    int nextTag = 10*int(nTag/10) + 10;
    int colNew  = nextTag + int(colOld%10 + rndmPtr->flat()*8)%9 + 1;
    if (inh12) { // new1 inherits colour
      while (colNew%10 == colR%10) 
        colNew = nextTag + int(colOld%10 + rndmPtr->flat()*8)%9 + 1;   
      trialPtr->new1.col(colOld);
      trialPtr->new2.acol(colOld);
      trialPtr->new2.col(colNew);
      trialPtr->new3.acol(colNew);
    } else { // new3 inherits colour
      while (colNew%10 == colL%10) 
        colNew = nextTag + int(colOld%10 + rndmPtr->flat()*8)%9 + 1;
      trialPtr->new1.col(colNew);
      trialPtr->new2.acol(colNew);
      trialPtr->new2.col(colOld);
      trialPtr->new3.acol(colOld);
    }
  }
  // Xg -> X qbar q
  else if (trialPtr->new2.colType() == -1) {
    trialPtr->new1.col(colOld);
    trialPtr->new2.acol(colOld);
  } 
  // gX -> qbar q X
  else if (trialPtr->new2.colType() == 1) {
    trialPtr->new2.col(colOld);
    trialPtr->new3.acol(colOld);
  }
  return usedColTag;
}
    
//********* 
  
// Print header information
  
void TimeAntShower::header() {
    
  // Avoid printing header several times
  headerIsPrinted = true;

  cout<<setprecision(3);
  cout.setf(ios::left);
  cout << "\n";
  cout << " *-------  VINCIA "<<fixed<<setw(6)<<version
       <<" Global Initialization  ------------------------------------------*\n";
  cout << " | \n";
  cout << " | Vincia                               = "
       <<bool2str(settingsPtr->flag("Vincia"),9) <<"\n |\n";
    
  if (!settingsPtr->flag("Vincia")) {
    cout << " |\n *-------  End VINCIA Global Initialization  "
         <<"---------------------------------------------*\n\n";
    cout.setf(ios::right);
    return;
  }
    
  // Print header information about Matching 
  cout << " | Matrix Element Corrections           = ";
  if (meCorrections && matchingLO >= 1) {
    cout << bool2str(true,9)<<"\n"; 
    cout <<" |                 matchingFullColor    = ";
    cout << bool2str(matchingFullColor,9)<<"\n";
    cout <<" |                 LO QCD               = ";
    cout << num2str(matchingLO,9)<<"  (coupling powers requested)\n";
    cout <<" |                 regOrder             = "
         << num2str(matchingRegOrder,9)<<endl;
    if (matchingScaleIsAbs) 
      cout <<" |                 regScale             = "
           << num2str(matchingScale,9)<<endl;
    else 
      cout <<" |                 regScaleRatio        = "
           << num2str(matchingScaleRatio,9)<<endl;
    if (verbose >= 2) {
      cout <<" |                 regShape             = "
           << num2str(matchingRegShape,9)<<endl;
    }
    cout <<" |                 MG4interface         = "
         << bool2str(useMG4,9)<<"\n";
    cout <<" |                 MG5interface         = "
         << bool2str(useMG5,9)<<"\n";
    cout <<" |                 useMHVamplitudes     = "
         << bool2str(useMHV,9)<<"\n";
  }    
  else {
    cout<<bool2str(false,9)<<"\n";
  }
    
  // Print header information about shower
  cout << " |\n";      
  cout << " | Shower:         evolutionType (FSR)  = "
       << num2str(settingsPtr->mode("Vincia:evolutionType"),9)
       <<"  (1:pT, 2:mD)\n";
  cout << " |                 evolveAllInPT (ISR)  = "
       << bool2str(settingsPtr->flag("Vincia:evolveAllInPT"),9)<<"\n";
  cout << " |                 helicityShower       = "
       << bool2str(settingsPtr->flag("Vincia:helicityShower"),9)<<"\n";
  cout << " |                 nGluonToQuarkF (FSR) = "
       << num2str(nGluonToQuarkF,9)
    //<<"  (massCorrectionSplitF = "<<massCorrectionSplit<<")"
       <<"\n";
  cout << " |                 nGluonToQuarkI (ISR) = "
       << num2str(settingsPtr->mode("Vincia:nGluonToQuarkI"),9)<<"\n";
    
  // Print header information about alphaS
  cout << " |\n | Alpha_s:        alphaS(mZ)|MSbar     = "
       << num2str(alphaSvalue,9)<<"\n"
       << " |                 order                = "
       << num2str(alphaSorder,9)<<"\n";
  if (alphaSorder >= 1) {
    cout << " |                 LambdaQCD[nF]        = "
         << num2str(alphaSptr->Lambda3(),9)<<"[3] "
         << num2str(alphaSptr->Lambda4(),7)<<"[4] "
         << num2str(alphaSptr->Lambda5(),7)<<"[5] "
         << num2str(alphaSptr->Lambda6(),7)<<"[6]\n";
    if (useCMW) {
      cout << " |                 LambdaQCD[nF] (MSbar)= "
	   << num2str(vinComPtr->alphaStrong.Lambda3(),9)<<"[3] "
	   << num2str(vinComPtr->alphaStrong.Lambda4(),7)<<"[4] "
	   << num2str(vinComPtr->alphaStrong.Lambda5(),7)<<"[5] "
	   << num2str(vinComPtr->alphaStrong.Lambda6(),7)<<"[6]\n";
    }
    cout << " |                 CMW rescaling        = "
         << num2str(settingsPtr->flag("Vincia:useCMW"),9)<<"\n";
    cout << " |                 kMuF                 = "
         << num2str(settingsPtr->parm("Vincia:alphaSkMuF"),9)
         <<" (FSR emissions)\n";
    cout << " |                 kMuSplitF            = "
         << num2str(settingsPtr->parm("Vincia:alphaSkMuSplitF"),9)
         <<" (FSR splittings)\n";
    cout << " |                 kMuI                 = "
         << num2str(settingsPtr->parm("Vincia:alphaSkMuI"),9)
         <<" (ISR emissions)\n";
    cout << " |                 kMuSplitI            = "
         << num2str(settingsPtr->parm("Vincia:alphaSkMuSplitI"),9)
         <<" (ISR splittings)\n";
    cout << " |                 kMuConv              = "
         << num2str(settingsPtr->parm("Vincia:alphaSkMuConv"),9)
         <<" (ISR conversions)\n";

    cout << " |                 alphaSmuFreeze       = "
         << num2str(alphaSmuFreeze,9)<<"\n";
    cout << " |                 alphaSmax            = "
         << num2str(alphaSmax,9)<<"\n";
  }
    
  // Print header information about IR regularization
  cout << " |\n"
       << " | IR Reg.:        cutoffType           = "
       << num2str(settingsPtr->mode("Vincia:cutoffTypeF"),9)<<"\n" 
       << " |                 cutoffScaleFF        = "
       << num2str(cutoffScale,9)<<"\n"
       << " |                 cutoffScaleII        = "
       << num2str(settingsPtr->parm("Vincia:cutoffScaleII"),9)<<"\n"
       << " |                 cutoffScaleIF        = "
       << num2str(settingsPtr->parm("Vincia:cutoffScaleIF"),9)<<"\n"
       << " |                 matchingIRcutoff     = "
       << num2str(matchingIRcutoff,9)<<"\n";
         
  // Print header information from Colour Reconnections 
  reconnectionsPtr->header();
  cout<<" |\n";
      
  // Print header information about Antennae
  if (verbose >= 2) {
    for (int iAnt=0;iAnt<=4;++iAnt) {
      cout.setf(ios::left);
      cout<<setprecision(2);
      cout<<" | *------------------------------------------------------------------------\n";
      cout<<" | * "<<getAnt(iAnt)->vinciaName()<<" ["
          <<getAnt(iAnt)->humanName()<<"]"<<endl;
      // Print color/charge factor
      double chargeFac = getAnt(iAnt)->getChargeFac();
      if (chargeFac > 0.) { 
        cout<<" |   chargeFactor       = "<<fixed
            <<setw(6)<<chargeFac<<"\n";
        int kineMapType = getAnt(iAnt)->getKineMapType();
        cout<<" |   kineMapType        = "
            <<setw(6)<<kineMapType<<"\n";
        cout << " | \n";
      } else { // Switched off
        cout<<" |   OFF \n | \n";
      }
    }
  }
  // Print references
  cout <<     " |-------------------------------------------"
       << "------------------------------------------\n |\n";
  cout << " | References :"<<endl;
  cout << " |    VINCIA     : Fischer, Prestel, Ritzmann, Skands, EPJC76(2016)589, arXiv:1605.06142"<<endl;
  cout << " |    PYTHIA 8   : Sjostrand et al., CPC191(2015)159, arXiv:1410.3012"<<endl;
  if (meCorrections && matchingLO >= 1 && ( useMG4 || useMG5 ) ) {
    cout << " | The MADGRAPH interface relies on:"<<endl;
    if (useMG4) {
      cout << " |    HELAS      : Murayama et al., KEK-91-11 (1992)"<<endl;
      cout << " |    MADGRAPH 4 : Alwall et al., JHEP09(2007)028, arXiv:0706.2334"<<endl;
    }
    if (useMG5) {
      cout << " |    MADGRAPH 5 : Alwall et al., JHEP06(2011)128, arXiv:1106.0522 "<<endl;
    }
  }
  
  cout << " |\n *-------  End VINCIA FSR Global Initialization  "
       << "-----------------------------------------*\n\n";
  cout.setf(ios::right);  
}  
  
//*********
  
// Print list of antennae

void TimeAntShower::list(bool includeSaved) const {
  // Loop over dipole list and print it.
  for (int iSys = 0; iSys < (int)branchelementals.size(); iSys++) {
    for (int iAnt = 0; iAnt < (int)branchelementals[iSys].size(); ++iAnt) {
      if (branchelementals[iSys].size() == 1) branchelementals[iSys][iAnt].list(true,true);
      else if ( iAnt == 0 ) branchelementals[iSys][iAnt].list(true,false);
      else if ( iAnt == int(branchelementals[iSys].size())-1 ) branchelementals[iSys][iAnt].list(false,true);
      else branchelementals[iSys][iAnt].list();
    }
  }
  if (!includeSaved || savedBranchings.size() == 0) return;
  // Add list of saved (postponed) trial branchings
  cout<<endl<<" --------  List of saved (postponed) trial branchings --------------------------------------------------------------------------------------"<<endl;
  for (int i=0; i<(int)savedBranchings.size(); ++i)
    savedBranchings[i].list();
  cout<<endl<<" --------  End List of saved (postponed) trial branchings ----------------------------------------------------------------------------------"<<endl;
    
}

//*********
  
// Print info

void TimeAntShower::printInfo(bool pluginCall) {
  if (!isInit) return;
    
  // Weight info
  int nTotWeightsInt         = weightsPtr->nTotWeights;
  int nWeights               = weightsPtr->nWeights();
  int nNonunityInitialWeight = weightsPtr->nNonunityInitialWeight;
  int nNegativeInitialWeight = weightsPtr->nNegativeInitialWeight;
  int nNonunityWeight        = weightsPtr->nNonunityWeight;
  int nNegativeWeight        = weightsPtr->nNegativeWeight;
  double nTotWeights         = max(1.0,((double)(nTotWeightsInt)));
  vector<double> weightSum   = weightsPtr->weightSum;
  vector<double> weightMax   = weightsPtr->weightsMax;
  vector<double> weightMin   = weightsPtr->weightsMin;
    
  cout << "\n";
  if (pluginCall)
    cout << " *--------  VINCIA FSR and ISR Statistics  ------------------------------------------------------------------------*\n";
  else
    cout << " *--------  VINCIA FSR Statistics  --------------------------------------------------------------------------------*\n";
  cout << " |                                                                                                                 |\n"; 
  cout << " |                                                                        "<<setw(40)<<" "<<" |\n";
  cout << " | Total Number of (accepted) events                           = "
       <<num2str(nTotWeightsInt,9)<<setw(40)<<" "<<" |\n";
  cout << " | Number of events with nonunity initial weight               = "
       <<( (nNonunityInitialWeight <= 0) ? "     none                                " 
         : num2str(nNonunityInitialWeight,9)+" <-- (INITIALLY) WEIGHTED EVENTS" )
       <<setw(8)<<" "<<" |\n";
  cout << " | Number of events with negative initial weight               = "
       <<( (nNegativeInitialWeight <= 0) ? "     none" 
         : num2str(nNegativeWeight,9))
       <<setw(40)<<" "<<" |\n";
  cout << " | Number of events with nonunity reweight                     = "
       <<( (nNonunityWeight <= 0) ? "     none                      " 
         : num2str(nNonunityWeight,9)+" <-- REWEIGHTED EVENTS" )
       <<setw(18)<<" "<<" |\n";
  cout << " | Number of events with negative reweight                     = "
       <<( (nNegativeWeight <= 0) ? "     none" 
         : num2str(min(nNegativeWeight,nNonunityWeight),9))
       <<setw(40)<<" "<<" |\n";
  cout << " |                                                                        "<<setw(40)<<" "<<" |\n";   
  cout << " |                      weight(i)          Avg Wt   Avg Dev  rms(dev)      kUnwt          Expected effUnw          |\n";
  cout << " | This run               i =     IsUnw       <w>     <w-1>                1/<w>        Max Wt  <w>/MaxWt   Min Wt |\n";
  cout << " |"<<setw(4)<<" "<<"User settings            0 "; 
  cout << ( (abs(1.0-weightSum[0]/nTotWeights) < TINY) ? "   yes " : "    no " );
  cout << num2str(weightSum[0]/nTotWeights,9)<<" ";
  cout << num2str(weightSum[0]/nTotWeights-1.0,9)<<" ";
  cout << setw(9)<<"-"<<"  ";
  cout << ((weightSum[0] != 0.0) ? num2str(nTotWeights/weightSum[0],9): num2str(0.0,9));
  cout << num2str(weightMax[0],14)<<" ";
  cout << ((weightMax[0] != 0.0) ? num2str(weightSum[0]/nTotWeights/weightMax[0],9) : num2str(0.0,9))<<" ";
  cout << num2str(weightMin[0],9)<<" ";
  cout << "|\n";
  if (uncertaintyBands && nWeights > 1) {
    for (int iVar=1; iVar<nWeights; iVar++) {
      cout << " |"<<setw(4)<<" "<<"Var : "<<setw(12)<<weightsPtr->varLabels[iVar-1]<<"     "<<num2str(iVar,3);
      cout <<( (abs(1.0-weightSum[iVar]/nTotWeights) < TINY) ? "    yes " : "     no " );
      cout << num2str(weightSum[iVar]/nTotWeights,9)<<" ";
      cout << num2str(weightSum[iVar]/nTotWeights-1.0,9)<<" ";
      cout << setw(9)<<"-"<<"  ";
      cout << ((weightSum[iVar] != 0.0) ? num2str(nTotWeights/weightSum[iVar],9): num2str(0.0,9));
      cout << num2str(weightMax[iVar],14)<<" ";
      cout << ((weightMax[iVar] != 0.0) ? num2str(weightSum[iVar]/nTotWeights/weightMax[iVar],9): num2str(0.0,9))<<" ";
      cout << num2str(weightMin[iVar],9)<<" ";
      cout << "|\n";
    }
  }
  cout << " |                                                                        "<<setw(40)<<" "<<" |\n";
  if (pluginCall) {
    cout << " | - - - -  FSR only Statistics  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - |\n";
    cout << " |                                                                        "<<setw(40)<<" "<<" |\n";
  }
  if (verbose >= 2 && nTrialsSum > 0) {
    cout << " | Trial Veto Rates:"<<fixed<<setprecision(3)
         <<"   Hull = "<<nTrialsHull*1.0/nTrialsSum
         <<"   Massive = "<<nTrialsHullMassive*1.0/nTrialsSum
         <<"   Had M+Q = "<<nTrialsCutM*1.0/nTrialsSum<<" + "<<nTrialsCutQ*1.0/nTrialsSum
         <<"   HQ PS = "<<nClosePSforHQ*1.0/nTrialsSum
         <<"   Veto = "<<nTrialsVeto*1.0/nTrialsSum
         <<"     |"<<endl;
    cout << " |                                                                        "<<setw(40)<<" "<<" |\n";
  }
  string accProb[3] = { " pure shower,                          ", 
    " with MECs,                            ", " with planned but not carried out MECs," };
  int    nMCviol    = nMCviols[0]+nMCviols[1]+nMCviols[2];
  double maxMCviol  = max(max(maxMCviols[0],maxMCviols[1]),maxMCviols[2]);
  int    nNegPa     = nNegPas[0]+nNegPas[1]+nNegPas[2];
  double minNegPa   = min(min(minNegPas[0],minNegPas[1]),minNegPas[2]);
  if (nMCviol > 0) {
      cout<< " | Encountered "<<setw(6)<<nMCviol<<" MC violations,                         "
          <<"with a maximum violation of "<<setw(15)<<maxMCviol<<setw(10)<<" "<<" |\n";
    for (int i=0; i<3; i++) if (nMCviols[i] > 0)
      cout<< " |             "<<setw(6)<<nMCviols[i]<<accProb[i]<<" with a maximum violation of "
          <<setw(15)<<maxMCviols[i]<<setw(10)<<" "<<" |\n";
    cout<< " |                                                                        "<<setw(40)<<" "<<" |\n";
  }
  if (nNegPa > 0) {
    cout<< " | Encountered "<<setw(6)<<nNegPa<<" negative accept probabilities,         "
        <<"with a minimum value of "<<setw(15)<<minNegPa<<setw(14)<<" "<<" |\n";
    for (int i=0; i<3; i++) if (nNegPas[i] > 0)
      cout<< " |             "<<setw(6)<<nNegPas[i]<<accProb[i]<<" with a minimum value of "
          <<setw(15)<<minNegPas[i]<<setw(14)<<" "<<" |\n";
    cout<< " |                                                                        "<<setw(40)<<" "<<" |\n";
  }
  if (!pluginCall)
    cout << " *--------  End VINCIA FSR Statistics -----------------------------------------------------------------------------*\n\n";
}

} // end namespace VINCIA
